/*********************************************************************NVMH1****
File:
hufo_fur_setup.cpp

Copyright (C) 1999, 2000 NVIDIA Corporation
This file is provided without support, instruction, or implied warranty of any
kind.  NVIDIA makes no guarantee of its fitness for a particular purpose and is
not liable under any circumstances for any damages or loss whatsoever arising
from the use or inability to use this file or items derived from it.

Comments:
    This programming example shows how to make use of the NV_vertex_program
    extension in order to setup the register combiners to achieve bump mapping

Author:
    Sbastien Domin
    sdomine@nvidia.com
    
******************************************************************************/

#include <windows.h>

#include <iostream>
#include <nv_util.h>
#include <nv_algebra.h>

// Disable the STL debug information warnings
#pragma warning (disable:4786)
// use glh for easy extension
#define GLH_EXT_SINGLE_FILE
#define GLH_NVEB_USING_NVPARSE
#include <glh_nveb.h>
#include <glh_extensions.h>
#include <glh_glut.h>

#include <nvparse.h>
#include <read_text_file.h>

#include <vector>


NVEB_EFFECT_NAME("Fur Effect with Dynamic Per Pixel Lighting");
NVEB_EFFECT_VERSION("1.0");
NVEB_EFFECT_LOCATION("3rd Party\\Hair");
NVEB_EFFECT_ABOUT( 0, NULL, "Fur Effect", "NVEffectsExplained.htm#Fur_Effect");
NVEB_EFFECT_ABOUT( 1, NULL, "Fur Effect with Dynamic Per Pixel Lighting", "http://hufo.planet-d.net/fur");
NVEB_EFFECT_ABOUT( 2, NULL, "Jeremie Allard (Hufo / N.A.A.)", "http://hufo.planet-d.net/");
NVEB_EFFECT_ABOUT( 3, NULL, "NVIDIA Shader Contest", "http://www.nvidia.com//developer");

#include "hufo_fur_setup.h"

using namespace std;

// near and far clip planes
nv_scalar near_z = 1.0;
nv_scalar far_z = 100.0;
nv_scalar fovy = 60.0;

// client area size
int width;
int height;

// position of the mouse when the user clicked the button
typedef enum
{
    NV_TUMBLE,
    NV_DOLLY,
} track_modes;
vec2 mouse_start;
bool mouse_track = false;
quat track_quat;
quat delta_quat;
mat4 * track_mat;
nv_scalar * track_z;
nv_scalar track_start_z;
int spin = 1;
vec2 mouse_pt_t0;
vec2 mouse_pt_t1;
track_modes track_mode;
// modelview matrix
mat4 mv;
nv_scalar cam_z;
// wireframe
bool  fill_mode = true;
bool  vis_mode = false;
// model geometry
nv_mesh_array   mesh_array;
ase::model *    model;
// vertex program id
GLuint  vpid;
// vertex state program id
GLuint  vspid;

typedef enum
{
    NV_WIREFRAME,
    NV_VISUALIZE_TANGENT_BASIS,
    NV_LIGHT_ROTATION,
    NV_INCREASE_FUR_LAYER,
    NV_DECREASE_FUR_LAYER,
    NV_INCREASE_FUR_SIZE,
    NV_DECREASE_FUR_SIZE
} render_modes;

void init_opengl();
void display();
void idle();
void mouse(int button, int state, int x, int y);
void motion(int x, int y);
void menu(int item);
void keyboard(unsigned char key, int x, int y);
void reshape(int w, int h);
void load_model();

void GLErrorReport()
{
    GLenum errnum;
    const GLubyte *errstr;
    while (errnum = glGetError()) 
    {
        errstr = gluErrorString(errnum);
        printf((const char *)errstr);
    }
    return;
}

int main(int argc, char* argv[])
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB );
    glutInitWindowSize(640,480);
	glutCreateWindow("Fur Effect with Dynamic Per Pixel Lighting");

  	init_opengl();
    load_model();

    glutKeyboardFunc(keyboard);
    glutReshapeFunc(reshape);
  	glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);
	glutDisplayFunc(display);
	glutMainLoop();

    return 0;
}

void init_ext(const char * ext)
{
	if(!glh_init_extension(ext))
	{ 
        cerr << "Failed to initialize " << ext << "!" << endl; 
        exit(0); 
    }
}

void init_opengl()
{
    cout << glGetString(GL_EXTENSIONS) << endl;
 	init_ext("GL_ARB_multitexture");
 	init_ext("GL_VERSION_1_2");
    init_ext("GL_NV_register_combiners");
    init_ext("GL_NV_vertex_program");

    glutCreateMenu(menu);
    glutAddMenuEntry("Toggle wireframe [w]", NV_WIREFRAME);
    glutAddMenuEntry("Toggle tangent basis display [b]", NV_VISUALIZE_TANGENT_BASIS);
    glutAddMenuEntry("Toggle Light Rotation [l]", NV_LIGHT_ROTATION);
    glutAddMenuEntry("Increase Fur Layer [+]", NV_INCREASE_FUR_LAYER);
    glutAddMenuEntry("Decrease Fur Layer [-]", NV_DECREASE_FUR_LAYER);
    glutAddMenuEntry("Increase Fur Length [*]", NV_INCREASE_FUR_SIZE);
    glutAddMenuEntry("Decrease Fur Length [/]", NV_DECREASE_FUR_SIZE);
    glutAttachMenu(GLUT_RIGHT_BUTTON);

    glEnable(GL_DEPTH_TEST);
    glClearColor(.25f, .25f, .25f, 1);
    glEnable(GL_TEXTURE_2D);

    // initialize the modelview matrix to identity
    mv = mat4_id;
    delta_quat = quat_id;
    cam_z = -6.0f;
    track_mode = NV_TUMBLE;

    // create the vertex state program.
    glGenProgramsNV( 1, &vspid );
    // enable the vertex state program.

    char * sprogram_buffer = read_text_file("hufo_fur_setup.vsp");
    nvparse( sprogram_buffer, vspid );
    delete [] sprogram_buffer;

    // create the vertex program.
    glGenProgramsNV( 1, &vpid );
    // enable the vertex program.
    glBindProgramNV( GL_VERTEX_PROGRAM_NV, vpid );

    char * program_buffer = read_text_file("hufo_fur_setup.vp");
    nvparse( program_buffer );
    delete [] program_buffer;

    // initialize the vertex program paramter consistently
    // Track the concatenation of the modelview and projection matrix in registers 0-3.
    glTrackMatrixNV( GL_VERTEX_PROGRAM_NV, 0, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV );
    // Track the modelview matrix in registers 4-7.
    glTrackMatrixNV( GL_VERTEX_PROGRAM_NV, 4, GL_MODELVIEW, GL_IDENTITY_NV );
    // Track the inverse transpose of the modelview matrix in registers 8-11.
    glTrackMatrixNV( GL_VERTEX_PROGRAM_NV, 8, GL_MODELVIEW, GL_INVERSE_TRANSPOSE_NV );
    // Track the inverse modelview matrix in registers 12-15.
    glTrackMatrixNV( GL_VERTEX_PROGRAM_NV, 12, GL_MODELVIEW, GL_INVERSE_NV );
    // c[18]            contains useful constants ( 0.5, 1.0, 2.0, 0.0 )
    glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, 18, .5, 1.0, 2.0, 0.0 );  

}

unsigned char * make_normal_map(ase::texdata * image, nv_scalar scale)
{
    int i,j;
	unsigned char * nmap = new unsigned char[image->width*image->height*4];
	unsigned char * ip = nmap;
    nv_scalar unitw = nv_one / nv_scalar(image->width);
    nv_scalar unith = nv_one / nv_scalar(image->height);
    vec3 bi,ds,dt;

    nv_scalar scale_over_w_times_255 = scale / (255.f * nv_scalar(image->width));
    nv_scalar scale_over_h_times_255 = scale / (255.f * nv_scalar(image->height));
	for( j = 0; j < image->height; ++j)
	{
		for( i = 0; i < image->width; ++i)
		{
            ds.x = unitw;
            ds.y = nv_zero;
            dt.x = nv_zero;
            dt.y = unith;
            // take the first channel of the input texture
            nv_scalar h = image->pixels[(( j % image->height) * image->width + i % image->width)  * image->components];
            nv_scalar h_ds = image->pixels[(( j % image->height) * image->width + (i + 1) % image->width)  * image->components] - h;
            nv_scalar h_dt = image->pixels[((( j + 1)% image->height) * image->width + i % image->width)  * image->components] - h;
            // compute the two edges of the triangle
            ds.z = h_ds * scale_over_w_times_255;
            dt.z = h_dt * scale_over_h_times_255;
            // need to renormalize ds and dt
            normalize(ds);
            normalize(dt);
            cross(bi,ds,dt);
            normalize(bi);
            //*ip++ = 128; *ip++ = 128; *ip++ = 255;
            *ip++ = bi.x * 127 + 128;
            *ip++ = bi.y * 127 + 128;
            *ip++ = bi.z * 127 + 128;
            *ip++ = (i&j&1)?255:0;
		}
	}
    free(image->pixels);
    image->pixels = nmap;
    image->components = 4;
    image->pixelfmt = GL_RGBA;
    return nmap;
}

// looks for the materials that have a bump map.
// converts the bump map into normal map and upload
// the normal map to gl, while updating the material
// reference
void load_bump_tex(ase::model * model)
{
    ase::mat_it it = model->mat.begin();
    while(it != model->mat.end())
    {
        ase::matobj * matobj = (*it);
        ase::map_it map_it = matobj->map_bump.begin();
        while (map_it != matobj->map_bump.end())
        {
            ase::tex_it tex_it = model->tex.find((*map_it)->bitmap);
            if (tex_it != model->tex.end())
            {
                ase::texdata * txdata = (*tex_it).second;
                glBindTexture(GL_TEXTURE_2D, txdata->textid);
                make_normal_map(txdata, 10.0);
    	        glTexImage2D(GL_TEXTURE_2D, 0, txdata->pixelfmt, txdata->width, txdata->height, 0, txdata->pixelfmt, GL_UNSIGNED_BYTE, txdata->pixels);
                glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
                glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
            }
            ++map_it;
        }
        ++it;
    }
}

void load_tex(const ase::model * model)
{
    ase::tex_cit it = model->tex.begin();
    while(it != model->tex.end())
    {
        ase::texdata * txdata = (*it).second;
        glGenTextures(1,&txdata->textid);
        glBindTexture(GL_TEXTURE_2D, txdata->textid);

        switch(txdata->components)
        {
            case 1:
                txdata->pixelfmt = GL_INTENSITY;
                break;
            case 3:
                txdata->pixelfmt = GL_RGB;
                break;
            case 4:
                txdata->pixelfmt = GL_RGBA;
                break;
            default:
                assert(false);
                break;
        }
        gluBuild2DMipmaps(GL_TEXTURE_2D, 
                   txdata->pixelfmt, 
                   txdata->width, txdata->height,
                   txdata->pixelfmt, 
                   GL_UNSIGNED_BYTE, txdata->pixels);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        ++it;
    }
}

void load_model()
{
    mat3 basis;
    vec3 face_normal;
    int j,j_times_3;

    unsigned int size;
    unsigned char * buf;
    // Load the .ASE files from ZIP volume
    //buf = unzip::open("./../../data/models/ase/demo_models.zip","rap.ase", &size);
    buf = unzip::open("./../../data/models/ase/hufo_fur_models.zip","bunny.ase", &size);
    if (buf == NULL)
    {
        cerr << "Failed to find the ASE model" << endl;
		MessageBox(GetDesktopWindow(), "Couldn't load model", "Error", MB_OK);
        exit(0);
    }

    model = ase::load((char *)buf,size, 0.1);
    delete [] buf;

    //model = ase::load("c:\\bunny.ase", 0.1);
    // Load the textures
    ase::load_tex("./../../data/images/", model);
    load_tex(model);
    load_bump_tex(model);

    hufo_fur_texture_init();

    // Process the ase::model
    ase::geom_it it = model->geom.begin();
    while (it != model->geom.end())
    {
        ase::geomobj * geom = *it;
        // we need vertices and texture coordinates
        // in order to construct the tangent basis
        //  required for bump mapping...
        if (geom->numv && geom->numt)
        {
            nv_mesh * mesh = new nv_mesh;
            
            mesh->name = geom->name;
            mesh->numv = geom->numv;
            mesh->v = new float[mesh->numv * 3];
            mesh->tangent = new float[mesh->numv * 3];
            mesh->binormal = new float[mesh->numv * 3];
            mesh->normal = new float[mesh->numv * 3];
            memset(mesh->tangent,0, mesh->numv * sizeof(float) * 3);
            memset(mesh->binormal,0, mesh->numv * sizeof(float) * 3);
            memset(mesh->normal,0, mesh->numv * sizeof(float) * 3);
            memcpy(mesh->v, geom->v, mesh->numv * sizeof(float) * 3);

            mesh->t = new float[mesh->numv * 2];

            mesh->numf = geom->numf;
            mesh->f = new int[mesh->numf * 3];
            memcpy(mesh->f, geom->f, mesh->numf * sizeof(int) * 3);

            mesh->matidx = geom->matidx;

            for (j = 0; j < mesh->numf; ++j)
            {
                j_times_3 = j * 3;
                int v0_idx = geom->f[j_times_3];
                int v1_idx = geom->f[j_times_3 + 1];
                int v2_idx = geom->f[j_times_3 + 2];
                int n0_idx = geom->fvn[j_times_3];
                int n1_idx = geom->fvn[j_times_3 + 1];
                int n2_idx = geom->fvn[j_times_3 + 2];
                int t0_idx = geom->tf[j_times_3];
                int t1_idx = geom->tf[j_times_3 + 1];
                int t2_idx = geom->tf[j_times_3 + 2];

                float  * n0 = &geom->n[n0_idx * 3];
                float  * n1 = &geom->n[n1_idx * 3];
                float  * n2 = &geom->n[n2_idx * 3];

                // Compute the face normal based on vertex normals
                face_normal.x = n0[0] + n1[0] + n2[0];
                face_normal.y = n0[1] + n1[1] + n2[1];
                face_normal.z = n0[2] + n1[2] + n2[2];

                normalize(face_normal);

                mesh->t[v0_idx * 2] = geom->t[t0_idx * 3];
                mesh->t[v0_idx * 2 + 1] = geom->t[t0_idx * 3 + 1];
                mesh->t[v1_idx * 2] = geom->t[t1_idx * 3];
                mesh->t[v1_idx * 2 + 1] = geom->t[t1_idx * 3 + 1];
                mesh->t[v2_idx * 2] = geom->t[t2_idx * 3];
                mesh->t[v2_idx * 2 + 1] = geom->t[t2_idx * 3 + 1];

                int v0_times_3 = v0_idx * 3;
                int v1_times_3 = v1_idx * 3;
                int v2_times_3 = v2_idx * 3;
                vec3 v0(&geom->v[v0_times_3]);
                vec3 v1(&geom->v[v1_times_3]);
                vec3 v2(&geom->v[v2_times_3]);
                vec2 t0(&mesh->t[v0_idx * 2]);
                vec2 t1(&mesh->t[v1_idx * 2]);
                vec2 t2(&mesh->t[v2_idx * 2]);
                
                tangent_basis( basis, v0, v1, v2, t0, t1, t2, face_normal);

                mesh->tangent[v0_times_3] += basis.a00;
                mesh->tangent[v0_times_3 + 1] += basis.a01;
                mesh->tangent[v0_times_3 + 2] += basis.a02;
                
                mesh->binormal[v0_times_3] += basis.a10;
                mesh->binormal[v0_times_3 + 1] += basis.a11;
                mesh->binormal[v0_times_3 + 2] += basis.a12;

                //mesh->normal[v0_times_3] += basis.a20;
                //mesh->normal[v0_times_3 + 1] += basis.a21;
                //mesh->normal[v0_times_3 + 2] += basis.a22;
                mesh->normal[v0_times_3] += n0[0];
                mesh->normal[v0_times_3 + 1] += n0[1];
                mesh->normal[v0_times_3 + 2] += n0[2];

                mesh->tangent[v1_times_3] += basis.a00;
                mesh->tangent[v1_times_3 + 1] += basis.a01;
                mesh->tangent[v1_times_3 + 2] += basis.a02;
                
                mesh->binormal[v1_times_3] += basis.a10;
                mesh->binormal[v1_times_3 + 1] += basis.a11;
                mesh->binormal[v1_times_3 + 2] += basis.a12;

                //mesh->normal[v1_times_3] += basis.a20;
                //mesh->normal[v1_times_3 + 1] += basis.a21;
                //mesh->normal[v1_times_3 + 2] += basis.a22;
                mesh->normal[v1_times_3] += n1[0];
                mesh->normal[v1_times_3 + 1] += n1[1];
                mesh->normal[v1_times_3 + 2] += n1[2];

                mesh->tangent[v2_times_3] += basis.a00;
                mesh->tangent[v2_times_3 + 1] += basis.a01;
                mesh->tangent[v2_times_3 + 2] += basis.a02;
                
                mesh->binormal[v2_times_3] += basis.a10;
                mesh->binormal[v2_times_3 + 1] += basis.a11;
                mesh->binormal[v2_times_3 + 2] += basis.a12;

                //mesh->normal[v2_times_3] += basis.a20;
                //mesh->normal[v2_times_3 + 1] += basis.a21;
                //mesh->normal[v2_times_3 + 2] += basis.a22;
                mesh->normal[v2_times_3] += n2[0];
                mesh->normal[v2_times_3 + 1] += n2[1];
                mesh->normal[v2_times_3 + 2] += n2[2];
            }
            for (j = 0; j < mesh->numv; ++j)
            {
                j_times_3 = j * 3;
                vec3 tangent(&mesh->tangent[j_times_3]);
                normalize(tangent);
                mesh->tangent[j_times_3] = tangent.x;
                mesh->tangent[j_times_3 + 1] = tangent.y;
                mesh->tangent[j_times_3 + 2] = tangent.z;

                vec3 binormal(&mesh->binormal[j_times_3]);
                normalize(binormal);
                mesh->binormal[j_times_3] = binormal.x;
                mesh->binormal[j_times_3 + 1] = binormal.y;
                mesh->binormal[j_times_3 + 2] = binormal.z;

                vec3 normal(&mesh->normal[j_times_3]);
                normalize(normal);
                mesh->normal[j_times_3] = normal.x;
                mesh->normal[j_times_3 + 1] = normal.y;
                mesh->normal[j_times_3 + 2] = normal.z;
            }
            mesh_array.push_back(mesh);
        }
        ++it;
    }
}

bool light_rotate=false;
int nlayer=10;
float fursize=0.15f;

void menu(int item)
{
    switch (item) 
    {
    case NV_VISUALIZE_TANGENT_BASIS:
        keyboard('b', 0, 0);
        break;
    case NV_WIREFRAME:
        keyboard('w', 0, 0);
        break;
    case NV_LIGHT_ROTATION:
        keyboard('l', 0, 0);
        break;
    case NV_INCREASE_FUR_LAYER:
        keyboard('+', 0, 0);
        break;
    case NV_DECREASE_FUR_LAYER:
        keyboard('-', 0, 0);
        break;
    case NV_INCREASE_FUR_SIZE:
        keyboard('*', 0, 0);
        break;
    case NV_DECREASE_FUR_SIZE:
        keyboard('/', 0, 0);
        break;
    }
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
    if (key == 27) exit(0);
    if (key == 'w' || key == 'W')
    {
        fill_mode = !fill_mode;
        glPolygonMode(GL_FRONT_AND_BACK, fill_mode ? GL_FILL : GL_LINE);
    }
    if (key == 'b' || key == 'B')
    {
        vis_mode = !vis_mode;
    }
    if (key == 'l' || key == 'L')
    {
        light_rotate=!light_rotate;
    }
    if (key == '+')
    {
        ++nlayer;
    }
    if (key == '-' && nlayer>0)
    {
        --nlayer;
    }
    if (key == '*')
    {
      fursize*=1.2f;
    }
    if (key == '/')
    {
      fursize/=1.2f;
    }
}

void mouse(int button, int state, int x, int y)
{
    if (button != GLUT_LEFT_BUTTON) return;
    
    mouse_pt_t1.x = x;
    mouse_pt_t1.y = y;
    if (state == GLUT_DOWN) 
    {
        quat tmp_quat;

        if (glutGetModifiers() == GLUT_ACTIVE_CTRL)
        {
            track_mode = NV_DOLLY;
            track_start_z = cam_z;
            track_z = &cam_z;
        }
        else
        {
            track_mode = NV_TUMBLE;
            track_mat = &mv;
            track_mat->get_rot(track_quat);
        }

        mouse_start.x = x;
        mouse_start.y = y;
        mouse_pt_t0.x = x;
        mouse_pt_t0.y = y;
        mouse_track = true;
    }
    else
    {
        mouse_track = false;
        if (track_mode == NV_TUMBLE && spin)
        {
            if (mouse_pt_t0.x - x > 2 || mouse_pt_t0.x - x < -2 || mouse_pt_t0.y - y > 2 || mouse_pt_t0.y - y < -2 )
            {
                vec2 pt1( ( nv_two * mouse_pt_t0.x - width) / (nv_two * width),
                          ( height - nv_two * mouse_pt_t0.y ) / (nv_two * height));
                vec2 pt2( ( nv_two * x - width) / (nv_two * width),
                          ( height - nv_two * y  ) / (nv_two * height));

                trackball( delta_quat, pt1, pt2, 0.8);
            }
            else
                delta_quat = quat_id;
        }
    }
}

void motion(int x, int y)
{
    if (mouse_track) 
    {
        quat curquat, tmpquat;
        vec2 pt1( ( nv_two * mouse_start.x - width) / width,
                  ( height - nv_two * mouse_start.y ) / height);
        vec2 pt2( ( nv_two * x - width) / width,
                  ( height - nv_two * y  ) / height);

        if (track_mode == NV_DOLLY)
        {
            *track_z = track_start_z + nv_scalar(y - mouse_start.y)/ height * ( far_z - near_z);
            *track_z = nv_clamp(*track_z,-far_z,-near_z);
        }
        else
        {
            trackball( tmpquat, pt1, pt2, 0.8);
            add_quats( curquat, tmpquat, track_quat);
            track_mat->set_rot( curquat );
        }

        glutPostRedisplay();
        mouse_pt_t0.x = mouse_pt_t1.x;
        mouse_pt_t0.y = mouse_pt_t1.y;
        mouse_pt_t1.x = x;
        mouse_pt_t1.y = y;
    }
}

void reshape(int w, int h)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( fovy, (GLdouble)w/(GLdouble)h, near_z, far_z);
    width = w; height = h;
}

void idle()
{
    if (mouse_track == false && spin && track_mat)
    {
        quat tmpquat;
        track_mat->get_rot(track_quat);
        add_quats(tmpquat, delta_quat, track_quat );
        normalize(tmpquat);
        track_mat->set_rot(tmpquat);
    }
    glutPostRedisplay();
}

void set_mesh_mat(const nv_mesh * pmesh,const ase::model * model)
{
    ase::matobj * matobj = model->mat[pmesh->matidx];
    ase::map_it map_it = matobj->map_bump.begin();
    if (map_it != matobj->map_bump.end())
    {
        ase::tex_cit tex_it = model->tex.find((*map_it)->bitmap);
        if (tex_it != model->tex.end())
        {
            ase::texdata * txdata = (*tex_it).second;
            glBindTexture(GL_TEXTURE_2D, txdata->textid);
        }
    }
}

void set_reg_comb_bump()
{
    glEnable(GL_REGISTER_COMBINERS_NV);

    nvparse(   
    "!!RC1.0\n"
    "/* register combiner for bump lighting */\n"
    "/* col0=L col1=H tex0=N tex1=C */\n"
    "/* output: amb + (L.N)*diff + spec*(H.N)^2 */\n"
    "const0 =(0.2,0.2,0.2,1.0); /* ambient, alpha=specular level */ \n"
    "const1 =(0.4,0.35,0.2,0.75); /* diffuse, alpha=specular shininess */ \n"
    "{\n"
    "	rgb {\n"
    "		col0 = expand(col0) . expand(tex0);\n"
    "		col1 = expand(col1) . expand(tex0);\n"
    "	}\n"
    "	alpha {\n"
    "		spare0 =  col0 * tex0;\n"
    "	}\n"
    "}\n"
    "{\n"
    "	rgb {\n"
    "		discard = const0 * unsigned_invert(zero);\n"
    "		discard = const1 * unsigned(col0);\n"
    "		col0 = sum();\n"
    "	}\n"
    "	alpha {\n"
    "		discard = unsigned_invert(col1.b) * const1.a;\n"
    "		discard = unsigned_invert(col1.b) * const1.a;\n"
    "		col1 = sum(); \n"
    "		scale_by_four(); \n"
    "	}\n"
    "}\n"

    "final_product = unsigned_invert(col1.a) * unsigned_invert(col1.a);\n"
    "out.rgb = lerp(const0.a,final_product,zero) + col0;\n"
    "out.a = spare0.a;\n"
    );

    static bool first=true;
    if (first)
    {
        for (char ** errors=(char**)nvparse_get_errors(); *errors; errors++)
        {
            fprintf( stderr, *errors);
            fprintf( stderr, "\n");
        }
        first=false;
    }

}

void set_reg_comb_fur()
{
    glEnable(GL_REGISTER_COMBINERS_NV);
    nvparse(   
    "!!RC1.0\n"
    "/* register combiner for fur lighting */\n"
    "/* col0=L col1=H tex0=N tex1=C */\n"
    "/* output: sin(acos(L.N))*tex1 + sin(acos(H.N))^2 */\n"
    "/* sin(acos(x)) approximated by 1-x */\n"
#if FUR_COLOR_TEXTURE==1
    "const0 =(1,1,1,0.0); /* alpha=1-specular level */ \n"
#else
    "const0 =(0.0,0.8,1.0,0.0); /* diffuse color & alpha=1-specular level */ \n"
#endif
    "{\n"
    "	rgb { /* first part of specular sinus computation */ \n"
    "		spare0 =  expand(col1) . expand(tex0);\n"
    "		spare1 = -expand(col1) . expand(tex0);\n"
    "		scale_by_two();\n"
    "	}\n"
    "	alpha {\n"
    "		spare0 =  col0 * tex0;\n"
    "	}\n"
    "}\n"
    "{\n"
    "	rgb { /* first part of diffuse sinus computation */ \n"
    "		spare0 =  expand(col0) . expand(tex0);\n"
    "		col1   = -expand(col0) . expand(tex0);\n"
    "	}\n"
    "	alpha { /* second part of specular sinus computation */ \n"
    "		discard = unsigned_invert(spare0.b) * unsigned_invert(spare1.b);\n"
    "		discard = -const0.a * unsigned_invert(zero);\n"
    "		spare1 = sum();\n"
    "	}\n"
    "}\n"
    "final_product = unsigned(spare1.a) * unsigned(spare1.a); /* increment specular power */ \n"
    "/* color_sum = second part of diffuse sinus computation */\n"
    //"out.rgb = lerp(tex1,unsigned_invert(color_sum),zero) + final_product;\n"
    //"out.rgb = lerp ( tex1 , unsigned_invert(color_sum) , zero ) + final_product;\n"
#if FUR_COLOR_TEXTURE==1
    "out.rgb = lerp ( tex1 , unsigned_invert ( color_sum ) , zero ) + final_product;\n"
#else
    "out.rgb = lerp ( const0 , unsigned_invert ( color_sum ) , zero ) + final_product;\n"
#endif
    //"out.rgb = color_sum + final_product;\n"
    //"out.rgb = lerp(tex1,color_sum,zero);\n"
    "out.a = spare0.a;\n"
    );

    static bool first=true;
    if (first)
    {
        for (char ** errors=(char**)nvparse_get_errors(); *errors; errors++)
        {
            fprintf( stderr, *errors);
            fprintf( stderr, "\n");
        }
        first=false;
    }

	glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, 2);
	{
		GLfloat constColor0[] = { 0.0, 0.8, 1.0, 0.0f};

		glCombinerParameterfvNV(GL_CONSTANT_COLOR0_NV, constColor0);
	}
	{
		GLfloat constColor1[] = { 0, 0, 0, 0};
		glCombinerParameterfvNV(GL_CONSTANT_COLOR1_NV, constColor1);
	}
	glCombinerParameteriNV(GL_COLOR_SUM_CLAMP_NV, GL_TRUE);
	glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV, GL_SECONDARY_COLOR_NV, GL_EXPAND_NORMAL_NV, GL_RGB);
	glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0_ARB, GL_EXPAND_NORMAL_NV, GL_RGB);
	glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_C_NV, GL_SECONDARY_COLOR_NV, GL_EXPAND_NEGATE_NV, GL_RGB);
	glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_D_NV, GL_TEXTURE0_ARB, GL_EXPAND_NORMAL_NV, GL_RGB);
	glCombinerOutputNV(GL_COMBINER0_NV, GL_RGB, GL_SPARE0_NV, GL_SPARE1_NV, GL_DISCARD_NV, GL_SCALE_BY_TWO_NV, GL_NONE, GL_TRUE, GL_TRUE, GL_FALSE);
	glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_A_NV, GL_PRIMARY_COLOR_NV, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_B_NV, GL_TEXTURE0_ARB, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_D_NV, GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	glCombinerOutputNV(GL_COMBINER0_NV, GL_ALPHA, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);
	glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV, GL_PRIMARY_COLOR_NV, GL_EXPAND_NORMAL_NV, GL_RGB);
	glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0_ARB, GL_EXPAND_NORMAL_NV, GL_RGB);
	glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_C_NV, GL_PRIMARY_COLOR_NV, GL_EXPAND_NEGATE_NV, GL_RGB);
	glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_D_NV, GL_TEXTURE0_ARB, GL_EXPAND_NORMAL_NV, GL_RGB);
	glCombinerOutputNV(GL_COMBINER1_NV, GL_RGB, GL_SPARE0_NV, GL_SECONDARY_COLOR_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_TRUE, GL_TRUE, GL_FALSE);
	glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_A_NV, GL_SPARE0_NV, GL_UNSIGNED_INVERT_NV, GL_BLUE);
	glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_B_NV, GL_SPARE1_NV, GL_UNSIGNED_INVERT_NV, GL_BLUE);
	glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_C_NV, GL_CONSTANT_COLOR0_NV, GL_SIGNED_NEGATE_NV, GL_ALPHA);
	glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_D_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_ALPHA);
	glCombinerOutputNV(GL_COMBINER1_NV, GL_ALPHA, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE1_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);
#if FUR_COLOR_TEXTURE==1
	glFinalCombinerInputNV(GL_VARIABLE_A_NV, GL_TEXTURE1_ARB, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
#else
	glFinalCombinerInputNV(GL_VARIABLE_A_NV, GL_CONSTANT_COLOR0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
#endif
	glFinalCombinerInputNV(GL_VARIABLE_B_NV, GL_SPARE0_PLUS_SECONDARY_COLOR_NV, GL_UNSIGNED_INVERT_NV, GL_RGB);
	glFinalCombinerInputNV(GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	glFinalCombinerInputNV(GL_VARIABLE_D_NV, GL_E_TIMES_F_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	//glFinalCombinerInputNV(GL_VARIABLE_D_NV, GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
	glFinalCombinerInputNV(GL_VARIABLE_E_NV, GL_SPARE1_NV, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	glFinalCombinerInputNV(GL_VARIABLE_F_NV, GL_SPARE1_NV, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
	glFinalCombinerInputNV(GL_VARIABLE_G_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);

}

void render_mesh(const nv_mesh * pmesh, const ase::model * model, const vec3 & eye, const vec3 & light)
{
    int i_times_3;
    glActiveTextureARB(GL_TEXTURE0_ARB);
    set_mesh_mat(pmesh,model);

    glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, 16, light.x, light.y, light.z, 1.0 ); // light

    float lightpos[4] = { light.x, light.y, light.z, 1.0f };
    glExecuteProgramNV( GL_VERTEX_STATE_PROGRAM_NV, vspid, lightpos );

    //float furparams[4]={0.0f,0.0f,0.0f,0.0f};
    
    glEnable( GL_VERTEX_PROGRAM_NV );

    glVertexAttribPointerNV( 8, 2, GL_FLOAT, 0, pmesh->t );
    glVertexAttribPointerNV( 3, 3, GL_FLOAT, 0, pmesh->binormal );
    glVertexAttribPointerNV( 2, 3, GL_FLOAT, 0, pmesh->normal );
    glVertexAttribPointerNV( 1, 3, GL_FLOAT, 0, pmesh->tangent );
    glVertexAttribPointerNV( 0, 3, GL_FLOAT, 0, pmesh->v );

    glEnableClientState( GL_VERTEX_ATTRIB_ARRAY8_NV );
    glEnableClientState( GL_VERTEX_ATTRIB_ARRAY3_NV );
    glEnableClientState( GL_VERTEX_ATTRIB_ARRAY2_NV );
    glEnableClientState( GL_VERTEX_ATTRIB_ARRAY1_NV );
    glEnableClientState( GL_VERTEX_ATTRIB_ARRAY0_NV );

    set_reg_comb_bump();

    //hufo_fur_texture_use(0);

    glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, 19, 0.0f, 1.0f, 1.0f, 1.0f ); // fur & texture scale

    glDrawElements( GL_TRIANGLES, pmesh->numf * 3, GL_UNSIGNED_INT, pmesh->f ); 


    float dsize=fursize/nlayer;

    set_reg_comb_fur();

    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glDepthMask(0);

    for (int i=0;i<nlayer;i++)
    {
      hufo_fur_texture_use((i*TNLAYER)/nlayer);

      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, 19, dsize*(i+1), 1.0f //float(nlayer-i)/float(nlayer) // fur
                                                      ,10.0f, 1.0f ); // texture scale
      glDrawElements( GL_TRIANGLES, pmesh->numf * 3, GL_UNSIGNED_INT, pmesh->f ); 
    }
    glDepthMask(1);
    glDisable(GL_BLEND);

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);

    glDisableClientState( GL_VERTEX_ATTRIB_ARRAY8_NV );
    glDisableClientState( GL_VERTEX_ATTRIB_ARRAY3_NV );
    glDisableClientState( GL_VERTEX_ATTRIB_ARRAY2_NV );
    glDisableClientState( GL_VERTEX_ATTRIB_ARRAY1_NV );
    glDisableClientState( GL_VERTEX_ATTRIB_ARRAY0_NV );

    glDisable( GL_VERTEX_PROGRAM_NV );
    glDisable(GL_REGISTER_COMBINERS_NV);

    if (vis_mode)
    {
        float scale = 0.15f;
        glBegin(GL_LINES);
        glColor3f(1.0,0.0,0.0);
        for (int i = 0; i < pmesh->numv; ++i)
        {
            i_times_3 = i * 3;
            glVertex3fv(&pmesh->v[i_times_3]);
            glVertex3f(pmesh->v[i_times_3] + pmesh->normal[i_times_3] * scale,
                pmesh->v[i_times_3 + 1] + pmesh->normal[i_times_3 + 1] * scale,
                pmesh->v[i_times_3 + 2] + pmesh->normal[i_times_3 + 2] * scale);
        }
        glColor3f(0.0,1.0,0.0);
        for (i = 0; i < pmesh->numv; ++i)
        {
            i_times_3 = i * 3;
            glVertex3fv(&pmesh->v[i_times_3]);
            glVertex3f(pmesh->v[i_times_3] + pmesh->binormal[i_times_3] * scale,
                pmesh->v[i_times_3 + 1] + pmesh->binormal[i_times_3 + 1] * scale,
                pmesh->v[i_times_3 + 2] + pmesh->binormal[i_times_3 + 2] * scale);
        }
        glColor3f(0.0,0.0,1.0);
        for (i = 0; i < pmesh->numv; ++i)
        {
            i_times_3 = i * 3;
            glVertex3fv(&pmesh->v[i_times_3]);
            glVertex3f(pmesh->v[i_times_3] + pmesh->tangent[i_times_3] * scale,
                pmesh->v[i_times_3 + 1] + pmesh->tangent[i_times_3 + 1] * scale,
                pmesh->v[i_times_3 + 2] + pmesh->tangent[i_times_3 + 2] * scale);
        }
        glEnd();
    }
}

void render(const ase::model * model, const vec3 & eye, const vec3 & light)
{
    nv_mesh_cit it = mesh_array.begin();
    while (it != mesh_array.end())
    {
        render_mesh(*it,model,eye,light);
        ++it;
    }
}

void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(0, 0, width, height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0, 0, cam_z);
    glMultMatrixf(mv.mat_array);
    vec3 eye = vec3_null;

    static float light_ang=0;

    if (light_rotate)
      light_ang+=3.141593f/100;

    vec3 light(10.0*cos(light_ang),10.0,10.0*sin(light_ang));

    render(model, eye, light);

    glutSwapBuffers();
}
