/*
 * Non-Photorealistic Rendering
 * Szczepan Kuzniarz, June 2001
 */


#ifdef WIN32
# include <windows.h>
#endif

#include <iostream>

#define GLH_EXT_SINGLE_FILE
#define GLH_NVEB_USING_NVPARSE

#include <glh_nveb.h>
#include <glh_extensions.h>
#include <glh_glut.h>
#include <glh_obs.h>
#include <glh_cube_map.h>

#include <array_texture.h>
#include <nvparse.h>

#include "teapot_lowres.h"
#include "gui_interactor.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

using namespace std;
using namespace glh;


NVEB_EFFECT_NAME( "Non-Photorealistic Rendering" );
NVEB_EFFECT_VERSION( "1.0" );
NVEB_EFFECT_LOCATION( "3rd Party\\Stylistic" );
NVEB_EFFECT_ABOUT( 0, NULL, "Non-Photorealistic Rendering", "NVEffectsExplained.htm#npr");
NVEB_EFFECT_ABOUT( 1, "Date", "June 2001", NULL );
NVEB_EFFECT_ABOUT( 2, "Author", "Szczepan Kuzniarz", "http://gemini.wcss.wroc.pl/~sis/" );

// we will use 2 vertex program ...
GLuint programId[2];

// vertex program parameters
#define LIGHT_DIRECTION_PARAM 12
#define COOL_COLOR_PARAM      13
#define WARM_COLOR_PARAM      14
#define SPECULAR_COLOR_PARAM  16

// ... and 2 register combiners
GLuint combinerId;


// cube maps used in silhouette rendering
tex_object_cube_map normalizationMap;
tex_object_cube_map warpZMap;


// tunable parameters (see 'Description.txt')
float lodBias = 3.0;
float alphaScale = 0.45;
float alpha_power = 1.0;
bool edges = true;

GLfloat warm = 0.4, cool = 0.65;
GLfloat alpha = 0.35, beta = 0.6;
GLfloat r = 1.0, g = 0.2, b = 0.2;


// GLUT functions
glut_callbacks cb;
glut_perspective_reshaper reshaper;
glut_simple_mouse_interactor mouse;
gui_interactor gui;




// GUI elements ids/shortcut keys
#define COOL_UP_ID    'C'
#define COOL_DOWN_ID  'c'
#define WARM_UP_ID    'W'
#define WARM_DOWN_ID  'w'
#define DARK_UP_ID    'A'
#define DARK_DOWN_ID  'a'
#define LIGHT_UP_ID   'B'
#define LIGHT_DOWN_ID 'b'
#define SPECULAR_ID   ' '
#define COLOR_ID      'o'
#define EDGES_ID      'e'
#define LOD_UP_ID     'L'
#define LOD_DOWN_ID   'l'
#define SCALE_UP_ID   'S'
#define SCALE_DOWN_ID 's'


// GUI objects
colorGradient *cool2WarmGradient;
colorGradient *dark2LightGradient;
colorGradient *finalGradient;

checkBox *colorCheck;
checkBox *specularCheck;
checkBox *edgesCheck;

label *lodLabel;
label *scaleLabel;




// This class comes form 'Mipmap-based toon shading' effect,
// it helps generating cubemaps
struct alpha_vector {
  typedef GLfloat Type;
  int components;
  GLenum type;
  GLenum format;
  GLfloat a;
  alpha_vector( float alpha = 1.0 ):
    components( 4 ),
    format( GL_RGBA ),
    type( GL_FLOAT ),
    a( alpha ) {
    }
	
  void operator()( const vec3f &v, Type *t ) {
    vec3f v2 = v;
    v2 *= 0.5f;
    v2 += .5;
    t[0] = v2[0];
    t[1] = v2[1];
    t[2] = v2[2];
    t[3] = a;
  }
};


// Two functions below generate cubemaps. There is no texture
// coordinates setup, because we use vertex programs.

void createNormalizationMap() {
  glActiveTextureARB( GL_TEXTURE0_ARB );
  normalizationMap.bind();
  make_cube_map( alpha_vector(), GL_RGB8, 32, 0 );
  make_cube_map( alpha_vector(), GL_RGB8, 16, 1 );
  make_cube_map( alpha_vector(), GL_RGB8,  8, 2 );
  make_cube_map( alpha_vector(), GL_RGB8,  4, 3 );
  make_cube_map( alpha_vector(), GL_RGB8,  2, 4 );
  make_cube_map( alpha_vector(), GL_RGB8,  1, 5 );
  normalizationMap.parameter( GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
  normalizationMap.parameter( GL_TEXTURE_MAX_ANISOTROPY_EXT, 2.0f );
  normalizationMap.parameter( GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
  normalizationMap.parameter( GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
  normalizationMap.parameter( GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );
  normalizationMap.enable();
}


void createWarpZMap() {
  glActiveTextureARB( GL_TEXTURE1_ARB );
  warpZMap.bind();
  float a = 0.0;
  for( int i = 0 ; i < 8 ; i++ ) {
    make_cube_map( alpha_vector( exp( alpha_power * log( a ) ) ), GL_RGBA8, 1 << (7 - i), i );
    a += 0.2;
    if( a > 1.0 ) {
      a = 1.0;
    };
  };
  warpZMap.parameter( GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
  warpZMap.parameter( GL_TEXTURE_MAX_ANISOTROPY_EXT, 2.0f );
  warpZMap.parameter( GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
  warpZMap.parameter( GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
  warpZMap.parameter( GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );
  glTexEnvf( GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, lodBias ); 
  warpZMap.enable();
}


// This function creates two register combiners. The thirst one
// takes into account object color (col0 register), the second
// one doesn't.

void createCombiners() {
  combinerId = glGenLists( 2 );

  glNewList( combinerId, GL_COMPILE );
  nvparse(
    "!!RC1.0\n"
    "{\n"
    "  rgb {\n"
    "    spare0 = expand(tex0) . expand(tex1);\n"
    "  }\n"
    "  alpha {\n"
    "    spare0 = unsigned_invert(tex1) * const0;\n"
    "  }\n"
    "}\n"
    "{\n"
    "  rgb {\n"
    "    discard = spare0;\n"
    "    discard = spare0.a;\n"
    "    spare0  = sum();\n"
    "    scale_by_two();\n"
    "  }\n"
    "}\n"
    "out.rgb = spare0 * col0.rgb;\n"
    "out.a   = col0.a;\n"
  );
  nvparse_print_errors( stdout );
  glEndList();

  glNewList( combinerId + 1, GL_COMPILE );
  nvparse(
    "!!RC1.0\n"
    "{\n"
    "  rgb {\n"
    "    spare0 = expand(tex0) . expand(tex1);\n"
    "  }\n"
    "  alpha {\n"
    "    spare0 = unsigned_invert(tex1) * const0;\n"
    "  }\n"
    "}\n"
    "{\n"
    "  rgb {\n"
    "    discard = spare0;\n"
    "    discard = spare0.a;\n"
    "    spare0  = sum();\n"
    "    scale_by_two();\n"
    "  }\n"
    "}\n"
    "out.rgb = spare0;\n"
    "out.a   = spare0.a;\n"
  );
  nvparse_print_errors( stdout );
  glEndList();

  glCallList( combinerId );

  glCombinerParameterNV( GL_CONSTANT_COLOR0_NV, vec3f(), alphaScale );
}


// This function creates two vertex programs. They both calculate
// color and texture coordinates, but the first one does not calculate
// specular component.

void createProgram() {
  glGenProgramsNV( 2, programId );

  glBindProgramNV( GL_VERTEX_PROGRAM_NV, programId[0] );
  nvparse(
    "!!VP1.0\n"
    "# transform vertex to clip space and output it\n"
    "# c[0]-c[3] MODELVIEW_PROJECTION\n"
    "DP4 o[HPOS].x, v[OPOS], c[0];\n"
    "DP4 o[HPOS].y, v[OPOS], c[1];\n"
    "DP4 o[HPOS].z, v[OPOS], c[2];\n"
    "DP4 o[HPOS].w, v[OPOS], c[3];\n\n"

    "# transform vertex to eye space\n"
    "# c[4]-c[7] MODELVIEW\n"
    "DP4 R4.x, v[OPOS], c[4];\n"
    "DP4 R4.y, v[OPOS], c[5];\n"
    "DP4 R4.z, v[OPOS], c[6];\n"
    "DP4 R4.w, v[OPOS], c[7];\n\n"

    "# output GL_EYE_LINEAR texture coordinates\n"
    "MOV o[TEX0], -R4;\n\n"

    "# transform normal\n"
    "# c[8]-c[11] INVERSE_TRANSPOSE_MODELVIRW\n"
    "DP3 R1.x, v[NRML], c[8];\n"
    "DP3 R1.y, v[NRML], c[9];\n"
    "DP3 R1.z, v[NRML], c[10];\n\n"

    "# normalize transformed normal\n"
    "DP3 R1.w, R1, R1;\n"
    "RSQ R1.w, R1.w;\n"
    "MUL R1, R1, R1.w;\n"
    "MOV o[TEX1], R1;\n\n"

    "# transform and normalize light direction\n"
    "MOV R3, c[12];\n"
    "DP4 R2.x, R3, c[4];\n"
    "DP4 R2.y, R3, c[5];\n"
    "DP4 R2.z, R3, c[6];\n"
    "DP4 R2.w, R3, c[7];\n"
    "DP3 R2.w, R2, R2;\n"
    "RSQ R2.w, R2.w;\n"
    "MUL R2, R2, R2.w;\n\n"

    "# normalized light direction (dot) normal\n"
    "DP3 R1.w, R1, R2;\n\n"

    "# 1 + normal (dot) light direction\n"
    "ADD R1.x, R1.w, c[15].x;\n\n"

    "# 0.5 * (1 + normal (dot) light)\n"
    "MUL R1.x, R1.x, c[15].y;\n\n"

    "# cool shade\n"
    "MUL R2, R1.xxxx, c[13];\n\n"

    "# 1.0 - 0.5 * (1 + normal (dot) light)\n"
    "ADD R1.y, -R1.x, c[15].x;\n\n"

    "# cool + warm shades\n"
    "MAD o[COL0], R1.yyyy, c[14], R2;\n\n"

    "END"
  );
  nvparse_print_errors( stdout );

  glBindProgramNV( GL_VERTEX_PROGRAM_NV, programId[1] );
  nvparse(
    "!!VP1.0\n"
    "# transform vertex to clip space and output it\n"
    "# c[0]-c[3] MODELVIEW_PROJECTION\n"
    "DP4 o[HPOS].x, v[OPOS], c[0];\n"
    "DP4 o[HPOS].y, v[OPOS], c[1];\n"
    "DP4 o[HPOS].z, v[OPOS], c[2];\n"
    "DP4 o[HPOS].w, v[OPOS], c[3];\n\n"

    "# transform vertex to eye space\n"
    "# c[4]-c[7] MODELVIEW\n"
    "DP4 R4.x, v[OPOS], c[4];\n"
    "DP4 R4.y, v[OPOS], c[5];\n"
    "DP4 R4.z, v[OPOS], c[6];\n"
    "DP4 R4.w, v[OPOS], c[7];\n\n"

    "# output GL_EYE_LINEAR texture coordinates\n"
    "MOV o[TEX0], -R4;\n\n"

    "# normalize eye vertex\n"
    "DP3 R4.w, R4, R4;\n"
    "RSQ R4.w, R4.w;\n"
    "MUL R4, R4, R4.w;\n\n"

    "# transform normal\n"
    "# c[8]-c[11] INVERSE_TRANSPOSE_MODELVIRW\n"
    "DP3 R1.x, v[NRML], c[8];\n"
    "DP3 R1.y, v[NRML], c[9];\n"
    "DP3 R1.z, v[NRML], c[10];\n\n"

    "# normalize transformed normal\n"
    "DP3 R1.w, R1, R1;\n"
    "RSQ R1.w, R1.w;\n"
    "MUL R1, R1, R1.w;\n"
    "MOV o[TEX1], R1;\n\n"

    "# transform and normalize light direction\n"
    "MOV R3, c[12];\n"
    "DP4 R2.x, R3, c[4];\n"
    "DP4 R2.y, R3, c[5];\n"
    "DP4 R2.z, R3, c[6];\n"
    "DP4 R2.w, R3, c[7];\n"
    "DP3 R2.w, R2, R2;\n"
    "RSQ R2.w, R2.w;\n"
    "MUL R2, R2, R2.w;\n\n"

    "# half vector - add light to eye vertex, then normalize\n"
    "ADD R4, -R4, c[15].wwxw;\n"
    "DP3 R4.w, R4, R4;\n"
    "RSQ R4.w, R4.w;\n"
    "MUL R4, R4, R4.w;\n\n"

    "# half vector (dot) normal\n"
    "DP3 R4.z, R4, R1;\n"
    "SGE R5.z, R4.z, c[15].w;\n"
    "MUL R4.z, R4.z, R5.z;\n\n"

    "# specular component\n"
    "LOG R4, R4.z;\n"
    "MUL R4.z, R4.z, c[15].z;\n"
    "EXP R4, R4.z;\n"
    "MUL R4, R4.z, c[16];\n\n"

    "# normalized light direction (dot) normal\n"
    "DP3 R1.w, R1, R2;\n\n"

    "# 1 + normal (dot) light direction\n"
    "ADD R1.x, R1.w, c[15].x;\n\n"

    "# 0.5 * (1 + normal (dot) light)\n"
    "MUL R1.x, R1.x, c[15].y;\n\n"

    "# cool shade\n"
    "MUL R2, R1.xxxx, c[13];\n\n"

    "# 1.0 - 0.5 * (1 + normal (dot) light)\n"
    "ADD R1.y, -R1.x, c[15].x;\n\n"

    "# cool + warm shades\n"
    "MAD R2, R1.yyyy, c[14], R2;\n\n"

    "# output color - cool + warm + specular\n"
    "ADD o[COL0], R2, R4;\n\n"

    "END"
  );
  nvparse_print_errors( stdout );

  glTrackMatrixNV( GL_VERTEX_PROGRAM_NV, 0,
	               GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV );

  glTrackMatrixNV( GL_VERTEX_PROGRAM_NV, 4,
	               GL_MODELVIEW, GL_IDENTITY_NV );

  glTrackMatrixNV( GL_VERTEX_PROGRAM_NV, 8,
	               GL_MODELVIEW, GL_INVERSE_TRANSPOSE_NV );

  glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, LIGHT_DIRECTION_PARAM,
	                      0.0, -0.5, -1.0, 0.0 );
  glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, COOL_COLOR_PARAM,
	                      r * alpha, g * alpha, b * alpha + cool, 0.0 );
  glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, WARM_COLOR_PARAM,
	                      r * beta + warm, g * beta + warm, b * beta, 0.0 );

  // some constants
  glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, 15,
                          1.0, 0.5, 40.0, 0.0 );

  glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, SPECULAR_COLOR_PARAM,
                          0.8, 0.8, 0.8, 0.0 );

  glEnable( GL_VERTEX_PROGRAM_NV );
}


// Initialization of extensions and the rest of OpenGL stuff

void initOpenGL() {
  if( !glh_init_extensions( " GL_NV_vertex_program"
                            " GL_ARB_multitexture"
                            " GL_NV_register_combiners" ) ) {
    cout << "Necessary extensions were not supported: " << endl
         << glh_get_unsupported_extensions() << endl << endl;
    exit( -1 );
  };

  glClearColor( 1.0, 1.0, 1.0, 0.0 );

  createNormalizationMap();
  createWarpZMap();

  createCombiners();

  createProgram();

  glEnable( GL_DEPTH_TEST );
}


// This function handles GUI objects. Action performed depends
// of object's unique id.

void guiAction( guiObject *o ) {
  bool repaint;

  switch( o->getId() ) {
    case COOL_UP_ID: {
      cool += 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, COOL_COLOR_PARAM,
                              r * alpha, g * alpha, cool + b * alpha, 0.0 );
      cool2WarmGradient->setLeftColor( 0.0, 0.0, cool );
      repaint = true;
    };
    break;

    case COOL_DOWN_ID: {
      cool -= 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, COOL_COLOR_PARAM,
                              r * alpha, g * alpha, cool + b * alpha, 0.0 );
      cool2WarmGradient->setLeftColor( 0.0, 0.0, cool );
      repaint = true;
    };
    break;

    case WARM_UP_ID: {
      warm += 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, WARM_COLOR_PARAM,
                              warm + r * beta, warm + g * beta, b * beta, 0.0 );
      cool2WarmGradient->setRightColor( warm, warm, 0.0 );
      repaint = true;
    };
    break;

    case WARM_DOWN_ID: {
      warm -= 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, WARM_COLOR_PARAM,
                              warm + r * beta, warm + g * beta, b * beta, 0.0 );
      cool2WarmGradient->setRightColor( warm, warm, 0.0 );
      repaint = true;
    };
    break;

    case DARK_UP_ID: {
      alpha += 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, COOL_COLOR_PARAM,
                              r * alpha, g * alpha, cool + b * alpha, 0.0 );
      dark2LightGradient->setLeftColor( r * alpha, g * alpha, b * alpha );
      repaint = true;
    };
    break;

    case DARK_DOWN_ID: {
      alpha -= 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, COOL_COLOR_PARAM,
                              r * alpha, g * alpha, cool + b * alpha, 0.0 );
      dark2LightGradient->setLeftColor( r * alpha, g * alpha, b * alpha );
      repaint = true;
    };
    break;

    case LIGHT_UP_ID: {
      beta += 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, WARM_COLOR_PARAM,
                              warm + r * beta, warm + g * beta, b * beta, 0.0 );
      dark2LightGradient->setRightColor( r * beta, g * beta, b * beta );
      repaint = true;
    };
    break;

    case LIGHT_DOWN_ID: {
      beta -= 0.01;
      glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, WARM_COLOR_PARAM,
                              warm + r * beta, warm + g * beta, b * beta, 0.0 );
      dark2LightGradient->setRightColor( r * beta, g * beta, b * beta );
      repaint = true;
    };
    break;

    case SPECULAR_ID: {
      if( specularCheck->isChecked() ) {
        glBindProgramNV( GL_VERTEX_PROGRAM_NV, programId[1] );
      }
      else {
        glBindProgramNV( GL_VERTEX_PROGRAM_NV, programId[0] );
      };
      repaint = true;
    };
    break;

    case COLOR_ID: {
      if( colorCheck->isChecked() ) {
        glCallList( combinerId );
      }
      else {
        glCallList( combinerId + 1 );
      };
      repaint = true;
    };
    break;

    case EDGES_ID: {
      edges = edgesCheck->isChecked();
      repaint = true;
    };
    break;

    case LOD_UP_ID: {
      lodBias += 0.01;
      char buffer[100];
      sprintf( buffer, "LOD bias: %1.2f", lodBias );
      lodLabel->setLabel( buffer );
      glActiveTextureARB( GL_TEXTURE1_ARB );
      glTexEnvf( GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, lodBias ); 
      repaint = true;
    };
    break;

    case LOD_DOWN_ID: {
      lodBias -= 0.01;
      char buffer[100];
      sprintf( buffer, "LOD bias: %1.2f", lodBias );
      lodLabel->setLabel( buffer );
      glActiveTextureARB( GL_TEXTURE1_ARB );
      glTexEnvf( GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, lodBias ); 
      repaint = true;
    };
    break;

    case SCALE_UP_ID: {
      alphaScale += 0.01;
      if( alphaScale < 0.0 ) {
        alphaScale = 0.0;
      };
      char buffer[100];
      sprintf( buffer, "alpha scale: %1.2f", alphaScale );
      scaleLabel->setLabel( buffer );
      glCombinerParameterNV( GL_CONSTANT_COLOR0_NV, vec3f(), alphaScale );
      repaint = 1;
    };
    break;

    case SCALE_DOWN_ID: {
      alphaScale -= 0.01;
      if( alphaScale > 1.0 ) {
        alphaScale = 1.0;
      };
      char buffer[100];
      sprintf( buffer, "alpha scale: %1.2f", alphaScale );
      scaleLabel->setLabel( buffer );
      glCombinerParameterNV( GL_CONSTANT_COLOR0_NV, vec3f(), alphaScale );
      repaint = 1;
    };
    break;
  };

  if( repaint ) {
    glutPostRedisplay();
  };
}


// GUI setup

void createGUI() {
  gui.setGUISize( 40, 50 );

  guiObject *o;
  
  gui.addGUIObject( o = new upButton( COOL_UP_ID, 2.5, 5.25, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( o = new downButton( COOL_DOWN_ID, 2.5, 4.0, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );

  gui.addGUIObject( cool2WarmGradient = new colorGradient( 6.0, 4.0, 25.0, 2.5 ) );
  cool2WarmGradient->setLeftColor( 0.0, 0.0, cool );
  cool2WarmGradient->setRightColor( warm, warm, 0 );

  gui.addGUIObject( o = new upButton( WARM_UP_ID, 31.5, 5.25, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( o = new downButton( WARM_DOWN_ID, 31.5, 4.0, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );

  gui.addGUIObject( o = new upButton( DARK_UP_ID, 2.5, 15.25, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( o = new downButton( DARK_DOWN_ID, 2.5, 14.0, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );

  gui.addGUIObject( dark2LightGradient = new colorGradient( 6.0, 14.0, 25.0, 2.5 ) );
  dark2LightGradient->setLeftColor( r * alpha, g * alpha, b * alpha );
  dark2LightGradient->setRightColor( r * beta, g * beta, b * beta );

  gui.addGUIObject( o = new upButton( LIGHT_UP_ID, 31.5, 15.25, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( o = new downButton( LIGHT_DOWN_ID, 31.5, 14.0, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );

  gui.addGUIObject( finalGradient = new colorGradient( 6.0, 9.0, 25.0, 2.5 ) );
  finalGradient->setLeftColor( r * alpha, g * alpha, b * alpha + cool );
  finalGradient->setRightColor( r * beta + warm, g * beta + warm, b * beta );

  gui.addGUIObject( specularCheck = new checkBox( SPECULAR_ID, 2.5, 37.0, 2.5, 2.5, "Specular", true ) );
  specularCheck->setActionFunc( &guiAction );

  gui.addGUIObject( edgesCheck = new checkBox( EDGES_ID, 2.5, 41.0, 2.5, 2.5, "Edges", true ) );
  edgesCheck->setActionFunc( &guiAction );

  gui.addGUIObject( colorCheck = new checkBox( COLOR_ID, 2.5, 45.0, 2.5, 2.5, "Color", true ) );
  colorCheck->setActionFunc( &guiAction );

  gui.addGUIObject( o = new upButton( LOD_UP_ID, 2.5, 29.25, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( o = new downButton( LOD_DOWN_ID, 2.5, 28.0, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( lodLabel = new label( 6.5, 28.0, 2.5, 2.5, "LOD bias: 3.00" ) );

  gui.addGUIObject( o = new upButton( SCALE_UP_ID, 2.5, 25.25, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( o = new downButton( SCALE_DOWN_ID, 2.5, 24.0, 3.0, 1.25 ) );
  o->setActionFunc( &guiAction );
  gui.addGUIObject( scaleLabel = new label( 6.5, 24.0, 2.5, 2.5, "alpha scale: 0.45" ) );

  glut_add_interactor( &gui);
}


// Draws teapot. We can't use GLUT built-in function because
// evaluators don't work with vertex programs.

void drawModel() {
  glBegin( GL_TRIANGLES );
  for( int f = 0 ; f < ntris ; f++ ) {
    int v0 = indicies[f * 3 + 0];
    int v1 = indicies[f * 3 + 1];
    int v2 = indicies[f * 3 + 2];
    glNormal3fv( normaldata[v0] );
    glVertex3fv( vertexdata[v0] );
    glNormal3fv( normaldata[v1] );
    glVertex3fv( vertexdata[v1] );
    glNormal3fv( normaldata[v2] );
    glVertex3fv( vertexdata[v2] );
  };
  glEnd();
}


// GLUT display callback

void display() {
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  if( edges ) {
    glEnable( GL_REGISTER_COMBINERS_NV );
    glActiveTextureARB( GL_TEXTURE0_ARB );
    glEnable( GL_TEXTURE_CUBE_MAP_ARB );
    glActiveTextureARB( GL_TEXTURE1_ARB );
    glEnable( GL_TEXTURE_CUBE_MAP_ARB );
  }
  else {
    glDisable( GL_REGISTER_COMBINERS_NV );
    glActiveTextureARB( GL_TEXTURE0_ARB );
    glDisable( GL_TEXTURE_CUBE_MAP_ARB );
    glActiveTextureARB( GL_TEXTURE1_ARB );
    glDisable( GL_TEXTURE_CUBE_MAP_ARB );
  };
  glEnable( GL_VERTEX_PROGRAM_NV );
  glEnable( GL_DEPTH_TEST );
  glDisable( GL_BLEND );

  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glTranslatef( -modelcenter[0], -modelcenter[1], -modelcenter[2] - 10.0 );
  mouse.apply_transform();
  drawModel();
  glPopMatrix();
}


// and the main function

int main( int argc, char **argv ) {
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB );
  glutInitWindowSize( 600, 600 );
  glutCreateWindow( "Non-Photorealistic Rendering" );

  initOpenGL();

  glut_helpers_initialize();

  cb.display_function = display;
  glut_add_interactor( &cb );

  createGUI();

  mouse.configure_buttons( 1 );
  mouse.dolly.dolly[2] = -3.0;
  glut_add_interactor( &mouse );

  reshaper.fovy = 30.0;
  reshaper.zFar = 30.0;
  glut_add_interactor( &reshaper );

  glutMainLoop();

  return 0;
}


// This is required when you compile program as a GLUT
// application rather than Effect

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow) {
  int argc = 1;
  char *argv[] = { "NPR" };
  return main( argc, argv );
}
