#ifndef __GUI_MOUSE_INTERACTOR_H
#define __GUI_MOUSE_INTERACTOR_H




using namespace glh;




void drawString( float x, float y, void *face, float size, char *string ) {
  glEnable( GL_LINE_SMOOTH );

  int line = 0;
  glPushMatrix();
  glTranslatef( x, y, 0.0 );
  glScalef( size, size, 1);
  while( *string != 0 ) {
    if( *string != '\n' ) {
      glutStrokeCharacter( face, *string );
    }
    else {
      line++;
      glPopMatrix();
      glPushMatrix();
      glTranslatef( x, y - 0.04 * line, 0.0 );
      glScalef( size, size, 1);  
    };
    string++;
  };
  glPopMatrix();

  glDisable( GL_LINE_SMOOTH );
}




class guiObject {
  void (*actionFunc)(guiObject *);

  guiObject();
  guiObject( const guiObject & );
  guiObject &operator=( const guiObject & );

protected:
  char *label;
  float labelLeft, labelBottom;
  void *labelFace;
  float labelSize;

  int id;
  float left, bottom;
  float width, height;

  virtual void drawField() {
    glBegin( GL_QUADS );
    glColor4f( 0.8, 0.8, 0.8, 0.8 );
    glVertex2f( left, bottom );
    glVertex2f( left + width, bottom );
    glVertex2f( left + width, bottom + height );
    glVertex2f( left, bottom + height);
    glEnd();
  }

  virtual void drawFrame() {
    glBegin( GL_LINE_LOOP );
    glColor4f( 0.5, 0.5, 0.5, 0.8 );
    glVertex2f( left, bottom );
    glVertex2f( left + width, bottom );
    glVertex2f( left + width, bottom + height );
    glVertex2f( left, bottom + height);
    glEnd();
  }

  virtual void drawLabel() {
    if( label != NULL ) {
      glColor4f( 0.3, 0.3, 0.3, 0.8 );
      drawString( left + labelLeft, bottom + labelBottom, labelFace, labelSize, label );
    };
  }

public:
  guiObject( float l, float b, float w, float h, char *lab = NULL ):
    label( NULL ),
    labelLeft( 0.0 ),
    labelBottom( 0.0 ),
    labelFace( GLUT_STROKE_ROMAN ),
    labelSize( h / 10000.0 ),
    actionFunc( NULL ),
    id( -1 ),
    left( l / 100.0 ),
    bottom( b / 100.0 ),
    width( w / 100.0 ),
    height( h / 100.0 ) {
      setLabel( lab );
    }

  guiObject( int i, float l, float b, float w, float h, char *lab = NULL ):
    label( NULL ),
    labelLeft( 0.0 ),
    labelBottom( 0.0 ),
    labelFace( GLUT_STROKE_ROMAN ),
    labelSize( h / 10000.0 ),
    actionFunc( NULL ),
    id( i ),
    left( l / 100.0 ),
    bottom( b / 100.0 ),
    width( w / 100.0 ),
    height( h / 100.0 ) {
      setLabel( lab );
    }

  inline void setLabel( char *l ) {
    if( l != NULL ) {
      delete[] label;
      label = new char[strlen( l ) + 1];
      strcpy( label, l );
    };
  }

  inline void setLabelFace( void *f ) {
    labelFace = f;
  }

  inline void setActionFunc( void (*f)(guiObject *) ) {
    actionFunc = f;
  }

  inline int getId() {
    return id;
  }

  virtual bool isOnOff() {
    return false;
  }

  virtual bool hit( float x, float y ) {
    if( (x >= left) && (x < (left + width)) && (y >= bottom) && (y < (bottom + height)) ) {
      action();
      return true;
    };
    return false;
  }

  virtual void action() {
    if( actionFunc != NULL ) {
      actionFunc( this );
    };
  }

  virtual void display() {
  }
};


class label : public guiObject {
public:
  label( float l, float b, float w, float h, char *lab,
         void *f = GLUT_STROKE_ROMAN ):
    guiObject( l, b, w, h, lab ) {
      labelFace = f;
    }

  label( int i, float l, float b, float w, float h, char *lab,
         void *f = GLUT_STROKE_ROMAN ):
    guiObject( i, l, b, w, h, lab ) {
      labelFace = f;
    }

  virtual void display() {
    drawLabel();
  }
};


class upButton : public guiObject {
public:
  upButton( float l, float b, float w, float h ):
    guiObject( l, b, w, h ) {
    }

  upButton( int i, float l, float b, float w, float h ):
    guiObject( i, l, b, w, h ) {
    }

  virtual void display() {
    drawField();
    drawFrame();

    glBegin( GL_TRIANGLES );
    glColor4f( 0.3, 0.3, 0.3, 0.8 );
    glVertex2f( left + 0.2 * width, bottom + 0.2 * height );
    glVertex2f( left + 0.8 * width, bottom + 0.2 * height );
    glVertex2f( left + 0.5 * width, bottom + 0.8 * height );
    glEnd();
  };
};


class downButton : public guiObject {
public:
  downButton( float l, float b, float w, float h ):
    guiObject( l, b, w, h ) {
    }

  downButton( int i, float l, float b, float w, float h ):
    guiObject( i, l, b, w, h ) {
    }

  virtual void display() {
    drawField();
    drawFrame();

    glBegin( GL_TRIANGLES );
    glColor4f( 0.3, 0.3, 0.3, 0.8 );
    glVertex2f( left + 0.2 * width, bottom + 0.8 * height );
    glVertex2f( left + 0.8 * width, bottom + 0.8 * height );
    glVertex2f( left + 0.5 * width, bottom + 0.2 * height );
    glEnd();
  };
};


class colorBox : public guiObject {
  float color[4];

public:
  colorBox( float l, float b, float w, float h,
            float r = 0.0, float g = 0.0, float bl = 0.0 ):
    guiObject( l, b, w, h ) {
      color[0] = r;
      color[1] = g;
      color[2] = bl;
      color[3] = 0.8;
    }

  colorBox( int i, float l, float b, float w, float h,
            float r = 0.0, float g = 0.0, float bl = 0.0 ):
    guiObject( i, l, b, w, h )
    {
      color[0] = r;
      color[1] = g;
      color[2] = bl;
      color[3] = 0.8;
    }

  inline void setColor( float r, float g, float b ) {
    color[0] = r;
    color[1] = g;
    color[2] = b;
  }

  virtual void display() {
    glBegin( GL_QUADS );
    glColor4fv( color );
    glVertex2f( left, bottom + height);
    glVertex2f( left, bottom );
    glVertex2f( left + width, bottom );
    glVertex2f( left + width, bottom + height );
    glEnd();

    drawFrame();
  }
};


class colorGradient : public guiObject {
  float color0[4], color1[4];

public:
  colorGradient( float l, float b, float w, float h,
                 float r0 = 0.0, float g0 = 0.0, float b0 = 0.0,
                 float r1 = 1.0, float g1 = 1.0, float b1 = 1.0 ):
    guiObject( l, b, w, h )
    {
      color0[0] = r0;
      color0[1] = g0;
      color0[2] = b0;
      color0[3] = 0.8;
      color1[0] = r1;
      color1[1] = g1;
      color1[2] = b1;
      color1[3] = 0.8;
    }

  colorGradient( int i, float l, float b, float w, float h,
                 float r0 = 0.0, float g0 = 0.0, float b0 = 0.0,
                 float r1 = 1.0, float g1 = 1.0, float b1 = 1.0 ):
    guiObject( i, l, b, w, h )
    {
      color0[0] = r0;
      color0[1] = g0;
      color0[2] = b0;
      color0[3] = 0.8;
      color1[0] = r1;
      color1[1] = g1;
      color1[2] = b1;
      color1[3] = 0.8;
    }

  inline void setLeftColor( float r, float g, float b ) {
    color0[0] = r;
    color0[1] = g;
    color0[2] = b;
  }

  inline void setRightColor( float r, float g, float b ) {
    color1[0] = r;
    color1[1] = g;
    color1[2] = b;
  }

  virtual void display() {
    glBegin( GL_QUADS );
    glColor4fv( color0 );
    glVertex2f( left, bottom + height);
    glVertex2f( left, bottom );
    glColor4fv( color1 );
    glVertex2f( left + width, bottom );
    glVertex2f( left + width, bottom + height );
    glEnd();

    drawFrame();
  }
};


class checkBox : public guiObject {
  bool checked;

public:
  checkBox( float l, float b, float w, float h, char *lab = NULL,
            bool c = false ):
    guiObject( l, b, w, h, lab ),
    checked( c ) {
      labelLeft = 1.25 * width + 0.01;
      labelBottom = 0.0;
      labelSize = h / 10000.0;
    }

  checkBox( int i, float l, float b, float w, float h, char *lab = NULL,
            bool c = false ):
    guiObject( i, l, b, w, h, lab ),
    checked( c ) {
      labelLeft = 1.25 * width + 0.01;
      labelBottom = 0.0;
      labelSize = h / 10000.0;
    }

  inline void setChecked( bool c ) {
    checked = c;
  }

  inline bool isChecked() {
    return checked;
  }

  virtual bool isOnOff() {
    return true;
  }

  virtual bool hit( float x, float y ) {
    bool result = false;;
    if( (x >= left) && (x < (left + width)) && (y >= bottom) && (y < (bottom + height)) ) {
      result = true;
    };
   
    if( result ) {
      checked = !checked;
      action();
    };
    return result;
  }

  virtual void display() {
    drawField();
    drawFrame();

    if( checked ) {
      glBegin( GL_QUADS );
      glColor4f( 0.5, 0.5, 0.5, 0.8 );
      glVertex2f( left + 0.2 * width, bottom + 0.2 * height );
      glVertex2f( left + 0.8 * width, bottom + 0.2 * height );
      glVertex2f( left + 0.8 * width, bottom + 0.8 * height );
      glVertex2f( left + 0.2 * width, bottom + 0.8 * height );
      glEnd();
    };

    drawLabel();
  }
};


class gui_interactor : public glut_interactor {
  int windowWidth, windowHeight;
  float aspect;

  float guiWidth, guiHeight;

  std::list<guiObject *> guiObjects;
  guiObject *hit;

  gui_interactor( const gui_interactor & );
  gui_interactor &operator=( const gui_interactor & );

public:
  gui_interactor():
    glut_interactor(),
    windowWidth( 0 ),
    windowHeight( 0 ),
    aspect( 1.0 ),
    guiWidth( 1.0 ),
    guiHeight( 1.0 ),
    hit( NULL )
    {
    }

  virtual void mouse( int mb, int state, int mx, int my ) {
    if( mb == GLUT_LEFT_BUTTON ) {
      float x = (float)(mx) / (float)(windowHeight);
      float y = (float)(windowHeight - 1 - my) / (float)(windowHeight);
      if( state == GLUT_DOWN ) {
        if( (x <= guiWidth) && (y <= guiHeight) ) {
          glut_event_processed();
        };
      };
      if( state == GLUT_UP ) {
        if( hit != NULL ) {
          hit = NULL;
          glut_idle( false );
          return;
        };
      };

      for( std::list<guiObject *>::iterator it = guiObjects.begin() ; it != guiObjects.end() ; it++ ) {
        if( (*it)->hit( x, y ) ) {
          hit = (*it);
          if( !(*it)->isOnOff() ) {
            glut_idle( true );
          };
          break;
        };
      };
    };
  }

  virtual void keyboard( unsigned char k, int x, int y ) {
    if( (k == 27) || (k == 'q') ) {
      exit( 0 );
    };

    for( std::list<guiObject *>::iterator it = guiObjects.begin() ; it != guiObjects.end() ; it++ ) {
      if( (*it)->getId() == k ) {
	  	(*it)->action();
        break;
      };
    };
  }

  virtual void reshape( int w, int h ) {
    windowWidth = w;
    windowHeight = h;
    aspect = (float)(w) / (float)(h);
  }

  virtual void display() {
    glMatrixMode( GL_PROJECTION );
    glPushMatrix();
    glLoadIdentity();
    glOrtho( 0.0, aspect, 0.0, 1.0, -1.0, 1.0 );
    glMatrixMode (GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();

    glDisable( GL_REGISTER_COMBINERS_NV );
    glDisable( GL_VERTEX_PROGRAM_NV );
    glDisable( GL_DEPTH_TEST );
    glActiveTextureARB( GL_TEXTURE0_ARB );
    glDisable( GL_TEXTURE_CUBE_MAP_ARB );
    glActiveTextureARB( GL_TEXTURE1_ARB );
    glDisable( GL_TEXTURE_CUBE_MAP_ARB );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glEnable( GL_BLEND );

    glBegin( GL_QUADS );
    glColor4f( 0.9, 0.9, 0.9, 0.8 );
    glVertex2f( 0.0, 0.0 );
    glVertex2f( 0.0, guiHeight );
    glColor4f( 0.9, 0.9, 0.9, 0.8 );
    glVertex2f( guiWidth, guiHeight );
    glVertex2f( guiWidth, 0.0 );
    glEnd();

    for( std::list<guiObject *>::iterator it = guiObjects.begin() ; it != guiObjects.end() ; it++ ) {
      (*it)->display();
    };

    glPopMatrix();
    glMatrixMode( GL_PROJECTION );
    glPopMatrix();

    glutSwapBuffers();
  }

  virtual void idle() {
    if( hit != NULL ) {
      hit->action();
    };
  }

  inline void setGUISize( float w, float h ) {
    guiWidth = w / 100.0;
    guiHeight = h / 100.0;
  }

  inline void addGUIObject( guiObject *o ) {
    guiObjects.push_back( o );
  }

};




#endif