//
// shader_BumpRefract.cpp
// Copyright (C) 2001 Tomohide Kano. All rights reserved.
//

#include "eb_effect.h"
#include "nvdevice.h"
#include "MouseUI.h"
#include "shader_BumpRefract.h"
#include "shader_BumpUtil.h"
#include "constants.h"
#include <stdio.h>
#include <stdlib.h>


DECLARE_EFFECT_MAIN()

extern "C" __declspec(dllexport)
unsigned int GetNumEffects()
{
	return 1;
}

extern "C" __declspec(dllexport)
EBEffect * CreateEffect(unsigned int EffectNum)
{
	return new CShaderBumpRefract();
}

// -----------------------------------------------------------------------------
// CShaderBumpRefract
// -----------------------------------------------------------------------------
CShaderBumpRefract::CShaderBumpRefract()
{
	m_strEffectName         = "Bump Refraction";
	m_strEffectLocation     = "3rd Party\\Bump Mapping";
	m_strEffectPixelShader  = GetFilePath("reflect.nvp");
	m_strEffectVertexShader = GetFilePath("reflect.nvv");
	m_strEffectVersion = "1.0";

	m_pMouseObj			= NULL;
	m_pMouseEnv			= NULL;

	m_pBoxVertexBuffer	= NULL;
	m_pBoxIndexBuffer	= NULL;
	m_pQuadVertexBuffer	= NULL;
	m_pQuadIndexBuffer	= NULL;
	m_pPatchBuffer		= NULL;

	m_pBumpMaps.clear();
	m_ppCurrentBumpMap	= m_pBumpMaps.begin();
	m_pCubeMap			= NULL;

	m_dwEnvBoxVertexShader			= 0;
	m_dwReflectVertexShader			= 0;
	m_dwReflectVertexShader_Patch	= 0;
	m_dwReflectPixelShader			= 0;
	m_dwRefractVertexShader			= 0;
	m_dwRefractVertexShader_Patch	= 0;
	m_dwRefractPixelShader			= 0;

	m_bPause			= false;
	m_bWireframe		= false;
	m_bUsePatch			= true;
	m_eRenderStyle		= RENDER_REFLECT_REFRACT;
	m_eMouseState		= MOUSE_OBJ;
	m_fRefractIndex		= 1.5f;
}

// -----------------------------------------------------------------------------
// ~CShaderBumpRefract
// -----------------------------------------------------------------------------
CShaderBumpRefract::~CShaderBumpRefract()
{
	Free();	
}

// -----------------------------------------------------------------------------
// ConfirmDevice
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::ConfirmDevice(D3DCAPS8 *pCaps, DWORD dwBehavior, D3DFORMAT Format)
{
	if (!(pCaps->TextureCaps & D3DPTEXTURECAPS_CUBEMAP))
	{
		m_strLastError = "Device does not support cubemaps!";
		return E_FAIL;
	}

	if (!(pCaps->MaxTextureBlendStages >= 4))
	{
		m_strLastError = "Not enough texture blend stages!";
		return E_FAIL;
	}

	if (D3DSHADER_VERSION_MAJOR(pCaps->PixelShaderVersion) < 1)
	{
		m_strLastError = "Device does not support pixel shaders!";
		return E_FAIL;
	}

	if (!(pCaps->DevCaps & D3DDEVCAPS_RTPATCHES))
	{
		m_strLastError = "Device does not support RTPATCHES!";
		return E_FAIL;
	}

	return S_OK;
}

// -----------------------------------------------------------------------------
// Initialize
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::Initialize(IDirect3DDevice8 *pDev)
{
	HRESULT hr;

	srand(GetTickCount());

	m_pD3DDev = pDev;
	m_pD3DDev->AddRef();

	// Mouse UI
	D3DVIEWPORT8 viewport;
	m_pD3DDev->GetViewport(&viewport);
	RECT rect;
	rect.left = rect.top = 0;
	rect.bottom = viewport.Height;
	rect.right = viewport.Width;
	m_pMouseObj = new MouseUI(rect);
	m_pMouseEnv = new MouseUI(rect);

	m_bSupportsPatches	= true;
	D3DDEVICE_CREATION_PARAMETERS CreationParams;
	m_pD3DDev->GetCreationParameters(&CreationParams);
	if (CreationParams.BehaviorFlags & D3DCREATE_SOFTWARE_VERTEXPROCESSING)
	{
		m_bSupportsPatches = false;
		m_bUsePatch        = false;
	}

    hr = GenerateEnvBox();
	if (FAILED(hr))
		return hr;

    hr = GenerateQuad(D3DXVECTOR3(0.0f, 0.0f, 0.0f), 0.4f);
	if (FAILED(hr))
		return hr;

    hr = GenerateBezierPatch();
	if (FAILED(hr))
		return hr;

	hr = CreateShaders();
	if (FAILED(hr))
		return hr;

	hr = CreateTextures();
	if (FAILED(hr))
		return hr;

	// Shader constants
	m_pD3DDev->SetVertexShaderConstant(CV_ZERO, D3DXVECTOR4(0.0f, 0.0f, 0.0f, 0.0f), 1);
	m_pD3DDev->SetVertexShaderConstant(CV_ONE,  D3DXVECTOR4(1.0f, 1.0f, 1.0f, 1.0f), 1);

	// Projection matrix
	D3DXMatrixPerspectiveFovLH(&m_matProj, D3DXToRadian(60.0f), 1.0f, 0.01f, 10.0f);

	// Backface culling
	// Note: Refraction is not correct on back faces
	m_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

	// Reset to defaults
	Keyboard(VK_HOME, 0, true);

	return S_OK;

}

// -----------------------------------------------------------------------------
// CreateShaders
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::CreateShaders()
{
	HRESULT hr;

	// Create the vertex shader for the environment box
	{
		DWORD Declaration[] =
		{
			D3DVSD_STREAM(0),
			D3DVSD_REG(0, D3DVSDT_FLOAT3), // Position
			D3DVSD_END()
		};

		hr = LoadAndCreateShader(GetFilePath("envbox.vso"),
			&Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwEnvBoxVertexShader);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load shader (envbox.vso)";
			return hr;
		}
	}

	// Create the vertex shaders for the quad
	{
		DWORD Declaration[] =
		{
			D3DVSD_STREAM(0),
			D3DVSD_REG(0, D3DVSDT_FLOAT3), // Position
			D3DVSD_REG(1, D3DVSDT_FLOAT2), // Texture
			D3DVSD_REG(2, D3DVSDT_FLOAT3), // dp/ds
			D3DVSD_REG(3, D3DVSDT_FLOAT3), // dp/dt
			D3DVSD_END()
		};

		hr = LoadAndCreateShader(GetFilePath("reflect.vso"),
			&Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwReflectVertexShader);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load shader (reflect.vso)";
			return hr;
		}

		hr = LoadAndCreateShader(GetFilePath("refract.vso"),
			&Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwRefractVertexShader);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load shader (refract.vso)";
			return hr;
		}
	}

	// Create the vertex shaders for the patch
	if (m_bSupportsPatches)
	{
		DWORD PatchDeclaration[] =
		{
			D3DVSD_STREAM(0),
			D3DVSD_REG(0, D3DVSDT_FLOAT3), // Position
			D3DVSD_REG(5, D3DVSDT_FLOAT2), // Texture (not used)
			D3DVSD_REG(2, D3DVSDT_FLOAT3), // dp/ds
			D3DVSD_REG(3, D3DVSDT_FLOAT3), // dp/dt
			D3DVSD_STREAM_TESS(),
			D3DVSD_TESSUV(1),
			D3DVSD_END()
		};

		hr = LoadAndCreateShader(GetFilePath("reflect.vso"),
			&PatchDeclaration[0], 0, SHADERTYPE_VERTEX, &m_dwReflectVertexShader_Patch);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load shader (reflect.vso)";
			return hr;
		}

		hr = LoadAndCreateShader(GetFilePath("refract.vso"),
			&PatchDeclaration[0], 0, SHADERTYPE_VERTEX, &m_dwRefractVertexShader_Patch);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load shader (refract.vso)";
			return hr;
		}
	}

	// Create the pixel shaders
	{
		hr = LoadAndCreateShader(GetFilePath("reflect.pso"),
			NULL, 0, SHADERTYPE_PIXEL, &m_dwReflectPixelShader);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load shader (reflect.pso)";
			return hr;
		}

		hr = LoadAndCreateShader(GetFilePath("refract.pso"),
			NULL, 0, SHADERTYPE_PIXEL, &m_dwRefractPixelShader);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load shader (refract.pso)";
			return hr;
		}
	}

	return S_OK;
}

// -----------------------------------------------------------------------------
// CreateTextures
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::CreateTextures()
{
	HRESULT hr;
	LPDIRECT3DTEXTURE8 pHeightMap, pTempTexture;

	bool bSupports8Bit  = false;
	bool bSupports16Bit = false;

	// Check available bumpmap formats
	{
		hr = D3DXCreateTexture(m_pD3DDev, 16, 16, 1, 0,
				D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &pHeightMap);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load texture!";
			return hr;
		}

		hr = CreateQ8W8V8U8FromHeight(m_pD3DDev, pHeightMap, &pTempTexture);
		bSupports8Bit = SUCCEEDED(hr);
		SAFE_RELEASE(pTempTexture);

		hr = CreateNVHSFromHeight(m_pD3DDev, pHeightMap, &pTempTexture);
		bSupports16Bit = SUCCEEDED(hr);
		SAFE_RELEASE(pTempTexture);

		if (!bSupports8Bit && !bSupports16Bit)
		{
			m_strLastError = "Could not find suitable bumpmap format!";
			return E_FAIL;
		}

		SAFE_RELEASE(pHeightMap);
	}

	FILE *fp = fopen(GetFilePath("bump_maps.txt").c_str(), "r");
	if (fp == NULL)
	{
		m_strLastError = "Could not open bump_maps.txt";
		return E_FAIL;
	}

	// Load bump normal maps
	while (true)
	{
		char bumpmap_name[80];
		int result = fscanf(fp, "%s", bumpmap_name);
		if (result == 0 || result == EOF)
			break;

		hr = D3DXCreateTextureFromFileEx(
			m_pD3DDev, GetFilePath(bumpmap_name).c_str(),
			D3DX_DEFAULT, D3DX_DEFAULT, 1, 0,
			D3DFMT_A8R8G8B8, D3DPOOL_MANAGED,
			D3DX_FILTER_LINEAR, D3DX_FILTER_LINEAR,
			0, NULL, NULL, &pHeightMap);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load texture (";
			m_strLastError += bumpmap_name;
			m_strLastError += ")";
			return hr;
		}

		if (bSupports16Bit)
			hr = CreateNVHSFromHeight(m_pD3DDev, pHeightMap, &pTempTexture);
		else
			hr = CreateQ8W8V8U8FromHeight(m_pD3DDev, pHeightMap, &pTempTexture);
		if (FAILED(hr))
		{
			m_strLastError = "Could not load texture (";
			m_strLastError += bumpmap_name;
			m_strLastError += ")";
			return hr;
		}

		m_pBumpMaps.push_back(pTempTexture);

		SAFE_RELEASE(pHeightMap);
	}

	fclose(fp);

	m_ppCurrentBumpMap = m_pBumpMaps.begin();

	// Load environment cube map
	hr = D3DXCreateCubeTextureFromFileExA(m_pD3DDev,
		GetFilePath("nvlobby_cube_mipmap.dds").c_str(), 
		D3DX_DEFAULT, 0, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
		D3DX_FILTER_LINEAR, D3DX_FILTER_LINEAR,
		0, NULL, NULL, &m_pCubeMap);
	if (FAILED(hr))
	{
		m_strLastError = "Could not load texture (nvlobby_cube_mipmap.dds)";
		return hr;
	}

	return S_OK;
}

// -----------------------------------------------------------------------------
// Free
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::Free()
{
	SAFE_DELETE(m_pMouseObj);
	SAFE_DELETE(m_pMouseEnv);

    SAFE_RELEASE(m_pBoxVertexBuffer);
    SAFE_RELEASE(m_pBoxIndexBuffer);
    SAFE_RELEASE(m_pQuadVertexBuffer);
    SAFE_RELEASE(m_pQuadIndexBuffer);
    SAFE_RELEASE(m_pPatchBuffer);

	while (!m_pBumpMaps.empty())
	{
		SAFE_RELEASE(m_pBumpMaps.back());
		m_pBumpMaps.pop_back();
	}
	SAFE_RELEASE(m_pCubeMap);

	if (m_pD3DDev)
	{
		if (m_dwEnvBoxVertexShader)
			m_pD3DDev->DeleteVertexShader(m_dwEnvBoxVertexShader);

		if (m_dwReflectVertexShader)
			m_pD3DDev->DeleteVertexShader(m_dwReflectVertexShader);
		if (m_dwReflectVertexShader_Patch)
			m_pD3DDev->DeleteVertexShader(m_dwReflectVertexShader_Patch);
		if (m_dwReflectPixelShader)
			m_pD3DDev->DeletePixelShader(m_dwReflectPixelShader);

		if (m_dwRefractVertexShader)
			m_pD3DDev->DeleteVertexShader(m_dwRefractVertexShader);
		if (m_dwRefractVertexShader_Patch)
			m_pD3DDev->DeleteVertexShader(m_dwRefractVertexShader_Patch);
		if (m_dwRefractPixelShader)
			m_pD3DDev->DeletePixelShader(m_dwRefractPixelShader);

		SAFE_RELEASE(m_pD3DDev);
	}

	return S_OK;
}

// -----------------------------------------------------------------------------
// Start
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::Start()
{
	return S_OK;
}

// -----------------------------------------------------------------------------
// GenerateEnvBox
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::GenerateEnvBox()
{
	HRESULT hr;

	SAFE_RELEASE(m_pBoxVertexBuffer);
	SAFE_RELEASE(m_pBoxIndexBuffer);

	m_dwBoxNumVertices = 8;
	m_dwBoxNumFaces    = 12;	

	hr = m_pD3DDev->CreateVertexBuffer(m_dwBoxNumVertices * sizeof(D3DXVECTOR3),
			0, D3DFVF_XYZ, D3DPOOL_MANAGED, &m_pBoxVertexBuffer);
	if (FAILED(hr))
	{
		m_strLastError = "Failed to generate environment box";
		return hr;
	}

	hr = m_pD3DDev->CreateIndexBuffer(3 * m_dwBoxNumFaces * sizeof(WORD),
			0, D3DFMT_INDEX16, D3DPOOL_MANAGED, &m_pBoxIndexBuffer);
	if (FAILED(hr))
	{
		m_strLastError = "Failed to generate environment box";
		return hr;
	}

	D3DXVECTOR3 *pVertices;
	WORD *pIndices;

	m_pBoxVertexBuffer->Lock(0, 0, (BYTE**)&pVertices, 0);
	m_pBoxIndexBuffer->Lock(0, 0, (BYTE**)&pIndices, 0);

	pVertices[0] = D3DXVECTOR3(-1.0f, -1.0f, -1.0f);
	pVertices[1] = D3DXVECTOR3( 1.0f, -1.0f, -1.0f);
	pVertices[2] = D3DXVECTOR3(-1.0f,  1.0f, -1.0f);
	pVertices[3] = D3DXVECTOR3( 1.0f,  1.0f, -1.0f);
	pVertices[4] = D3DXVECTOR3(-1.0f, -1.0f,  1.0f);
	pVertices[5] = D3DXVECTOR3( 1.0f, -1.0f,  1.0f);
	pVertices[6] = D3DXVECTOR3(-1.0f,  1.0f,  1.0f);
	pVertices[7] = D3DXVECTOR3( 1.0f,  1.0f,  1.0f);

	int i = 0;
	pIndices[i++] = 0; pIndices[i++] = 1; pIndices[i++] = 2;
	pIndices[i++] = 2; pIndices[i++] = 1; pIndices[i++] = 3;
	pIndices[i++] = 1; pIndices[i++] = 5; pIndices[i++] = 3;
	pIndices[i++] = 3; pIndices[i++] = 5; pIndices[i++] = 7;
	pIndices[i++] = 2; pIndices[i++] = 3; pIndices[i++] = 6;
	pIndices[i++] = 6; pIndices[i++] = 3; pIndices[i++] = 7;
	pIndices[i++] = 0; pIndices[i++] = 2; pIndices[i++] = 4;
	pIndices[i++] = 4; pIndices[i++] = 2; pIndices[i++] = 6;
	pIndices[i++] = 1; pIndices[i++] = 0; pIndices[i++] = 5;
	pIndices[i++] = 5; pIndices[i++] = 0; pIndices[i++] = 4;
	pIndices[i++] = 5; pIndices[i++] = 4; pIndices[i++] = 7;
	pIndices[i++] = 7; pIndices[i++] = 4; pIndices[i++] = 6;

	m_pBoxVertexBuffer->Unlock();
	m_pBoxIndexBuffer->Unlock();

    return S_OK;
}

// -----------------------------------------------------------------------------
// GenerateQuad
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::GenerateQuad(D3DXVECTOR3& vCenter, FLOAT fRadius)
{
	HRESULT hr;

	m_dwQuadNumVertices = 4;
	m_dwQuadNumFaces    = 2;	

	hr = m_pD3DDev->CreateVertexBuffer(m_dwQuadNumVertices * sizeof(BumpVertex),
			0, 0, D3DPOOL_MANAGED, &m_pQuadVertexBuffer);
	if (FAILED(hr))
	{
		m_strLastError = "Failed to generate quad";
		return hr;
	}

	hr = m_pD3DDev->CreateIndexBuffer(3 * m_dwQuadNumFaces * sizeof(WORD),
			0, D3DFMT_INDEX16, D3DPOOL_MANAGED, &m_pQuadIndexBuffer);
	if (FAILED(hr))
	{
		m_strLastError = "Failed to generate quad";
		return hr;
	}

	BumpVertex *pVertices;
	WORD *pIndices;

	m_pQuadVertexBuffer->Lock(0, 0, (BYTE**)&pVertices, 0);
	m_pQuadIndexBuffer->Lock(0, 0, (BYTE**)&pIndices, 0);

	pVertices[0].pos = D3DXVECTOR3(vCenter.x - fRadius, vCenter.y + fRadius, 0.0f);
	pVertices[1].pos = D3DXVECTOR3(vCenter.x + fRadius, vCenter.y + fRadius, 0.0f);
	pVertices[2].pos = D3DXVECTOR3(vCenter.x - fRadius, vCenter.y - fRadius, 0.0f);
	pVertices[3].pos = D3DXVECTOR3(vCenter.x + fRadius, vCenter.y - fRadius, 0.0f);

	pVertices[0].tex = D3DXVECTOR2(0.0f, 0.0f);
	pVertices[1].tex = D3DXVECTOR2(1.0f, 0.0f);
	pVertices[2].tex = D3DXVECTOR2(0.0f, 1.0f);
	pVertices[3].tex = D3DXVECTOR2(1.0f, 1.0f);

	pVertices[0].dpds = pVertices[1].dpds =
	pVertices[2].dpds = pVertices[3].dpds = D3DXVECTOR3(1.0f, 0.0f, 0.0f);

	pVertices[0].dpdt = pVertices[1].dpdt =
	pVertices[2].dpdt = pVertices[3].dpdt = D3DXVECTOR3(0.0f, -1.0f, 0.0f);

	pIndices[0] = 0; pIndices[1] = 1; pIndices[2] = 2;
	pIndices[3] = 2; pIndices[4] = 1; pIndices[5] = 3;

	m_pQuadVertexBuffer->Unlock();
	m_pQuadIndexBuffer->Unlock();

    return S_OK;
}

// -----------------------------------------------------------------------------
// GenerateBezierPatch
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::GenerateBezierPatch()
{
	HRESULT hr;

	hr = m_pD3DDev->CreateVertexBuffer(4 * 4 * sizeof(BumpVertex), 
			D3DUSAGE_RTPATCHES | D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY, 
			0, D3DPOOL_DEFAULT, &m_pPatchBuffer);
	if (FAILED(hr))
	{
		m_strLastError = "Failed to generate bezier patch";
		return hr;
	}

	m_fNumSegs[0] = 20.0f;
	m_fNumSegs[1] = 20.0f;
	m_fNumSegs[2] = 20.0f;
	m_fNumSegs[3] = 20.0f;

	m_patchInfo.StartVertexOffsetWidth  = 0;
	m_patchInfo.StartVertexOffsetHeight = 0;
	m_patchInfo.Width  = 4;
	m_patchInfo.Height = 4;
	m_patchInfo.Stride = 4;
	m_patchInfo.Basis  = D3DBASIS_BEZIER;
	m_patchInfo.Order  = D3DORDER_CUBIC;

	return hr;
}

// -----------------------------------------------------------------------------
// InitBezierParams
// -----------------------------------------------------------------------------
inline float randf(float max) { return max * rand() / RAND_MAX; }

void CShaderBumpRefract::InitBezierParams()
{
	// amplitude of the control points
	m_fA[0]  = 0.0f; m_fA[1]  = 0.2f; m_fA[2]  = 0.2f; m_fA[3]  = 0.0f;
	m_fA[4]  = 0.2f; m_fA[5]  = 1.0f; m_fA[6]  = 1.0f; m_fA[7]  = 0.2f;
	m_fA[8]  = 0.2f; m_fA[9]  = 1.0f; m_fA[10] = 1.0f; m_fA[11] = 0.2f;
	m_fA[12] = 0.0f; m_fA[13] = 0.2f; m_fA[14] = 0.2f; m_fA[15] = 0.0f;

	for (int i=0; i<16; i++)
	{
		m_fB[i] = randf(2*D3DX_PI); // initial phase
		m_fC[i] = randf(D3DX_PI);   // frequency
	};
}

// -----------------------------------------------------------------------------
// AnimateBezierPatch
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::AnimateBezierPatch(EBTimer *pTimer)
{
	int i, j;

	BumpVertex* v;
	m_pPatchBuffer->Lock(0, 0, (BYTE**)&v, D3DLOCK_DISCARD);

	// Re-initialize the control points
	for (j=0; j<4; j++)
	for (i=0; i<4; i++)
		v[4*j+i].pos = 0.25f * D3DXVECTOR3(i-1.5f, 1.5f-j, 0.0f);

	// Update the current time
	if (!m_bPause)
		m_fTime += min(0.5f, pTimer->GetSecsPerFrame());

	// Modify the control points
	for (j=0; j<4; j++)
	for (i=0; i<4; i++)
		v[4*j+i].pos.z = m_fA[4*j+i] * sinf(m_fB[4*j+i] + m_fC[4*j+i]*m_fTime);

	// dp/ds
	for (j=0; j<4; j++)
	{
		v[4*j+0].dpds = -9*v[4*j+0].pos + 9*v[4*j+1].pos;
		v[4*j+1].dpds = -4*v[4*j+0].pos + 3*v[4*j+2].pos + 1*v[4*j+3].pos;
		v[4*j+2].dpds = -1*v[4*j+0].pos - 3*v[4*j+1].pos + 4*v[4*j+3].pos;
		v[4*j+3].dpds = -9*v[4*j+2].pos + 9*v[4*j+3].pos;
	}

	// dp/dt
	for (i=0; i<4; i++)
	{
		v[4*0+i].dpdt = -9*v[4*0+i].pos + 9*v[4*1+i].pos;
		v[4*1+i].dpdt = -4*v[4*0+i].pos + 3*v[4*2+i].pos + 1*v[4*3+i].pos;
		v[4*2+i].dpdt = -1*v[4*0+i].pos - 3*v[4*1+i].pos + 4*v[4*3+i].pos;
		v[4*3+i].dpdt = -9*v[4*2+i].pos + 9*v[4*3+i].pos;
	}

	return S_OK;
}

// -----------------------------------------------------------------------------
// Tick
// -----------------------------------------------------------------------------
HRESULT CShaderBumpRefract::Tick(EBTimer *pTimer)
{
	m_pD3DDev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0f, 0);

	m_pD3DDev->SetRenderState(D3DRS_FILLMODE,
		m_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID);

	if (m_bUsePatch)
		AnimateBezierPatch(pTimer);

	UpdateMatrices(pTimer);

	DrawEnvBox();

	UpdateShaderConstants();
	DrawObject();

	return S_OK;
}

// -----------------------------------------------------------------------------
// UpdateMatrices
// -----------------------------------------------------------------------------
#define KEYDOWN(key) (GetAsyncKeyState(key) != 0)
const D3DXMATRIX env(-1,0,0,0, 0,1,0,0, 0,0,-1,0, 0,0,0,1);

void CShaderBumpRefract::UpdateMatrices(EBTimer *pTimer)
{
	static D3DXVECTOR3 velocity(0.0f, 0.0f, 0.0f);
	D3DXVECTOR3 vel(0.0f, 0.0f, 0.0f);

	if (KEYDOWN(VK_LEFT))  vel.x -= 0.5f; // Slide Left
	if (KEYDOWN(VK_RIGHT)) vel.x += 0.5f; // Slide Right
	if (KEYDOWN(VK_NEXT))  vel.y -= 0.5f; // Slide Down
	if (KEYDOWN(VK_PRIOR)) vel.y += 0.5f; // Slide Up
	if (KEYDOWN(VK_DOWN))  vel.z -= 0.5f; // Move Backward
	if (KEYDOWN(VK_UP))    vel.z += 0.5f; // Move Forward

	const float dt = min(0.5f, pTimer->GetSecsPerFrame());
	const float a  = min(1.0f, dt/0.2f);
	velocity = velocity * (1.0f - a) + vel * a;
	m_pMouseObj->Translate(velocity * dt);

	D3DXMatrixMultiply(&m_matObj, &m_pMouseObj->GetRotationMatrix(),
									&m_pMouseObj->GetTranslationMatrix());

	D3DXMatrixMultiply(&m_matEnv, &env, &m_pMouseEnv->GetRotationMatrix());
}

// -----------------------------------------------------------------------------
// DrawEnvBox
// -----------------------------------------------------------------------------
void CShaderBumpRefract::DrawEnvBox()
{
	D3DXMATRIX matTemp;

	D3DXMatrixMultiply(&matTemp, &m_matEnv, &m_matProj);
	D3DXMatrixTranspose(&matTemp, &matTemp);
	m_pD3DDev->SetVertexShaderConstant(CV_WORLDVIEWPROJ_0, matTemp, 4);

	m_pD3DDev->SetStreamSource(0, m_pBoxVertexBuffer, sizeof(D3DXVECTOR3));
	m_pD3DDev->SetIndices(m_pBoxIndexBuffer, 0);

	m_pD3DDev->SetVertexShader(m_dwEnvBoxVertexShader);
	m_pD3DDev->SetPixelShader(0);

	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pD3DDev->SetTexture(0, m_pCubeMap);

	// Do not write to Z buffer (i.e. the size of the environment box is infinite) 
	m_pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);

	m_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
		0, m_dwBoxNumVertices, 0, m_dwBoxNumFaces);

	m_pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
}

// -----------------------------------------------------------------------------
// UpdateShaderConstants
// -----------------------------------------------------------------------------
void CShaderBumpRefract::UpdateShaderConstants()
{
	D3DXMATRIX matTemp;

	// Projection from object space to clip space
	D3DXMatrixMultiply(&matTemp, &m_matObj, &m_matProj);
	D3DXMatrixTranspose(&matTemp, &matTemp);
	m_pD3DDev->SetVertexShaderConstant(CV_WORLDVIEWPROJ_0, matTemp, 4);

	// Transform matrix from object space to cubemap space
	D3DXMATRIX matCubemap;
	D3DXMatrixInverse(&matTemp, NULL, &m_matEnv);
	D3DXMatrixMultiply(&matCubemap, &m_matObj, &matTemp);
	D3DXMatrixTranspose(&matCubemap, &matCubemap);
	m_pD3DDev->SetVertexShaderConstant(CV_CUBEMAP_0, matCubemap, 3);
	D3DXMatrixTranspose(&matCubemap, &matCubemap);

	// eye position in object space
	D3DXVECTOR3 vEyeLocal;
	D3DXMatrixInverse(&matTemp, NULL, &m_matObj);
	vEyeLocal = &matTemp(3,0);
	m_pD3DDev->SetVertexShaderConstant(CV_EYE_LOCAL,
		D3DXVECTOR4(vEyeLocal.x, vEyeLocal.y, vEyeLocal.z, 1.0f), 1);

	// eye position in cubemap space
	D3DXVECTOR3 vEyeCubmap;
	D3DXVec3TransformCoord(&vEyeCubmap, &vEyeLocal, &matCubemap);
	m_pD3DDev->SetVertexShaderConstant(CV_EYE_CUBEMAP,
		D3DXVECTOR4(vEyeCubmap.x, vEyeCubmap.y, vEyeCubmap.z, 1.0f), 1);

	// Set bumpmap scale
	m_pD3DDev->SetVertexShaderConstant(CV_BUMP_SCALE,
		D3DXVECTOR4(0.0f, 0.0f, 0.0f, m_fBumpScale), 1);

	// Set index of refraction
	m_pD3DDev->SetVertexShaderConstant(CV_N_SQUARED,
		D3DXVECTOR4(0.0f, 0.0f, 0.0f, m_fRefractIndex*m_fRefractIndex), 1);

	// Set pixel shader constants
	switch (m_eRenderStyle)
	{
		case RENDER_REFLECT:
		{
			m_pD3DDev->SetPixelShaderConstant(
				CP_REFLECTION, &D3DXCOLOR(0.9f, 0.9f, 0.9f, 0.0f), 1);
		}
		break;

		case RENDER_REFRACT:
		{
			m_pD3DDev->SetPixelShaderConstant(
				CP_REFRACTION, &D3DXCOLOR(0.9f, 0.9f, 0.9f, 0.0f), 1);
		}
		break;

		case RENDER_REFLECT_REFRACT:
		{
			m_pD3DDev->SetPixelShaderConstant(
				CP_REFLECTION, &D3DXCOLOR(0.3f, 0.3f, 0.3f, 0.0f), 1);
			m_pD3DDev->SetPixelShaderConstant(
				CP_REFRACTION, &D3DXCOLOR(0.7f, 0.7f, 0.7f, 0.0f), 1);
		}
		break;
	}
}

// -----------------------------------------------------------------------------
// DrawObject
// -----------------------------------------------------------------------------
void CShaderBumpRefract::DrawObject()
{
	// Set the vertex data and the index data
    m_pD3DDev->SetStreamSource(0, m_bUsePatch ? m_pPatchBuffer : m_pQuadVertexBuffer, sizeof(BumpVertex));
	m_pD3DDev->SetIndices(m_bUsePatch ? NULL : m_pQuadIndexBuffer, 0);

	// Set the textures
	m_pD3DDev->SetTexture(0, *m_ppCurrentBumpMap);	// Bump normal map
	m_pD3DDev->SetTexture(1, NULL);
	m_pD3DDev->SetTexture(2, NULL);
	m_pD3DDev->SetTexture(3, m_pCubeMap);			// Env cube map

	// Draw reflection
	if (m_eRenderStyle & RENDER_REFLECT)
	{
		m_pD3DDev->SetVertexShader(m_bUsePatch ? m_dwReflectVertexShader_Patch : m_dwReflectVertexShader);
		m_pD3DDev->SetPixelShader(m_dwReflectPixelShader);

		if (m_bUsePatch)
			m_pD3DDev->DrawRectPatch(1, m_fNumSegs, &m_patchInfo);
		else
			m_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
				0, m_dwQuadNumVertices, 0, m_dwQuadNumFaces);
	}

	// Draw refraction
	if (m_eRenderStyle & RENDER_REFRACT)
	{
		m_pD3DDev->SetVertexShader(m_bUsePatch ? m_dwRefractVertexShader_Patch : m_dwRefractVertexShader);
		m_pD3DDev->SetPixelShader(m_dwRefractPixelShader);

		if (m_eRenderStyle & RENDER_REFLECT)
		{
			m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
			m_pD3DDev->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
			m_pD3DDev->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
		}

		if (m_bUsePatch)
			m_pD3DDev->DrawRectPatch(1, m_fNumSegs, &m_patchInfo);
		else
			m_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
				0, m_dwQuadNumVertices, 0, m_dwQuadNumFaces);
	}

	m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
}

#define STR_DECBUMPSCALE	"Decrease Bump Scale \t -"
#define STR_INCBUMPSCALE	"Increase Bump Scale \t +"
#define STR_NEXTBUMPMAP		"Next Bump Map \t N"
#define STR_RESET			"Reset to Defaults \t Home"
// -----------------------------------------------------------------------------
// UpdateProperties
// -----------------------------------------------------------------------------
void CShaderBumpRefract::UpdateProperties()
{
	EBEnumProperty *pEnumProp;

	EBEffect::UpdateProperties();

	AddProperty(new EBTriggerProperty(STR_DECBUMPSCALE));
	AddProperty(new EBTriggerProperty(STR_INCBUMPSCALE));
	AddProperty(new EBTriggerProperty(STR_NEXTBUMPMAP));
	AddProperty(new EBTriggerProperty(STR_RESET));
	AddProperty(new EBProperty(
		"Pause \t Space", OBJECT_MEMBER(m_bPause), EBTYPE_BOOL_PROP));
	AddProperty(new EBProperty(
		"Wireframe \t W", OBJECT_MEMBER(m_bWireframe), EBTYPE_BOOL_PROP));

	pEnumProp = new EBEnumProperty(
		"zzz Mouse", OBJECT_MEMBER(m_eMouseState), EBTYPE_DWORD_PROP);
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"Mouse - Move Object", (DWORD)MOUSE_OBJ, EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"Mouse - Move Environment", (DWORD)MOUSE_ENV, EBTYPE_DWORD_PROP));
	AddProperty(pEnumProp);

	float refraction[] = { 1.0f, 1.3f, 1.5f, 2.0f, 2.4f };
	pEnumProp = new EBEnumProperty(
		"zzz Reflection Index", OBJECT_MEMBER(m_fRefractIndex), EBTYPE_DWORD_PROP);
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"1.0 (Vacuum)", *(DWORD*)&refraction[0], EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"1.3 (Ice)", *(DWORD*)&refraction[1], EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"1.5 (Glass)", *(DWORD*)&refraction[2], EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"2.0 (Crystal)", *(DWORD*)&refraction[3], EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"2.4 (Diamond)", *(DWORD*)&refraction[4], EBTYPE_DWORD_PROP));
	AddProperty(pEnumProp);

	pEnumProp = new EBEnumProperty(
		"zzz Render", OBJECT_MEMBER(m_eRenderStyle), EBTYPE_DWORD_PROP);
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"Reflection only", (DWORD)RENDER_REFLECT, EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"Refraction only", (DWORD)RENDER_REFRACT, EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"Reflection and Refraction", (DWORD)RENDER_REFLECT_REFRACT, EBTYPE_DWORD_PROP));
	AddProperty(pEnumProp);

	pEnumProp = new EBEnumProperty(
		"zzz Shape", OBJECT_MEMBER(m_bUsePatch), EBTYPE_DWORD_PROP);
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
		"Simple Quad", (DWORD)FALSE, EBTYPE_DWORD_PROP));
	if (m_bSupportsPatches)
		pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp,
			"Dynamic Bezier-Patch", (DWORD)TRUE, EBTYPE_DWORD_PROP));
	AddProperty(pEnumProp);

	// Vertex shaders
	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum,
		"Reflection", GetFilePath("reflect.nvv"), EBTYPE_STRING_PROP));
	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum,
		"Refraction", GetFilePath("refract.nvv"), EBTYPE_STRING_PROP));

	// Pixel shaders
	m_pPixelShaderEnum->AddEnumerant(new EBEnumValue(m_pPixelShaderEnum,
		"Reflection", GetFilePath("reflect.nvp"), EBTYPE_STRING_PROP));
	m_pPixelShaderEnum->AddEnumerant(new EBEnumValue(m_pPixelShaderEnum,
		"Refraction", GetFilePath("refract.nvp"), EBTYPE_STRING_PROP));

	std::string aboutText = "file://";
    aboutText += GetFilePath("..\\..\\..\\NVEffectsExplained.htm");
    aboutText += "#bumprefraction";
	SetAboutInfo( NULL, _T("Bump Refraction"), _T(aboutText.c_str()));

}

// -----------------------------------------------------------------------------
// PropertyUpdateCallback
// -----------------------------------------------------------------------------
void CShaderBumpRefract::PropertyUpdateCallback(const EBProperty *pProperty, bool bWritten)
{
	if (!bWritten)
		return;

	EBString	name;

	if (pProperty->IsKindOf(EBTYPE_TRIGGER_PROP))
	{
		name = pProperty->GetPropertyName();
		
		if (name == EBString(STR_NEXTBUMPMAP))
			Keyboard('N', 0, true);
		else if (name == EBString(STR_DECBUMPSCALE))
			Keyboard(VK_ADD, 0, true);
		else if (name == EBString(STR_INCBUMPSCALE))
			Keyboard(VK_SUBTRACT, 0, true);
		else if (name == EBString(STR_RESET))
			Keyboard(VK_HOME, 0, true);
	}
}

// -----------------------------------------------------------------------------
// Keyboard
// -----------------------------------------------------------------------------
void CShaderBumpRefract::Keyboard(DWORD dwKey, UINT nFlags, bool bDown)
{
	eEBKeyAction Action = TranslateEffectKey(dwKey, nFlags, bDown);

	switch (Action)
	{
		case EB_HELP:
		{
			::MessageBoxEx(NULL,
				" Help : F1 - Help \n\n"
				" Home - Reset To Defaults \n\n"
				" W - Wireframe Toggle \n\n"
				" Space/Pause - Toggle Pause/Resume \n\n"
				" Left Button & Mouse - Rotate Object\n\n"
				" Shift Left Button & Mouse - Pan Object \n\n"
				" Ctrl Left Button & Mouse - Move Object In & Out\n\n",
				"Help", MB_ICONINFORMATION | MB_TASKMODAL,
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT));
		}
		break;

		case EB_WIREFRAME:
		{
			m_bWireframe = !m_bWireframe;
			m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
		}
		break;

		case EB_RESET:
		{
			m_pMouseObj->Reset();
			m_pMouseObj->SetTranslationalSensitivityFactor(0.1f);
			m_pMouseObj->SetRotationalSensitivityFactor(0.5f);
			m_pMouseObj->Translate(0.0f, 0.0f, 1.0f);

			m_pMouseEnv->Reset();
			m_pMouseEnv->SetRotationalSensitivityFactor(0.5f);

			m_fTime				= 0.0f;
			m_fBumpScale		= 0.5f;

			InitBezierParams();

			m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
		}
		break;

		case EB_PAUSE:
		{
			m_bPause = !m_bPause;
			m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
		}
		break;

		case EB_ADD:
		{
			m_fBumpScale += 0.02f;
			if (m_fBumpScale > 2.0f)
				m_fBumpScale = 2.0f;
		}
		break;

		case EB_SUBTRACT:
		{
			m_fBumpScale -= 0.02f;
			if (m_fBumpScale < 0.0f)
				m_fBumpScale = 0.0f;
		}
		break;

		case EB_NORMALMAP:
		{
			m_ppCurrentBumpMap++;
			if (m_ppCurrentBumpMap == m_pBumpMaps.end())
				m_ppCurrentBumpMap = m_pBumpMaps.begin();
		}
		break;
	}
}

// -----------------------------------------------------------------------------
// MouseButton
// -----------------------------------------------------------------------------
void CShaderBumpRefract::MouseButton(HWND hWnd, eButtonID button, bool bDown, int x, int y)
{
	if (button == MOUSE_LEFTBUTTON)
	{
		if (m_eMouseState == MOUSE_OBJ)
		{
			if (bDown)
				m_pMouseObj->OnLButtonDown(x, y);
			else
				m_pMouseObj->OnLButtonUp(x, y);
		}
		else if (m_eMouseState == MOUSE_ENV)
		{
			if (bDown)
				m_pMouseEnv->OnLButtonDown(x, y);
			else
				m_pMouseEnv->OnLButtonUp(x, y);
		}
	}
}

// -----------------------------------------------------------------------------
// MouseMove
// -----------------------------------------------------------------------------
void CShaderBumpRefract::MouseMove(HWND hWnd, int x, int y)
{
	if (m_eMouseState == MOUSE_OBJ)
		m_pMouseObj->OnMouseMove(x, y);
	else if (m_eMouseState == MOUSE_ENV)
		m_pMouseEnv->OnMouseMove(x, y);
}

