/******************************************************************************

  Copyright (C) 1999, 2000 NVIDIA Corporation
  This file is provided without support, instruction, or implied warranty of any
  kind.  NVIDIA makes no guarantee of its fitness for a particular purpose and is
  not liable under any circumstances for any damages or loss whatsoever arising
  from the use or inability to use this file or items derived from it.
  
    Comments:
    
      
        
******************************************************************************/

#include "eb_effect.h"
#include "shader_ParticleSystem.h"
#include "shader_ParticleSystem2.h"
#include "nvinc.h"
#include "shader_constants.h"
#include "ParticleDialog\common.h"

using namespace std;

DECLARE_EFFECT_MAIN()

#define MAX_PARTICLE_VERTS (kNumParticles * 3)

const float    PI		 = (float)(3.14159265358979323846264338327950288419716939937511);
const float    TWO_PI	 = (float)(2.0*PI);

static float RandomNumberInRange(const float& Min, const float& Max)
{
	float Range = Max - Min;
	float Random = (float)(abs(rand()));
	int value = rand();
	float fvalue = (float)value;
	float fprecision = (float)RAND_MAX;
	float fdecimal = fvalue/fprecision;

	Random += fdecimal;

	int iRange = (int)Range;
	float Multiplier = 1.0f;
	if (iRange > RAND_MAX)
	{
		Multiplier = (float)(iRange / RAND_MAX);
	}

	Random = fmod(Random,Range);
	Random *= Multiplier;
	Random += Min; 

	return (Random);
}

static void SetFromHeadingAndPitch(D3DXVECTOR4& Vector, float heading, float pitch)
{
	Vector = D3DXVECTOR4(0.0f, 1.0f, 0.0f, 0.0f);

	D3DXMATRIX d3dMatrix;
	D3DXMatrixRotationYawPitchRoll(&d3dMatrix, heading, pitch, 0.0f);

	D3DXVec4Transform(&Vector, &Vector, &d3dMatrix);
}

extern "C"
{

__declspec(dllexport) unsigned int GetNumEffects() { return 2; }

__declspec(dllexport) EBEffect* CreateEffect(unsigned int EffectNum)
{	
	switch(EffectNum)
	{
		case 0:
			return new CParticleSystemShader();
			break;
		case 1:
			return new CParticleSystemShader2();
			break;
		default:
			return NULL;
	}
}

}

CParticleSystemShader::CParticleSystemShader()
	: m_pVertexBuffer(NULL),
	m_pLogoTexture(NULL),
	m_dwCurrentShader(0),
	m_bWireframe(false),
	m_pUI(NULL),
	m_hDialog(NULL)
{
	m_strEffectLocation = "3rd Party\\Physics";
	m_strEffectName = "Particle System";
	m_strEffectVertexShader = GetFilePath("particlesystem.nvv");
	m_strEffectPixelShader = "";
	m_strEffectVersion = "1.0";
}

void CParticleSystemShader::UpdateProperties()
{
	EBEffect::UpdateProperties();

	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum, "Dynamic Particle System", GetFilePath("particlesystem.nvv"), EBTYPE_STRING_PROP));
	AddProperty(new EBProperty("Wireframe", OBJECT_MEMBER(m_bWireframe), EBTYPE_BOOL_PROP));

	AddProperty(new EBProperty("TriangleCount", OBJECT_MEMBER(m_iTris), EBTYPE_DWORD_PROP));
	AddProperty(new EBProperty("VertexCount", OBJECT_MEMBER(m_iVerts), EBTYPE_DWORD_PROP));

    string aboutText = "file://";
    aboutText += GetFilePath("..\\..\\..\\NVEffectsExplained.htm");
    aboutText += "#particles";
	SetAboutInfo( NULL, _T("Dynamic Particle System"), _T(aboutText.c_str()));
	SetAboutInfo( NULL, _T("Greg Snook"), _T("http://www.mightystudios.com"));
}

CParticleSystemShader::~CParticleSystemShader()
{
	if (m_hDialog)
	{
		if (m_DialogFunc.Free)
		{
			m_DialogFunc.Free();
			ZeroMemory(&m_DialogFunc, sizeof(m_DialogFunc));
		}

		FreeLibrary(m_hDialog);
		m_hDialog = NULL;
	}
	Free();	
}

HRESULT CParticleSystemShader::Initialize(LPDIRECT3DDEVICE8 pDev)
{
	// attach to the dialog app
	ZeroMemory(&m_DialogFunc, sizeof(OptionsDialogFunc));
	m_hDialog = LoadLibrary(GetFilePath("particledialog.dll").c_str());
	if (!m_hDialog || m_hDialog == (HMODULE)0xcdcdcdcd)
	{
		m_strLastError = "Could not load particledialog.dll";
		return E_FAIL;
	}

	m_DialogFunc.ShowDialog = (LPSHOWDIALOG)GetProcAddress(m_hDialog, "ParticleDialog_ShowDialog");
	m_DialogFunc.Initialise = (LPINITIALISE)GetProcAddress(m_hDialog, "ParticleDialog_Initialise");
	m_DialogFunc.Free = (LPFREE)GetProcAddress(m_hDialog, "ParticleDialog_Free");
	m_DialogFunc.GetSettings = (LPGETSETTINGS)GetProcAddress(m_hDialog, "ParticleDialog_GetSettings");
	m_DialogFunc.SetSettings = (LPSETSETTINGS)GetProcAddress(m_hDialog, "ParticleDialog_SetSettings");

	if (!m_DialogFunc.ShowDialog ||
		!m_DialogFunc.Initialise ||
		!m_DialogFunc.Free ||
		!m_DialogFunc.GetSettings ||
		!m_DialogFunc.SetSettings)
	{
		m_strLastError = "Could not resolve dialog box functions!";
		return E_FAIL;
	}
    
	
	ZeroMemory(&m_bKey[0], sizeof(bool) * kMaxVKey);

	HRESULT hr;

	m_pD3DDev = pDev;
	pDev->AddRef();

	//initialize mouse UI
	RECT rect;
	rect.left = rect.top = 0;
	D3DVIEWPORT8 viewport;
	m_pD3DDev->GetViewport(&viewport);
	rect.bottom = viewport.Height;
	rect.right  = viewport.Width;
	m_pUI = new MouseUI((const RECT)rect);

	vector<DWORD> Declaration;
	Declaration.push_back(D3DVSD_STREAM(0));
	Declaration.push_back(D3DVSD_REG(0, D3DVSDT_FLOAT4));
	Declaration.push_back(D3DVSD_END());

	hr = LoadAndCreateShader(GetFilePath("particlesystem.vso"), &Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwCurrentShader);
	if (FAILED(hr))
		return hr;

    hr = m_pD3DDev->CreateVertexBuffer( MAX_PARTICLE_VERTS * sizeof(ParticleVertex), D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &m_pVertexBuffer);
	if (FAILED(hr))
		return hr;
		
	hr = m_pD3DDev->SetStreamSource(0, m_pVertexBuffer, sizeof(ParticleVertex));
	if (FAILED(hr))
		return hr;

	hr = m_pD3DDev->SetVertexShader(m_dwCurrentShader);
	if (FAILED(hr))
		return hr;

	ParticleVertex* pBuff;
	if (m_pVertexBuffer)
	{
		float fScale = 1.0f;
		hr = m_pVertexBuffer->Lock(0, sizeof(ParticleVertex) * MAX_PARTICLE_VERTS,(BYTE**)&pBuff, 0);
		if (FAILED(hr))
		{
			m_strLastError = "Couldn't lock buffer!";
			return hr;
		}

		// fill vertex buffer with sample data
		for (int i=0; i<kNumParticles; ++i)
		{
			// generate random data for this particle
			D3DXVECTOR3 RandomVector(RandomNumberInRange(-1.0f, 1.0f), RandomNumberInRange(-1.0f, 1.0f), RandomNumberInRange(-1.0f, 1.0f));
			D3DXVECTOR3 LaunchVector;
			D3DXVec3Normalize(&LaunchVector, &RandomVector);

			float birth_offset = RandomNumberInRange(0.0f, 0.999f);

			// setup the tree verts that make up this particle triangle
			for (float v=0.0f;v<3.0f;v+= 1.0f)
			{
				pBuff->LaunchVector = LaunchVector;
				pBuff->Data = v + birth_offset;
				pBuff++;

			}
		}

		m_pVertexBuffer->Unlock();
	}

	// We always draw three verts or one triangles...
	m_iTris  = 1;
	m_iVerts = 3;

	// Create the particle texture
	lstrcpy(m_ParticleSettings.texture_path, "particle.dds");
	hr = D3DXCreateTextureFromFileEx(m_pD3DDev, 
		GetFilePath(m_ParticleSettings.texture_path).c_str(),
		D3DX_DEFAULT,
		D3DX_DEFAULT,
		0,
		0,
		D3DFMT_UNKNOWN,
		D3DPOOL_MANAGED,
		D3DX_FILTER_LINEAR,
		D3DX_FILTER_LINEAR,
		0,
		NULL,
		NULL,
		&m_pLogoTexture);
	if (FAILED(hr))
	{
		m_strLastError = "Could not create particle texture";
		return hr;
	}

	m_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

	m_pD3DDev->SetTexture(0, m_pLogoTexture);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

	m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);

	m_pD3DDev->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
	m_pD3DDev->SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

	m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);

	m_pD3DDev->SetRenderState( D3DRS_SPECULARENABLE, FALSE );
	m_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE,   TRUE );
	m_pD3DDev->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
	m_pD3DDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

	// Enable alpha testing (avoids drawing pixels with less than a certain
	// alpha.)
	m_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
	m_pD3DDev->SetRenderState( D3DRS_ALPHAREF,        0x08 );
	m_pD3DDev->SetRenderState( D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL );

	// setup table and constants
	m_ConstantValues = D3DXVECTOR4(32.0f, 1.0f/32.0f, 4.0f, 0.0f);

	float angle = 0.0f;
	float angle_delta = TWO_PI/32.0f;

	for (int i=0;i<32;++i)
	{
		float sine = sin(angle);
		float cosine = cos(angle);
		m_SineCosineTable[i] = D3DXVECTOR4(cosine, -sine, sine, cosine);
		angle += angle_delta;
	}

	D3DXVECTOR3 Normal(1.0f, 1.0f, 0.0f);
	D3DXVECTOR3 Offset;
	D3DXVec3Normalize(&Offset, &Normal);

	m_VertexData[0] =  D3DXVECTOR4(0.0f, Offset.y, 1.0f, 0.0f);
	m_VertexData[1] =  D3DXVECTOR4(Offset.x, -Offset.y, 0.0f, 0.0f);
	m_VertexData[2] =  D3DXVECTOR4(-Offset.x, -Offset.y, 0.5f, 1.0f);

	// setup the UV lookup table for our 4 animation frames
	m_UVTable[0] = D3DXVECTOR4(0.5f, 0.5f, 0.0f, 0.0f);
	m_UVTable[1] = D3DXVECTOR4(0.5f, 0.5f, 0.0f, 0.0f);
	m_UVTable[2] = D3DXVECTOR4(0.5f, 0.5f, 0.0f, 0.0f);
	m_UVTable[3] = D3DXVECTOR4(0.5f, 0.5f, 0.0f, 0.0f);

	m_UVTable[4] = D3DXVECTOR4(1.0f, 1.0f, 0.0f, 0.0f);
	m_UVTable[5] = D3DXVECTOR4(0.0f, 1.0f, 0.0f, 0.0f);
	m_UVTable[6] = D3DXVECTOR4(0.0f, 0.0f, 0.0f, 0.0f);
	m_UVTable[7] = D3DXVECTOR4(1.0f, 0.0f, 0.0f, 0.0f);

	m_UVTable[8] = D3DXVECTOR4(0.0f, 1.0f, 0.0f, 0.0f);
	m_UVTable[9] = D3DXVECTOR4(0.0f, 0.0f, 0.0f, 0.0f);
	m_UVTable[10] = D3DXVECTOR4(1.0f, 0.0f, 0.0f, 0.0f);
	m_UVTable[11] = D3DXVECTOR4(1.0f, 1.0f, 0.0f, 0.0f);

	// set default startup values
	m_ParticleSettings.total_particles = kStartParticles;
	m_ParticleSettings.lifetime = 6.0f;

	m_ParticleSettings.launch_speed = 2.0f;
	m_ParticleSettings.launch_accel = 0.95f;

	m_ParticleSettings.start_scale = 0.4f;
	m_ParticleSettings.end_scale = 0.8f;

	m_ParticleSettings.start_rotation = 0.0f;
	m_ParticleSettings.end_rotation = 200.0f;

	m_ParticleSettings.wind_heading = 2.00f;
	m_ParticleSettings.wind_pitch = 1.00f;
	m_ParticleSettings.wind_scale = 1.67f;
	m_ParticleSettings.wind_start_effect = 0.0f;
	m_ParticleSettings.wind_end_effect = 0.33f;

	m_ParticleSettings.gravity_heading = 0.0f;
	m_ParticleSettings.gravity_pitch = 6.0f;
	m_ParticleSettings.gravity_scale = 0.6f;
	m_ParticleSettings.gravity_acceleration = 1.21f;

	m_ParticleSettings.magnet_heading = 0.0f;
	m_ParticleSettings.magnet_pitch = 1.93f;
	m_ParticleSettings.magnet_distance = 10.0f;
	m_ParticleSettings.magnet_strength = 0.73f;
	m_ParticleSettings.magnet_start_effect = 0.0f;
	m_ParticleSettings.magnet_end_effect = 0.0f;

	m_ParticleSettings.start_red = 1.0f;
	m_ParticleSettings.start_green = 0.0f;
	m_ParticleSettings.start_blue = 0.0f;
	m_ParticleSettings.start_alpha = 1.0f;

	m_ParticleSettings.end_red = 1.0f;
	m_ParticleSettings.end_green = 1.0f;
	m_ParticleSettings.end_blue = 0.5f;
	m_ParticleSettings.end_alpha = 0.75f;

	m_Origin = D3DXVECTOR4(0.0f, 0.0f, 0.0f, 1.0f);

	SetFromHeadingAndPitch(m_Magnet, m_ParticleSettings.magnet_heading, m_ParticleSettings.magnet_pitch);
	m_Magnet *= m_ParticleSettings.magnet_distance;

	SetFromHeadingAndPitch(m_Gravity, m_ParticleSettings.gravity_heading, m_ParticleSettings.gravity_pitch);
	m_Gravity *= m_ParticleSettings.gravity_scale;

	SetFromHeadingAndPitch(m_Wind, m_ParticleSettings.wind_heading, m_ParticleSettings.wind_pitch);
	m_Wind *= m_ParticleSettings.wind_scale;

	m_StartColor = D3DXVECTOR4(	m_ParticleSettings.start_red,
								m_ParticleSettings.start_green,
								m_ParticleSettings.start_blue,
								m_ParticleSettings.start_alpha);

	m_EndColor = D3DXVECTOR4(	m_ParticleSettings.end_red - m_ParticleSettings.start_red,
								m_ParticleSettings.end_green - m_ParticleSettings.start_green,
								m_ParticleSettings.end_blue - m_ParticleSettings.start_blue,
								m_ParticleSettings.end_alpha - m_ParticleSettings.start_alpha);

	WriteVertexShaderConstants();



	D3DDEVICE_CREATION_PARAMETERS CreateParams;	
	pDev->GetCreationParameters(&CreateParams);
	HWND hParent = GetParent(CreateParams.hFocusWindow);
	if (!m_DialogFunc.Initialise(CreateParams.hFocusWindow))
	{
		m_strLastError = "Could not initialize dialog box!";
		return E_FAIL;
	}

	m_DialogFunc.ShowDialog(true);
	m_DialogFunc.SetSettings(&m_ParticleSettings);

	return S_OK;
}

void CParticleSystemShader::WriteVertexShaderConstants() const
{
    // All of the following are write-once!
	m_pD3DDev->SetVertexShaderConstant(CV_CONSTANTS, &m_ConstantValues, 1);
	m_pD3DDev->SetVertexShaderConstant(CV_VERTEX_0, &m_VertexData[0], 1);
	m_pD3DDev->SetVertexShaderConstant(CV_VERTEX_1, &m_VertexData[1], 1);
	m_pD3DDev->SetVertexShaderConstant(CV_VERTEX_2, &m_VertexData[2], 1);

	for (int i=0; i<32; ++i)
	{
		m_pD3DDev->SetVertexShaderConstant(CV_SINCOS_TABLE_0 + i, &m_SineCosineTable[i], 1);
	}

	for (i=0; i<12; ++i)
	{
		m_pD3DDev->SetVertexShaderConstant(CV_UV_TABLE_V0_FRAME0 + i, &m_UVTable[i], 1);
	}

	
}

HRESULT CParticleSystemShader::Free()
{
	SAFE_RELEASE(m_pLogoTexture);
	SAFE_RELEASE(m_pVertexBuffer);
	SAFE_DELETE(m_pUI);

	if (m_pD3DDev)
	{
		if (m_dwCurrentShader)
			m_pD3DDev->DeleteVertexShader(m_dwCurrentShader);
		SAFE_RELEASE(m_pD3DDev);
	}
	
	return S_OK;
}

HRESULT CParticleSystemShader::Start()
{
	return S_OK;
}

void CParticleSystemShader::SetViewFromKeyboardInput(EBTimer* pTimer, D3DXMATRIX &matView)
{
    float   const   kSpeed              = pTimer->GetSecsPerFrame() * 10.0f;
    float   const   moveSpeed           = kSpeed * 2.0f;

    D3DXVECTOR3 const   yVector(0.0f, 1.0f, 0.0f);
    D3DXVECTOR3         lookDir(0.0f, 0.0f, 1.0f);    
    D3DXVECTOR4         position;
    D3DXMATRIX          matLookRotations;

    D3DXMatrixTranspose(&matLookRotations, &(m_pUI->GetRotationMatrix()));
    D3DXVec3TransformNormal(&lookDir, &lookDir, &matLookRotations);

    float       const   h = D3DXVec3Dot(&yVector, &lookDir);
    D3DXVECTOR3         xzProjectedLookDirection(lookDir - h*yVector);

    D3DXVec3Normalize(&xzProjectedLookDirection, &xzProjectedLookDirection);
    
    D3DXVECTOR3         rightVector;         
    D3DXVec3Cross(&rightVector, &yVector, &xzProjectedLookDirection);

    if(m_bKey[VK_UP]    || m_bKey[VK_NUMPAD8])
        m_pUI->Translate(  moveSpeed*xzProjectedLookDirection);
	if(m_bKey[VK_DOWN]  || m_bKey[VK_NUMPAD2])
        m_pUI->Translate(- moveSpeed*xzProjectedLookDirection);
	if(m_bKey[VK_LEFT]  || m_bKey[VK_NUMPAD4])
        m_pUI->Translate(- 0.5f*moveSpeed*rightVector);
	if(m_bKey[VK_RIGHT] || m_bKey[VK_NUMPAD6])              
        m_pUI->Translate(  0.5f*moveSpeed*rightVector);
	if(m_bKey[VK_PRIOR])
        m_pUI->Translate(  0.25f*moveSpeed*yVector);
	if(m_bKey[VK_NEXT])
        m_pUI->Translate(- 0.25f*moveSpeed*yVector);
    

    D3DXVec3Transform(&position, &D3DXVECTOR3(0.0f, 8.0f, -20.0f), &m_pUI->GetTranslationMatrix());

    D3DXVECTOR3 const   camPosition = D3DXVECTOR3(position.x, max(0.1f, position.y), position.z);
    D3DXVECTOR3 const   vLookatPt   = camPosition + lookDir;         

    D3DXMatrixLookAtLH(&matView, &camPosition, &vLookatPt, &yVector);
}

HRESULT CParticleSystemShader::Tick(EBTimer* pTimer)
{
	HRESULT hr = S_OK;
	D3DXMATRIX matWorld;
    D3DXMATRIX matView;
	D3DXMATRIX matProj;
	D3DXMATRIX matWorldView;

	m_DialogFunc.GetSettings(&m_ParticleSettings);

	if (m_ParticleSettings.dirty_flag)
	{
		SetFromHeadingAndPitch(m_Magnet, m_ParticleSettings.magnet_heading, m_ParticleSettings.magnet_pitch);
		m_Magnet *= m_ParticleSettings.magnet_distance;
		m_Magnet += m_Origin;
		m_Magnet.w = 1.0f;

		SetFromHeadingAndPitch(m_Gravity, m_ParticleSettings.gravity_heading, m_ParticleSettings.gravity_pitch);
		m_Gravity *= m_ParticleSettings.gravity_scale;

		SetFromHeadingAndPitch(m_Wind, m_ParticleSettings.wind_heading, m_ParticleSettings.wind_pitch);
		m_Wind *= m_ParticleSettings.wind_scale;

		m_StartColor = D3DXVECTOR4(	m_ParticleSettings.start_red,
									m_ParticleSettings.start_green,
									m_ParticleSettings.start_blue,
									m_ParticleSettings.start_alpha);

		m_EndColor = D3DXVECTOR4(	m_ParticleSettings.end_red - m_ParticleSettings.start_red,
									m_ParticleSettings.end_green - m_ParticleSettings.start_green,
									m_ParticleSettings.end_blue - m_ParticleSettings.start_blue,
									m_ParticleSettings.end_alpha - m_ParticleSettings.start_alpha);

		m_pD3DDev->SetVertexShaderConstant(CV_COLOR_A, &m_StartColor, 1);
		m_pD3DDev->SetVertexShaderConstant(CV_COLOR_B, &m_EndColor, 1);

		D3DXVECTOR4 temp;

		temp = D3DXVECTOR4(m_ParticleSettings.start_scale, m_ParticleSettings.end_scale - m_ParticleSettings.start_scale, m_ParticleSettings.start_rotation, m_ParticleSettings.end_rotation - m_ParticleSettings.start_rotation);
		m_pD3DDev->SetVertexShaderConstant(CV_SCALE_AND_SPIN, &temp, 1);

		temp = D3DXVECTOR4(m_ParticleSettings.magnet_start_effect, m_ParticleSettings.magnet_end_effect - m_ParticleSettings.magnet_start_effect, 0.0f, 0.0f);
		m_pD3DDev->SetVertexShaderConstant(CV_MAGNET_EFFECT, &temp, 1);

		temp = D3DXVECTOR4(m_ParticleSettings.wind_start_effect, m_ParticleSettings.wind_end_effect - m_ParticleSettings.wind_start_effect, 0.0f, 0.0f);
		m_pD3DDev->SetVertexShaderConstant(CV_WIND_EFFECT, &temp, 1);

		temp = D3DXVECTOR4(m_ParticleSettings.launch_accel, m_ParticleSettings.launch_accel, m_ParticleSettings.launch_accel, 0.0f);
		m_pD3DDev->SetVertexShaderConstant(CV_TRAJECTORY_ACCEL, &temp, 1);

		temp = D3DXVECTOR4(m_ParticleSettings.gravity_acceleration, m_ParticleSettings.gravity_acceleration, m_ParticleSettings.gravity_acceleration, 0.0f);
		m_pD3DDev->SetVertexShaderConstant(CV_GRAVITY_ACCEL, &temp, 1);

		temp = D3DXVECTOR4(m_ParticleSettings.launch_speed, m_ParticleSettings.launch_speed, m_ParticleSettings.launch_speed, 1.0f);
		m_pD3DDev->SetVertexShaderConstant(CV_TRAJECTORY_SPEED, &temp, 1);

		temp = D3DXVECTOR4(m_ParticleSettings.magnet_strength, m_ParticleSettings.magnet_strength, m_ParticleSettings.magnet_strength, 0.0f);
		m_pD3DDev->SetVertexShaderConstant(CV_MAGNET_STRENGTH, &temp, 1);

		if (m_ParticleSettings.new_path_name)
		{
			m_pD3DDev->SetTexture(0, 0);
			SAFE_RELEASE(m_pLogoTexture);

			// load new texture
			// Create the particle texture
			hr = D3DXCreateTextureFromFileEx(m_pD3DDev, 
				GetFilePath(m_ParticleSettings.texture_path).c_str(),
				D3DX_DEFAULT,
				D3DX_DEFAULT,
				0,
				0,
				D3DFMT_UNKNOWN,
				D3DPOOL_MANAGED,
				D3DX_FILTER_LINEAR,
				D3DX_FILTER_LINEAR,
				0,
				NULL,
				NULL,
				&m_pLogoTexture);
			if (FAILED(hr))
			{
				m_strLastError = "Could not create particle texture";
				return hr;
			}

			m_pD3DDev->SetTexture(0, m_pLogoTexture);
		}
	}

	hr = m_pD3DDev->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB( 0xAA, 0xAA, 0xAA ), 1.0, 0);

	m_pD3DDev->SetRenderState(D3DRS_FILLMODE, m_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID);

	D3DXVECTOR3 vEyePt = D3DXVECTOR3( 0.0f, 0.0f, -5.0f );
	D3DXVECTOR3 vLookatPt = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	D3DXVECTOR3 vUp = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );

	// Setup the world, view and projection matrices
	D3DXMatrixIdentity(&matWorld);

    SetViewFromKeyboardInput(pTimer, matView);

	D3DXMatrixPerspectiveFovLH(&matProj, D3DXToRadian(60.0f), 1.0f, 1.0f, 1000.0f);
		
	// Setup the composite matrix to get vertices to camera space
	D3DXMatrixMultiply(&matWorldView, &matWorld, &matView);
	
	// Setup constants
	D3DXMatrixTranspose(&matProj, &matProj);
//	D3DXMatrixTranspose(&matWorldView, &matWorldView);

	m_pD3DDev->SetVertexShaderConstant(CV_PROJ_0, &matProj(0, 0), 4);
	m_pD3DDev->SetVertexShaderConstant(CV_WORLDVIEW_0, &matWorldView(0, 0), 4);

	m_ElapsedTime = pTimer->GetDuration();
	D3DXVECTOR4 vTime = D3DXVECTOR4(m_ParticleSettings.lifetime, m_ParticleSettings.lifetime*0.5f, 1/m_ParticleSettings.lifetime, m_ElapsedTime/m_ParticleSettings.lifetime);
	m_pD3DDev->SetVertexShaderConstant(CV_TIME, &vTime, 1);

	D3DXVECTOR4 Coordinate;

	D3DXVec4Transform(&Coordinate, &m_Origin, &matWorldView);
	m_pD3DDev->SetVertexShaderConstant(CV_ORIGIN, &Coordinate, 1);

	D3DXVec4Transform(&Coordinate, &m_Magnet, &matWorldView);
	m_pD3DDev->SetVertexShaderConstant(CV_MAGNET, &Coordinate, 1);

	D3DXVECTOR4 Vector;

	D3DXVec4Transform(&Vector, &m_Gravity, &matWorldView);
	Vector.w = 0.0f;
	m_pD3DDev->SetVertexShaderConstant(CV_GRAVITY, &Vector, 1);

	D3DXVec4Transform(&Vector, &m_Wind, &matWorldView);
	Vector.w = 0.0f;
	m_pD3DDev->SetVertexShaderConstant(CV_WIND, &Vector, 1);

	hr = m_pD3DDev->DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_ParticleSettings.total_particles);

	return hr;
}

HRESULT CParticleSystemShader::ConfirmDevice(D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format)
{
    // check vertex shading support
    if(D3DSHADER_VERSION_MAJOR(pCaps->VertexShaderVersion) < 1)
	{
		m_strLastError = "Device does not support vertex shaders!";
		return E_FAIL;
	}
    if(D3DSHADER_VERSION_MINOR(pCaps->VertexShaderVersion) < 1)
	{
		m_strLastError = "Device does not support 1.1 vertex shaders!";
		return E_FAIL;
	}

    // check simultaneous texture support (only need 1...)
    if(pCaps->MaxSimultaneousTextures < 1)
	{
		m_strLastError = "Device does not support 1 simultaneous textures!";
		return E_FAIL;
	}

    // only need very few, simple register combiners...
	if(pCaps->MaxTextureBlendStages < 2)
	{
		m_strLastError = "Device does not support 2 register combiners!";
		return E_FAIL;
	}

    // need mipmapping 
	if ((pCaps->TextureCaps & D3DPTEXTURECAPS_MIPMAP) == 0)
	{
		m_strLastError = "Device does not support mip-mapping!";
		return E_FAIL;
	}
    // and trilinear filtering between mipmaps
	if ((pCaps->TextureFilterCaps & D3DPTFILTERCAPS_MIPFLINEAR) == 0)
	{
		m_strLastError = "Device does not support trilinear mip-mapping!";
		return E_FAIL;
	}

	return S_OK;
}

void CParticleSystemShader::MouseButton(HWND hWnd, eButtonID button, bool bDown, int x, int y)
{
	if(button == MOUSE_LEFTBUTTON)
	{
		if(bDown)
		{
			m_pUI->OnLButtonDown(x, y);
		}
		else
		{
			m_pUI->OnLButtonUp(x, y);
		}
	}
	return;
}

void CParticleSystemShader::MouseMove(HWND hWnd, int x, int y)
{
	m_pUI->OnMouseMove(x, y);
	return;
}

void CParticleSystemShader::Keyboard(DWORD vKey, UINT nFlags, bool bDown)
{
	eEBKeyAction Action = TranslateEffectKey(vKey, nFlags, bDown);
	
    switch ( Action )
    {
		case EB_HELP:
		{
			::MessageBoxEx( NULL, " Help : F1 - Help \n\n Home - Reset To Defaults \n\n W - Wireframe Toggle \n\n Space/Pause Toggle Pausing Animation\n\n +/- - Increase/Reduce Number of Particles\n\n Left Button & Mouse - Rotate Camera\n\n Shift Left Button & Mouse - Pan Camera \n\n Ctrl Left Button & Mouse - Move Camera In & Out\n\n",
				   "Help", MB_ICONINFORMATION | MB_TASKMODAL, MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ) );
		}
		break;

		case EB_WIREFRAME:
        {
            m_bWireframe = !m_bWireframe;
            m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
        }
		break;

		case EB_RESET:
        {
            m_ParticleSettings.total_particles = kStartParticles;
            m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
            m_pUI->Reset();
        }
		break;

        case EB_ADD:
            m_ParticleSettings.total_particles += max(5, m_ParticleSettings.total_particles/10);
            if (m_ParticleSettings.total_particles > kNumParticles)
                m_ParticleSettings.total_particles = kNumParticles;
            break;

        case EB_SUBTRACT:
            m_ParticleSettings.total_particles -= max(5, m_ParticleSettings.total_particles/10);
            if (m_ParticleSettings.total_particles < 10)
                m_ParticleSettings.total_particles = 10;
            break;

        default:
            m_bKey[vKey] = bDown;
    }
}