/*****************************************************************************
File:  Sketch.cpp

  Copyright (c) 2001 David E. Tin Nyo
  http://www.meme-x.com

  This Effect for the NVIDIA Effect Browser is a non-photorealistic renderer
  that simulates a crosshatched pencil sketch.

  The basic idea is to modulate a crosshatch pattern with a grayscale pattern
  that varies in relation to the light source.
  I added a little more interest by having separate vertical and horizontal
  hatch patterns that can be modulated by separate lights.

  portions Copyright (C) 1999, 2000 NVIDIA Corporation
  This file is provided without support, instruction, or implied warranty of any
  kind.  NVIDIA makes no guarantee of its fitness for a particular purpose and is
  not liable under any circumstances for any damages or loss whatsoever arising
  from the use or inability to use this file or items derived from it.

Comments:


******************************************************************************/

#include "eb_effect.h"
#include "nvinc.h"
#include "Sketch.h"
#include "ComPtr.h"
#include "Constants.h"

using namespace std;

DECLARE_EFFECT_MAIN()

extern "C"
{
	
	__declspec(dllexport) unsigned int GetNumEffects() { return 1; }
	
	__declspec(dllexport) EBEffect* CreateEffect(unsigned int EffectNum)
	{	
		switch(EffectNum)
		{
		case 0:
			return new CShaderSketch();
			break;
		default:
			return NULL;
		}
	}
	
}

// ------------------------------------------------------------------------------
// CShaderSketch Constructor
//
// Description: Nothing to it -- just straight-forward construction
// ------------------------------------------------------------------------------ 
CShaderSketch::CShaderSketch()
: m_pVertexBuffer(NULL), 
  m_pIndexBuffer(NULL),
  m_pCross1Texture(NULL), 
  m_pCross2Texture(NULL),
  m_pShade1Texture(NULL),
  m_pShade2Texture(NULL),
  m_dwCurrentVShader(0),
  m_dwCurrentPShader(0),
  m_pAttributes(NULL),
  m_pUI(NULL),
  m_eDisplayOption(SKETCHDISPLAY_LIGHTPOS0),
  m_fRadius(0.0f),
  m_vecCenter(0.0f, 0.0f, 0.0f)
{
	m_strEffectLocation = "3rd Party\\Stylistic";
	m_strEffectName = "Pencil Sketch";
	m_strEffectVertexShader = GetFilePath("Sketch.nvv");
	m_strEffectPixelShader = GetFilePath("Sketch.nvp");
	m_strEffectVersion = "1.0";
	
	D3DXMatrixIdentity(&m_world);
	D3DXMatrixIdentity(&m_view);
	D3DXMatrixIdentity(&m_proj);
}

// ------------------------------------------------------------------------------
// CShaderSketch Destructor
//
// Description: Nothing to it -- just straight-forward destruction
// ------------------------------------------------------------------------------ 
CShaderSketch::~CShaderSketch()
{
	Free(); 
}

// ------------------------------------------------------------------------------
// CShaderSketch::UpdateProperties
//
// Description: This adds options to the menu (click on "File" or right-click
// on the display.
// ------------------------------------------------------------------------------ 
void CShaderSketch::UpdateProperties()
{
	EBEffect::UpdateProperties();
	EBEnumProperty* pEnumProp = new EBEnumProperty("Display Options", OBJECT_MEMBER(m_eDisplayOption), EBTYPE_DWORD_PROP);
	
	// Create the menu
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp, "Lights Offset", (DWORD)SKETCHDISPLAY_LIGHTPOS0, EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp, "Lights Opposite", (DWORD)SKETCHDISPLAY_LIGHTPOS1, EBTYPE_DWORD_PROP));
	pEnumProp->AddEnumerant(new EBEnumValue(pEnumProp, "Lights Aligned", (DWORD)SKETCHDISPLAY_LIGHTPOS2, EBTYPE_DWORD_PROP));
	AddProperty(pEnumProp);

	// Vertex shaders
	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum, "Pencil Sketch Shader", GetFilePath("Sketch.nvv"), EBTYPE_STRING_PROP));
	// Pixel shaders
	m_pPixelShaderEnum->AddEnumerant(new EBEnumValue(m_pPixelShaderEnum, "Pencil Sketch Shader", GetFilePath("Sketch.nvp"), EBTYPE_STRING_PROP));

    string aboutText = "file://";
    aboutText += GetFilePath("..\\..\\..\\NVEffectsExplained.htm");
    aboutText += "#pencilsketch";
	SetAboutInfo( NULL, _T("Pencil Sketch"), _T(aboutText.c_str()));
}


// ------------------------------------------------------------------------------
// CShaderSketch::ConfirmDevice
//
// Description: Performs caps-bit checking to make sure the selected device 
//		supports this demo.  
// ------------------------------------------------------------------------------ 
HRESULT CShaderSketch::ConfirmDevice(D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format)
{
	if (!(pCaps->TextureCaps & D3DPTEXTURECAPS_MIPMAP))
	{
		m_strLastError = "Device does not support mipmaps!";
		return E_FAIL;
	}

	if (!(pCaps->TextureFilterCaps & D3DPTFILTERCAPS_MIPFLINEAR))
	{
		m_strLastError = "Device does not support appropriate texture filtering!";
		return E_FAIL;
	}

	if(pCaps->MaxSimultaneousTextures < 4)
	{
		m_strLastError = "Device does not support four simultaneous textures!";
		return E_FAIL;
	}

	if(D3DSHADER_VERSION_MAJOR(pCaps->PixelShaderVersion) < 1)
	{
		m_strLastError = "Device does not support pixel shaders!";
		return E_FAIL;
	}

	return S_OK;
}

// ------------------------------------------------------------------------------
// CShaderSketch::Initialize
//
// Description: Initialize render-states, vertex shader and vertex buffers
// ------------------------------------------------------------------------------ 
HRESULT CShaderSketch::Initialize(IDirect3DDevice8* pDev)
{
	HRESULT hr;
	
	// get the D3D device 
	m_pD3DDev = pDev;
	pDev->AddRef();
	
	//initialize mouse UI
	RECT rect;
	rect.left = rect.top = 0;
	D3DVIEWPORT8 viewport;
	m_pD3DDev->GetViewport(&viewport);
	rect.bottom = viewport.Height;
	rect.right  = viewport.Width;
	m_pUI = new MouseUI((const RECT)rect);
	
	D3DCAPS8 DeviceCaps;
    m_pD3DDev->GetDeviceCaps(&DeviceCaps);
	if (!(DeviceCaps.TextureFilterCaps & D3DPTFILTERCAPS_MINFANISOTROPIC) ||
		!(DeviceCaps.TextureFilterCaps & D3DPTFILTERCAPS_MAGFANISOTROPIC))
	{
		m_bAnisotropic = false;
	}
	else
	{
		m_bAnisotropic = true;
	}

	vector<DWORD> Declaration;
	Declaration.push_back(D3DVSD_STREAM(0));
	Declaration.push_back(D3DVSD_REG(0, D3DVSDT_FLOAT3));
	Declaration.push_back(D3DVSD_REG(1, D3DVSDT_FLOAT3));
	Declaration.push_back(D3DVSD_REG(2, D3DVSDT_FLOAT2));
	Declaration.push_back(D3DVSD_REG(3, D3DVSDT_FLOAT2));
	Declaration.push_back(D3DVSD_REG(4, D3DVSDT_FLOAT1));
	Declaration.push_back(D3DVSD_REG(5, D3DVSDT_FLOAT1));
	Declaration.push_back(D3DVSD_END());
	
	// Load the vertex shader object file and create it.
	m_dwCurrentVShader = 0;
	hr = LoadAndCreateShader(GetFilePath("Sketch.vso"), &Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwCurrentVShader);
	if (FAILED(hr))
		return hr;

	// select the just created vertex shader
	hr = m_pD3DDev->SetVertexShader(m_dwCurrentVShader);
	if (FAILED(hr))
		return hr;

	m_dwCurrentPShader = 0;
	hr = LoadAndCreateShader(GetFilePath("Sketch.pso"), &Declaration[0], 0, SHADERTYPE_PIXEL, &m_dwCurrentPShader);
	if (FAILED(hr))
		return hr;
	
	// select the just created pixel shader
	hr = m_pD3DDev->SetPixelShader(m_dwCurrentPShader);
	if (FAILED(hr))
		return hr;
	
	DWORD tempFVF2;
	D3DXFVFFromDeclarator(&Declaration[0], &tempFVF2);
	
	// load the mesh
	DWORD tempFVF = D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX4 | 
		D3DFVF_TEXCOORDSIZE2(0) | D3DFVF_TEXCOORDSIZE2(1) | D3DFVF_TEXCOORDSIZE1(2) | D3DFVF_TEXCOORDSIZE1(3);
	hr = LoadXFile(GetFilePath("bigship1.x").c_str(), tempFVF);
	
	if(FAILED(hr))
		return hr;
	
	// use this vertex buffer as our source for vertices
	hr = m_pD3DDev->SetStreamSource(0, m_pVertexBuffer, sizeof(SkVertex));
	if (FAILED(hr))
		return hr;
	
	//set index buffer
	m_pD3DDev->SetIndices(m_pIndexBuffer, 0);
	
	//set render states
	m_pD3DDev->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	m_pD3DDev->SetRenderState(D3DRS_LIGHTING, FALSE);
	m_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
	
	//load textures
	// two cross hatch textures
	D3DXCreateTextureFromFile(m_pD3DDev, GetFilePath("vert.bmp").c_str(), &m_pCross1Texture);
	D3DXCreateTextureFromFile(m_pD3DDev, GetFilePath("horz.bmp").c_str(), &m_pCross2Texture);
	// two shade textures (grayscale ramps)
	// it might be interesting to use different ramps for each texture
	D3DXCreateTextureFromFile(m_pD3DDev, GetFilePath("64shade.bmp").c_str(), &m_pShade1Texture);
	D3DXCreateTextureFromFile(m_pD3DDev, GetFilePath("64shade.bmp").c_str(), &m_pShade2Texture);
	
	m_pD3DDev->SetTexture(0, m_pCross1Texture);
	m_pD3DDev->SetTexture(1, m_pCross2Texture);
	m_pD3DDev->SetTexture(2, m_pShade1Texture);
	m_pD3DDev->SetTexture(3, m_pShade2Texture);
	
	m_pD3DDev->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
	m_pD3DDev->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
	m_pD3DDev->SetTextureStageState(2, D3DTSS_TEXCOORDINDEX, 2);
	m_pD3DDev->SetTextureStageState(3, D3DTSS_TEXCOORDINDEX, 3);

	// wraparound 2d textures for crosshatches
	// use anisotropic filtering to get clean pencil lines
	m_pD3DDev->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP);
	if (m_bAnisotropic)
	{
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_ANISOTROPIC);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_ANISOTROPIC);
	}
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);

	m_pD3DDev->SetTextureStageState(1, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP);
	m_pD3DDev->SetTextureStageState(1, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP);
	if (m_bAnisotropic)
	{
		m_pD3DDev->SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_ANISOTROPIC);
		m_pD3DDev->SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_ANISOTROPIC);
	}
	m_pD3DDev->SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);

	// clamped 1d textures for shading grayscale, point filtering
	m_pD3DDev->SetTextureStageState(2, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	m_pD3DDev->SetTextureStageState(2, D3DTSS_MINFILTER, D3DTEXF_POINT);
	m_pD3DDev->SetTextureStageState(2, D3DTSS_MAGFILTER, D3DTEXF_POINT);
	m_pD3DDev->SetTextureStageState(2, D3DTSS_MIPFILTER, D3DTEXF_NONE);

	m_pD3DDev->SetTextureStageState(3, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	m_pD3DDev->SetTextureStageState(3, D3DTSS_MINFILTER, D3DTEXF_POINT);
	m_pD3DDev->SetTextureStageState(3, D3DTSS_MAGFILTER, D3DTEXF_POINT);
	m_pD3DDev->SetTextureStageState(3, D3DTSS_MIPFILTER, D3DTEXF_NONE);

	// light positioning		
	SetLightPositions();

	// setup some constants for the Vertex Shader
	// first three are just useful values.
	// last value is a scaling constant so the crosshatch
	// textures appear at a good scale on my 1600x1200 display
	// this should really be scaled relative to the window size.
	D3DXVECTOR4 constNums(0.0f, 1.0f, 0.5f, 5.0f );
	m_pD3DDev->SetVertexShaderConstant(CV_CONSTANTS, &constNums, 1);
	
	//set view matrix depending on the bounding sphere
	D3DXVECTOR3 eye, lookAt, up;
	
	eye.x    = m_vecCenter.x; eye.y    = m_vecCenter.y; eye.z    = -m_fRadius*2;
	lookAt.x = m_vecCenter.x; lookAt.y = m_vecCenter.y; lookAt.z = m_vecCenter.z;
	up.x	 = 0.0f;          up.y     = 1.0f;          up.z     = 0.0f;
		
	D3DXMatrixIdentity(&m_world);
	D3DXMatrixLookAtLH(&m_view, &eye, &lookAt, &up);
	
	D3DXMatrixPerspectiveFovLH(&m_proj,
		D3DXToRadian(60.0f),
		1,
		1.0f,
		100.0f);
	m_pD3DDev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX*)&m_proj);
	
	//matrices
	SetVertexShaderMatrices();
	
	return S_OK;
}

HRESULT CShaderSketch::SetLightPositions() 
{
	D3DXVECTOR4 light1Dir;
	D3DXVECTOR4 light2Dir;

	// setup the different light configurations depending on the menu selection:
	if ( m_eDisplayOption == SKETCHDISPLAY_LIGHTPOS1 ) {
		// opposing lights
		// lights are on diagonal opposites
		light1Dir = D3DXVECTOR4(1.0f, 1.0f, -1.0f, 0.0f);
		light2Dir = D3DXVECTOR4(-1.0f, -1.0f, -1.0f, 0.0f);
	} else if (m_eDisplayOption == SKETCHDISPLAY_LIGHTPOS2) {
		// aligned lights
		// both lights are in same position
		light1Dir = D3DXVECTOR4(1.0f, 1.0f, -1.0f, 0.0f);
		light2Dir = D3DXVECTOR4(1.0f, 1.0f, -1.0f, 0.0f);
	} else {
		// offset lights
		// lights are at a small angle from each other
		light1Dir = D3DXVECTOR4(1.0f, 1.0f, -1.0f, 0.0f);
		light2Dir = D3DXVECTOR4(0.97f, 0.26f, -1.0f, 0.0f);
	}


	D3DXVec4Normalize(&light1Dir, &light1Dir);
	D3DXVec4Normalize(&light2Dir, &light2Dir);
	
	m_pD3DDev->SetVertexShaderConstant(CV_LIGHT1_DIRECTION, &light1Dir, 1);
	m_pD3DDev->SetVertexShaderConstant(CV_LIGHT2_DIRECTION, &light2Dir, 1);

	return S_OK;
}

HRESULT CShaderSketch::SetVertexShaderMatrices()
{
	D3DXMATRIX worldViewProjMat, worldITMat, worldMat;
	D3DXMATRIX tempMat;
	
	D3DXMatrixMultiply(&tempMat, &m_pUI->GetRotationMatrix(), &m_pUI->GetTranslationMatrix());
	D3DXMatrixTranspose(&worldMat, &tempMat);
	
	D3DXMatrixIdentity(&worldViewProjMat);
	
	D3DXMatrixMultiply(&worldViewProjMat, &tempMat, &m_view);
	
	D3DXMatrixInverse(&worldITMat, NULL, &tempMat);
	
	D3DXMatrixMultiply(&worldViewProjMat, &worldViewProjMat, &m_proj);
	D3DXMatrixTranspose(&worldViewProjMat, &worldViewProjMat);
	
	m_pD3DDev->SetVertexShaderConstant(CV_WORLDVIEWPROJ_0, &worldViewProjMat, 4);
	
	m_pD3DDev->SetVertexShaderConstant(CV_WORLD_IT_0, &worldITMat, 4);
	
	return S_OK;
}

// ------------------------------------------------------------------------------
// CShaderSketch::Free
//
// Description: Called when we switch away from this demo.	Free up all the 
//		memory that is in use.
// ------------------------------------------------------------------------------ 
HRESULT CShaderSketch::Free()
{
	SAFE_RELEASE(m_pVertexBuffer);
	SAFE_RELEASE(m_pIndexBuffer);
	
	SAFE_RELEASE(m_pCross1Texture);
	SAFE_RELEASE(m_pCross2Texture);
	SAFE_RELEASE(m_pShade1Texture);
	SAFE_RELEASE(m_pShade2Texture);
	
	SAFE_DELETE_ARRAY(m_pAttributes);
	
	if (m_pD3DDev)
	{
		m_pD3DDev->DeleteVertexShader(m_dwCurrentVShader);
		m_pD3DDev->DeletePixelShader(m_dwCurrentPShader);
		SAFE_RELEASE(m_pD3DDev);
	}

	SAFE_DELETE(m_pUI);
	
	return S_OK;
}


HRESULT CShaderSketch::LoadXFile(const char* fileName, const DWORD dwFVF)
{
	ID3DXMesh* tempMesh, *tempMeshFVF, *tempMeshOpt;
	ID3DXBuffer* tempMaterials;
	
	HRESULT hr = D3DXLoadMeshFromX(const_cast<char*>(fileName), D3DXMESH_VB_SYSTEMMEM, m_pD3DDev, NULL, 
		&tempMaterials, &m_dwNumSections, &tempMesh);

	//comptr to ensure this object gets released
	ComPtr< ID3DXBuffer* > tempMaterialBuffer( tempMaterials, COMPTR_DONT_ADD_REF );
	ComPtr< ID3DXMesh* > tempMesh2( tempMesh, COMPTR_DONT_ADD_REF  );

	if(FAILED(hr))
		return hr;

	//compute bounding sphere
	SkVertex* pBuff;
	hr = tempMesh->LockVertexBuffer(D3DLOCK_READONLY, (BYTE**)&pBuff);
	if(FAILED(hr))
		return hr;
	D3DXComputeBoundingSphere(pBuff, tempMesh->GetNumVertices(), tempMesh->GetFVF(), &m_vecCenter, &m_fRadius);
	tempMesh->UnlockVertexBuffer();
	
	//optimize in attribute order
	hr = tempMesh->Optimize(D3DXMESHOPT_ATTRSORT, NULL, NULL, NULL, NULL, &tempMeshOpt);
	if(FAILED(hr))
		return hr;

	//comptr to ensure this object gets released
	ComPtr< ID3DXMesh* > tempMesh3( tempMeshOpt, COMPTR_DONT_ADD_REF  );

	//get attributes
	DWORD attribSize = m_dwNumSections;
	hr = tempMeshOpt->GetAttributeTable(NULL, &attribSize);
	if(FAILED(hr))
		return hr;
	
	m_pAttributes = new D3DXATTRIBUTERANGE[attribSize];
	
	hr = tempMeshOpt->GetAttributeTable(m_pAttributes, &attribSize);
	if(FAILED(hr))
		return hr;

	//convert to our format, and also make writeonly
	hr = tempMeshOpt->CloneMeshFVF(D3DXMESH_WRITEONLY, dwFVF, m_pD3DDev, &tempMeshFVF);
	if(FAILED(hr))
		return hr;

	//comptr to ensure this object gets released
	ComPtr< ID3DXMesh* > tempMesh4( tempMeshFVF, COMPTR_DONT_ADD_REF  );
	
	tempMeshFVF->GetVertexBuffer(&m_pVertexBuffer);
	tempMeshFVF->GetIndexBuffer(&m_pIndexBuffer);
	
	return S_OK;
}


// ------------------------------------------------------------------------------
// CShaderSketch::Start
//
// Description: Called to reset
// ------------------------------------------------------------------------------ 
HRESULT CShaderSketch::Start()
{
	return S_OK;
}


// ------------------------------------------------------------------------------
// CShaderSketch::Tick
//
// Description: This function is called every frame.  Here is where we write
//		and update constants, clear the z- and back-buffer and render our 
//		primitives
// ------------------------------------------------------------------------------ 
HRESULT CShaderSketch::Tick(EBTimer* pTimer)
{
	HRESULT hr = S_OK;

	SetLightPositions();

	SetVertexShaderMatrices();
  
	m_pD3DDev->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

	// Clear window to white
	m_pD3DDev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0x00FFFFFF, 1.0f, 0L);

	for(int i = 0; i < m_dwNumSections; i++)
	{
		m_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, m_pAttributes[i].VertexStart, 
			m_pAttributes[i].VertexCount, m_pAttributes[i].FaceStart*3, m_pAttributes[i].FaceCount);
	}

	return hr;
}


void CShaderSketch::MouseButton(HWND hWnd, eButtonID button, bool bDown, int x, int y)
{
	if(button == MOUSE_LEFTBUTTON)
	{
		if(bDown)
		{
			m_pUI->OnLButtonDown(x, y);
		}
		else
		{
			m_pUI->OnLButtonUp(x, y);
		}
	}
	return;
}


void CShaderSketch::MouseMove(HWND hWnd, int x, int y)
{
	m_pUI->OnMouseMove(x, y);
	return;
}

void CShaderSketch::Keyboard(DWORD dwKey, UINT nFlags, bool bDown)
{
	if(bDown)
	{
		switch(dwKey)
		{

            case 'H' :
            case VK_F1 :
            {
		        ::MessageBoxEx( NULL, " Help : F1 - Help \n\n Home - Reset To Defaults \n\n Space\\Pause - Toggle Pause/Resume \n\n Left Button & Mouse - Rotate Object\n\n Shift Left Button & Mouse - Pan Camera \n\n Ctrl Left Button & Mouse - Move Camera In & Out\n\n",
					   "Help", MB_ICONINFORMATION | MB_TASKMODAL, MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ) );
                break;
            }

            case VK_HOME :
            case VK_END :
			case VK_NUMPAD7 :
			case VK_NUMPAD1 :
            case '7' :
            case '1' :
            {
                m_pUI->Reset();
				m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
                break;
            }

            default :
            {
                break;
            }
        }
	}
	SetVertexShaderMatrices();
	return;
}
