/******************************************************************************
  Vortex - by Jim Bumgardner

	Converts polar texture coordinates to cartesian, with vortex twisting.

	The trick here is that the texture coordinates are precomputed as polar coordinates,
	which reduces the required processing inside the vertex shader.

	The mesh has smaller polygons in the center, where more detail is needed.

  Based on code Copyright (C) 1999, 2000 NVIDIA Corporation
        
******************************************************************************/

#include "eb_effect.h"
#include "shader_Vortex.h"
#include "nvinc.h"

using namespace std;

DECLARE_EFFECT_MAIN()

extern "C"
{

__declspec(dllexport) unsigned int GetNumEffects() { return 1; }

__declspec(dllexport) EBEffect* CreateEffect(unsigned int EffectNum)
{	
	switch(EffectNum)
	{
		case 0:
			return new CShaderVortex();
			break;
		default:
			return NULL;
	}
}

}

CShaderVortex::CShaderVortex()
	: m_pVertexBuffer(NULL),
	m_pLogoTexture(NULL),
	m_dwCurrentShader(0),
	m_bWireframe(false),
	m_pUI(NULL)
{
	m_strEffectLocation = "3rd Party\\Procedural";
	m_strEffectName = "Vortex Texture Mapping";
	m_strEffectVertexShader = GetFilePath("vortex.nvv");
	m_strEffectPixelShader = "";
	m_strEffectVersion = "1.0";
}

void CShaderVortex::UpdateProperties()
{
	EBEffect::UpdateProperties();

	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum, "Polar Coords -> tu,tv", GetFilePath("vortex.nvv"), EBTYPE_STRING_PROP));
	AddProperty(new EBProperty("Wireframe", OBJECT_MEMBER(m_bWireframe), EBTYPE_BOOL_PROP));

    string aboutText = "file://";
    aboutText += GetFilePath("..\\..\\..\\..\\NVEffectsExplained.htm");
    aboutText += "#vortex";
	SetAboutInfo( NULL, _T("Vortex"), _T(aboutText.c_str()));
}

CShaderVortex::~CShaderVortex()
{
	Free();	
}


// Function to compute polar coordinates
//
static void ComputePolarCoordinates(FLOAT x, FLOAT y, D3DXVECTOR2* coords);

static void ComputePolarCoordinates(FLOAT x, FLOAT y, D3DXVECTOR2* coords)
{
	FLOAT	r = sqrtf(x*x+y*y),a;
	if (r == 0.0)
		a = 0.0f;
	else {
		if (y > 0)
			a =  (D3DX_PI*2.0f)-acosf(x/r);
		else
			a = acosf(x/r);
	}
	coords->x = r;
	coords->y = a;
}

// GAIN() is a mapping function which is used to make the grid cells near the
// center smaller (more detail is needed there)

#define log_half				logf(0.5f)
#define BIAS(x, p)			(powf( (x), logf(p) / log_half ))
#define GAIN(x, p)			(x < .5? BIAS( 2*x, 1-p)/2 : 1 - BIAS(2-2*x,1-p)/2 )

HRESULT CShaderVortex::Initialize(LPDIRECT3DDEVICE8 pDev)
{
	HRESULT hr;

	m_pD3DDev = pDev;
	pDev->AddRef();

	//initialize mouse UI
	RECT rect;
	rect.left = rect.top = 0;
	D3DVIEWPORT8 viewport;
	m_pD3DDev->GetViewport(&viewport);
	rect.bottom = viewport.Height;
	rect.right  = viewport.Width;
	m_pUI = new MouseUI((const RECT)rect);

	vector<DWORD> Declaration;
	Declaration.push_back(D3DVSD_STREAM(0));
	Declaration.push_back(D3DVSD_REG(0, D3DVSDT_FLOAT3));
	Declaration.push_back(D3DVSD_REG(1, D3DVSDT_FLOAT3));
	Declaration.push_back(D3DVSD_REG(2, D3DVSDT_FLOAT2));
	Declaration.push_back(D3DVSD_END());

	hr = LoadAndCreateShader(GetFilePath("vortex.vso"), &Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwCurrentShader);
	if (FAILED(hr))
		return hr;

	hr = m_pD3DDev->SetVertexShader(m_dwCurrentShader);
	if (FAILED(hr))
		return hr;

	hr = m_pD3DDev->CreateVertexBuffer( kMaxVertices * sizeof(SpyroVertex), D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &m_pVertexBuffer);
	if (FAILED(hr))
		return hr;
		
  // initialize all vertices with the proper position, normal, and uv coordinates
	//
	SpyroVertex* pBuff;
	if (m_pVertexBuffer)
	{
		hr = m_pVertexBuffer->Lock(0, sizeof(SpyroVertex) * kMaxVertices,(BYTE**)&pBuff, 0);
		if (FAILED(hr))
		{
			m_strLastError = "Couldn't lock buffer!";
			return hr;
		}

		int	x,y;
		FLOAT	fx,fy,fy2;

		SpyroVertex* pb = pBuff;

// This computes a grid with more detail in the center
#define CENTER_GAIN	0.36f

		for (y = 0; y < kHeightInTiles; ++y)
		{
			// ry = (FLOAT) y / (FLOAT) kHeightInTiles;
			fy = -1.0 + 2*GAIN((FLOAT) y / (FLOAT) kHeightInTiles,CENTER_GAIN);
			// ry = (FLOAT) (y+1)/(FLOAT) kHeightInTiles;
			fy2 = -1.0 + 2*GAIN((FLOAT) (y+1)/(FLOAT) kHeightInTiles,CENTER_GAIN);

			for (x = 0; x <= kHeightInTiles; ++x)
			{
				// rx =  x / (FLOAT) kWidthInTiles;
				fx = -1.0 + 2*GAIN(x / (FLOAT) kWidthInTiles,CENTER_GAIN);
				
				pb->pos = D3DXVECTOR3(fx,fy2,0);
				pb->normal= D3DXVECTOR3(0.0f,0.0f,-1.0f);
				ComputePolarCoordinates(fx * .5,fy2 * .5,&pb->texture);
				++pb;

				pb->pos = D3DXVECTOR3(fx,fy,0);
				pb->normal= D3DXVECTOR3(0.0f,0.0f,-1.0f);
				ComputePolarCoordinates(fx * .5,fy * .5,&pb->texture);
				++pb;

			}
			
		}

		m_pVertexBuffer->Unlock();
	}

	hr = m_pD3DDev->SetStreamSource(0, m_pVertexBuffer, sizeof(SpyroVertex));
	if (FAILED(hr))
		return hr;


	// Create the Logo texture
	hr = D3DXCreateTextureFromFileEx(m_pD3DDev, 
		GetFilePath("nvlogo_only.dds").c_str(),
		D3DX_DEFAULT,
		D3DX_DEFAULT,
		0,
		0,
		D3DFMT_UNKNOWN,
		D3DPOOL_MANAGED,
		D3DX_FILTER_LINEAR,
		D3DX_FILTER_LINEAR,
		0,
		NULL,
		NULL,
		&m_pLogoTexture);

	if (FAILED(hr))
	{
		m_strLastError = "Could not create nvlogo_only.dds";
		return hr;
	}

	m_pD3DDev->SetTexture(0, m_pLogoTexture);

	m_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

	m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_ADD);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);

	m_pD3DDev->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
	m_pD3DDev->SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

	m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
  m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);

  // Setup Shader Constants
  D3DXVECTOR4 vZero( 0.0f, 0.0f, 0.0f, 0.0f );
  D3DXVECTOR4 vOne( 1.0f, 0.5f, 0.2f, 0.05f );
	D3DXVECTOR4 vPI(D3DX_PI, 1.f/2.f, 2.f*D3DX_PI, 1.f/(2.f*D3DX_PI) );
	D3DXVECTOR4 vCosStuff( -1.0f, 1.f/2.f, -1.f/24.f, 1.f/720.f );
	D3DXVECTOR4 vSinStuff( 1.0f, -1.f/6.f, 1.f/120.f, -1.f/5040.f );

  m_pD3DDev->SetVertexShaderConstant(  20, &vZero,     1 );
  m_pD3DDev->SetVertexShaderConstant(  21, &vOne,      1 );
  m_pD3DDev->SetVertexShaderConstant(  22, &vPI,       1 );
  m_pD3DDev->SetVertexShaderConstant(  30, &vCosStuff, 1 );
  m_pD3DDev->SetVertexShaderConstant(  31, &vSinStuff, 1 );
	
	return S_OK;

}

HRESULT CShaderVortex::Free()
{
	SAFE_RELEASE(m_pLogoTexture);
	SAFE_RELEASE(m_pVertexBuffer);
	SAFE_DELETE(m_pUI);

	if (m_pD3DDev)
	{
		if (m_dwCurrentShader)
			m_pD3DDev->DeleteVertexShader(m_dwCurrentShader);
		SAFE_RELEASE(m_pD3DDev);
	}
	
	return S_OK;
}

HRESULT CShaderVortex::Start()
{
	return S_OK;
}

HRESULT CShaderVortex::Tick(EBTimer* pTimer)
{
	HRESULT hr = S_OK;
	D3DXMATRIX matWorld;
  D3DXMATRIX matView;
	D3DXMATRIX matProj;
	D3DXMATRIX matTemp;
	D3DXMATRIX matWorldViewProj;
	D3DXMATRIX matWorldView;
	D3DXMATRIX matWorldI;

  // Setup Vortex Constants Here
	// x = amount of twist, y = overall rotation
	FLOAT fTime = pTimer->GetDuration();
	const FLOAT fTwistMagnitude = 4.0f;
  D3DXVECTOR4 vRotateF( fTwistMagnitude*sinf(fTime), fTime, 0.0f, 0.0f );
  m_pD3DDev->SetVertexShaderConstant(  24, &vRotateF, 1 );
	
	hr = m_pD3DDev->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB( 0xAA, 0xAA, 0xAA ), 1.0, 0);

	m_pD3DDev->SetRenderState(D3DRS_FILLMODE, m_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID);

	D3DXVECTOR3 vEyePt = D3DXVECTOR3( 0.0f, 0.0f, -5.0f );
	D3DXVECTOR3 vLookatPt = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	D3DXVECTOR3 vUp = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );

	// Setup the world, view and projection matrices
	D3DXMatrixMultiply( &matWorld, &m_pUI->GetRotationMatrix(), &m_pUI->GetTranslationMatrix() );

	D3DXMatrixLookAtLH(&matView, &vEyePt, &vLookatPt, &vUp);
	D3DXMatrixPerspectiveFovLH(&matProj, D3DXToRadian(60.0f), 1.0f, 1.0f, 50.0f);
		
	// Setup the composite matrix to get vertices to projection space
	D3DXMatrixMultiply(&matTemp, &matWorld, &matView);
	D3DXMatrixMultiply(&matWorldViewProj, &matTemp, &matProj);
	D3DXMatrixMultiply(&matWorldView, &matWorld, &matView);
	
	// Create a 3x3 inverse of the worldview for the light transformation back to model space
	D3DXMatrixInverse(&matWorldI, NULL, &matWorld);
	D3DXMatrixTranspose(&matWorldViewProj, &matWorldViewProj);

	// Setup constants
	m_pD3DDev->SetVertexShaderConstant(0, &matWorldViewProj(0, 0), 4);
	
	// Transform from world space to model space and point it away from the vertex towards the light
	// Light is pointing into the world
	D3DXVECTOR3 vecLight(0.0f, 0.0f, 1.0f);
	D3DXVec3TransformNormal(&vecLight, &vecLight, &matWorldI);
	D3DXVec3Normalize((D3DXVECTOR3*)&vecLight, (D3DXVECTOR3*)&vecLight);
	vecLight = -vecLight;
	
	m_pD3DDev->SetVertexShaderConstant(4, &vecLight.x, 1);

	// The material color
	D3DXVECTOR4 matColor(1.0f, 1.0f, 1.0f, 0.0f);
	m_pD3DDev->SetVertexShaderConstant(5, &matColor, 1);

	for (int y = 0;y < kHeightInTiles; ++y)
    m_pD3DDev->DrawPrimitive( D3DPT_TRIANGLESTRIP, kVerticesPerRow*y, kWidthInTiles << 1 );

	return hr;
}

HRESULT CShaderVortex::ConfirmDevice(D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format)
{
	return S_OK;
}

void CShaderVortex::MouseButton(HWND hWnd, eButtonID button, bool bDown, int x, int y)
{
	if(button == MOUSE_LEFTBUTTON)
	{
		if(bDown)
		{
			m_pUI->OnLButtonDown(x, y);
		}
		else
		{
			m_pUI->OnLButtonUp(x, y);
		}
	}
	return;
}

void CShaderVortex::MouseMove(HWND hWnd, int x, int y)
{
	m_pUI->OnMouseMove(x, y);
	return;
}

void CShaderVortex::Keyboard(DWORD dwKey, UINT nFlags, bool bDown)
{
	eEBKeyAction Action = TranslateEffectKey(dwKey, nFlags, bDown);
	
    switch ( Action )
    {
		case EB_HELP:
		{
			::MessageBoxEx( NULL, " Help : F1 - Help \n\n Home - Reset To Defaults \n\n W - Wireframe Toggle \n\n Left Button & Mouse - Rotate Object\n\n Shift Left Button & Mouse - Pan Camera \n\n Ctrl Left Button & Mouse - Move Camera In & Out\n\n",
				   "Help", MB_ICONINFORMATION | MB_TASKMODAL, MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ) );
		}
		break;

		case EB_WIREFRAME:
        {
            m_bWireframe = !m_bWireframe;
			m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
            
        }
		break;

		case EB_RESET:
        {
            m_pUI->Reset();
            m_bWireframe = false;
			m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
            
        }
		break;

        default :
            break;
    }
}
