/*----------------------------------------------------------------------
 * Project:		Fur/Grass
 * File:		shader_Fur.cpp
 * Description:	Fur/Grass VS Renderer
 * Created by:	Sebastien St-Laurent
 *				on 3/18/01
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Overview:
 *	This is a simple implementation of grass/fur rendering using a simple
 *  vertex shader to automate the task. The concept used to render the 
 *  grass/fur is based on the grass sample from nVidia and Huges Hoppe's 
 *  research on realtime fur rendering. Of course, this renderer is 
 *  simplified to acheive better performance. The principle used is fairly
 *  simple, we first render the basic texture for the mesh then we render
 *  successive layers (hulls) with different alpha values corresponding to
 *  the height of the fur/grass for that hull (note that each hull is rendered
 *  slightly above the original mesh to give a 3D effect). The vertex shader 
 *  is used to render the hulls at different elevations (without having to
 *  manipulate the original mesh), compute the proper alpha values for each
 *  layer and perform a simple lighting pass of the mesh.
 *
 *  Note : This effect used the "shader_aniso" as a template so it's structure
 *         is similar in many ways...
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *  NAME                   DATE      DESCRIPTION
 *  Sebastien St-Laurent   3/18/01   Created
 *  ins
 *----------------------------------------------------------------------*/

#include "eb_effect.h"
#include "nvinc.h"
#include "shader_fur.h"
using namespace std;

DECLARE_EFFECT_MAIN()

extern "C"
{
	
	__declspec(dllexport) unsigned int GetNumEffects() { return 2; }
	
	__declspec(dllexport) EBEffect* CreateEffect(unsigned int EffectNum)
	{	
		switch(EffectNum)
		{
		case 0:
			return new CShaderFuzz(0);
			break;
		case 1:
			return new CShaderFuzz(1);
			break;
		default:
			return NULL;
		}
	}
	
};

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::CShaderFuzz
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Constructor for the effect
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
CShaderFuzz::CShaderFuzz(int CurrentModel)
: m_bWireframe(false),
  m_pUI(NULL),
  m_bUseTrilinear(true)
{
	for(int i=0;i<NUM_MODELS;i++)
	{
		m_pVertexBuffer[i]=NULL;
		m_pIndexBuffer[i]=NULL;
		m_pShadeTexture[i]=NULL;
		m_pAttributes[i]=NULL;
	}

	// Setup the effect for the browser
	m_strEffectLocation = "3rd Party\\Hair";
	switch(CurrentModel)
	{
		case 0:
			m_strEffectName = "Fur";
			break;
		case 1:
			m_strEffectName = "Grass";
			break;
		default:
			assert(0);
			break;
	}
	m_strEffectVertexShader = GetFilePath("Fuzz.nvv");
	m_strEffectPixelShader = "";
	m_strEffectVersion = "1.0";

	// Init our matrices
	D3DXMatrixIdentity(&m_world);
	D3DXMatrixIdentity(&m_view);
	D3DXMatrixIdentity(&m_proj);

	m_strFileName[0] = "tiger.x";
	m_strFileName[1] = "envcube.x";
	m_strTextureName[0] = "fuzz3.dds";
	m_strTextureName[1] = "fuzz2.tga";
	m_fOffset[0] = 0.005f;
	m_fOffset[1] = -0.005f;
	m_nLayers = 8;
	m_nCurModel = CurrentModel;

}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::~CShaderFuzz
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Destructor for the effect
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
CShaderFuzz::~CShaderFuzz()
{
	Free(); 
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::UpdateProperties
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		This adds options to the menu (click on "File" or right-click
 *				on the display.
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
void CShaderFuzz::UpdateProperties()
{
	EBEffect::UpdateProperties();

	AddProperty(new EBProperty("Wireframe", OBJECT_MEMBER(m_bWireframe), EBTYPE_BOOL_PROP));
	AddProperty(new EBTriggerProperty("More Layers"));
	AddProperty(new EBTriggerProperty("Less Layers"));

	// Vertex shaders
	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum, "Fuzz/Fur Shading", GetFilePath("fuzz.nvv"), EBTYPE_STRING_PROP));

	
    string aboutText = "file://";
    aboutText += GetFilePath("..\\..\\..\\NVEffectsExplained.htm");
	if(m_nCurModel == 0)
	{
		aboutText += "#fur";
		SetAboutInfo( NULL, _T("Fur"), _T(aboutText.c_str()));
	}
	else
	{
		aboutText += "#grass";
		SetAboutInfo( NULL, _T("Grass"), _T(aboutText.c_str()));
	}
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::ConfirmDevice
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Performs caps-bit checking to make sure the selected device 
 *				supports this demo.  In this particular case we check for nothing!
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::ConfirmDevice(D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format)
{
	if (!(pCaps->TextureCaps & D3DPTEXTURECAPS_MIPMAP))
	{
		m_strLastError = "Device does not support mipmaps!";
		return E_FAIL;
	}

	if (!(pCaps->TextureCaps & D3DPTFILTERCAPS_MIPFLINEAR))
	{
		m_bUseTrilinear = false;
	}

	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::Initialize
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Initialize render-states, vertex shader and vertex buffers
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::Initialize(IDirect3DDevice8* pDev)
{
	HRESULT hr;
	
	// get the D3D device 
	m_pD3DDev = pDev;
	pDev->AddRef();
	
	//initialize mouse UI
	RECT rect;
	rect.left = rect.top = 0;
	D3DVIEWPORT8 viewport;
	m_pD3DDev->GetViewport(&viewport);
	rect.bottom = viewport.Height;
	rect.right  = viewport.Width;
	m_pUI = new MouseUI((const RECT)rect);
	
	vector<DWORD> Declaration;
	Declaration.push_back(D3DVSD_STREAM(0));
	Declaration.push_back(D3DVSD_REG(D3DVSDE_POSITION, D3DVSDT_FLOAT3)); // position
	Declaration.push_back(D3DVSD_REG(D3DVSDE_NORMAL, D3DVSDT_FLOAT3)); // normal
	Declaration.push_back(D3DVSD_REG(D3DVSDE_TEXCOORD0, D3DVSDT_FLOAT2)); // texture coords 0
	Declaration.push_back(D3DVSD_END());
	
	// Load the vertex shader object file and create it.
	m_dwCurrentShader = 0;
	hr = LoadAndCreateShader(GetFilePath("fuzz.vso"), &Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwCurrentShader);
	if (FAILED(hr))
		return hr;
	
	// select the just created vertex shader as the thing to transform and light vertices
	hr = m_pD3DDev->SetVertexShader(m_dwCurrentShader);
	if (FAILED(hr))
		return hr;
	
	for (int i=0;i<NUM_MODELS;i++)
	{
		hr = LoadXFile(i,GetFilePath( m_strFileName[i] ).c_str(), FuzzVertex::FVF_Flags );
		if(FAILED(hr))
			return hr;
	}
	
	//set render states
	m_pD3DDev->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	m_pD3DDev->SetRenderState(D3DRS_LIGHTING, FALSE);
	m_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	m_pD3DDev->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
	m_pD3DDev->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	
	//load textures
	for (i=0;i<NUM_MODELS;i++)
	{
		D3DXCreateTextureFromFile(m_pD3DDev, GetFilePath(m_strTextureName[i]).c_str(), &m_pShadeTexture[i]);
	}
	
	
	m_pD3DDev->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_MIRROR);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
	
	if(m_bUseTrilinear)
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
		

	// write the constant constants to constant memory now
	m_pD3DDev->SetVertexShaderConstant(9, D3DXVECTOR4(0.0f, 0.5f, 1.0f, -1.0f), 1);
	D3DXVECTOR4 lightDir(0.0f, 0.0f, -1.0f, 0.0f);
	D3DXVec4Normalize(&lightDir, &lightDir);
	
	m_pD3DDev->SetVertexShaderConstant(4, &lightDir, 1);
	
	D3DXVECTOR4 constNums(0.0f, 0.5f, 1.0f, -1.0f);
	m_pD3DDev->SetVertexShaderConstant(9, &constNums, 1);
	
	D3DXMatrixIdentity(&m_world);
	
	D3DXMatrixPerspectiveFovLH(&m_proj,
		D3DXToRadian(60.0f),
		1,
		1.0f,
		50.0f);
	m_pD3DDev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX*)&m_proj);
	
	//matrices
	SetVertexShaderMatrices();
	
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::SetVertexShaderMatrices
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Sets up the matrices used by the vertex shader
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::SetVertexShaderMatrices()
{
	D3DXMATRIX mvpMat, worldITMat, worldMat;
	
	//D3DXMatrixTranspose(&worldMat, &m_world);
	D3DXMATRIX tempMat;
	D3DXMatrixMultiply( &tempMat, &m_pUI->GetRotationMatrix(), &m_pUI->GetTranslationMatrix() );

	D3DXMatrixTranspose(&worldMat, &tempMat);

	D3DXMatrixIdentity(&mvpMat);
	
	D3DXMatrixMultiply(&mvpMat, &tempMat, &m_view);
	
	D3DXMatrixInverse(&worldITMat, NULL, &tempMat );
	
	D3DXMatrixMultiply(&mvpMat, &mvpMat, &m_proj);
	D3DXMatrixTranspose(&mvpMat, &mvpMat);
	
	m_pD3DDev->SetVertexShaderConstant(0, &mvpMat, 4);
	
	m_pD3DDev->SetVertexShaderConstant(5, &worldITMat, 4);
	m_pD3DDev->SetVertexShaderConstant(11, &worldMat, 4);
	
	return S_OK;
}


/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::PropertyUpdateCallback
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Called when some property is changed from the UI
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
void CShaderFuzz::PropertyUpdateCallback(const EBProperty* pProperty, bool bWritten)
{
	if (!bWritten)
		return;

	EBString name;
	if (pProperty->IsKindOf(EBTYPE_TRIGGER_PROP))
	{

		name = pProperty->GetPropertyName();
		
		if(name == EBString("Next Mesh"))
		{
			Keyboard('N', 0, true);
		}
		else if (name == EBString("Previous Mesh"))
		{
			Keyboard('P', 0, true);
		}
		else if (name == EBString("More Layers"))
		{
			Keyboard('M', 0, true);
		}
		else if (name == EBString("Less Layers"))
		{
			Keyboard('L', 0, true);
		}
	}
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::Free
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Called when we switch away from this demo.	Free up all the 
 *				memory that is in use.
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::Free()
{
	for (int i=0;i<NUM_MODELS;i++)
	{
		SAFE_RELEASE(m_pVertexBuffer[i]);
		SAFE_RELEASE(m_pIndexBuffer[i]);
		SAFE_RELEASE(m_pShadeTexture[i]);
		SAFE_DELETE_ARRAY(m_pAttributes[i]);
	}
	
	if (m_pD3DDev)
	{
		m_pD3DDev->DeleteVertexShader(m_dwCurrentShader);
		SAFE_RELEASE(m_pD3DDev);
	}
	delete m_pUI, m_pUI = 0;
	
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::LoadXFile
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Loads a mesh from a .X file
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::LoadXFile(int idx,const char* fileName, const DWORD dwFVF)
{
	ID3DXMesh *tempMesh, *tempMeshFVF, *tempMeshOpt;
		
	HRESULT hr = D3DXLoadMeshFromX(const_cast<char*>(fileName), D3DXMESH_SYSTEMMEM , m_pD3DDev, NULL, 
		NULL, &m_dwNumSections[idx], &tempMesh);
	
	//convert to our format
	hr = tempMesh->CloneMeshFVF(D3DXMESH_DYNAMIC, dwFVF, m_pD3DDev, &tempMeshFVF);

	D3DXComputeNormals(tempMeshFVF);

	tempMeshFVF->Optimize(D3DXMESHOPT_ATTRSORT, NULL, NULL, NULL, NULL, &tempMeshOpt);
	DWORD attribSize = m_dwNumSections[idx];
	hr = tempMeshOpt->GetAttributeTable(NULL, &attribSize);

	SAFE_DELETE_ARRAY( m_pAttributes[idx] );

	m_pAttributes[idx] = new D3DXATTRIBUTERANGE[attribSize];
	
	hr = tempMeshOpt->GetAttributeTable(m_pAttributes[idx], &attribSize);


	SAFE_RELEASE( m_pVertexBuffer[idx] );
	SAFE_RELEASE( m_pIndexBuffer[idx] );

	tempMeshOpt->GetVertexBuffer(&m_pVertexBuffer[idx]);
	tempMeshOpt->GetIndexBuffer(&m_pIndexBuffer[idx]);

	SAFE_RELEASE(tempMesh);
	SAFE_RELEASE(tempMeshFVF);
	SAFE_RELEASE(tempMeshOpt);
		
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::Start
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Called when the effect is inovked
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::Start()
{
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::SetTextureStageStates
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Sets the texture renderstates required to render our mesh
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::SetTextureStageStates(bool Fuzz)
{
	if (Fuzz==false)
	{
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
		m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		m_pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);

		m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
	}
	else
	{
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
		m_pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
		m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);

		m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
	}

	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderFuzz::Tick
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		This function is called every frame.  Here is where we write
 *				and update constants, clear the z- and back-buffer and render our 
 *				primitives
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderFuzz::Tick(EBTimer* pTimer)
{
	HRESULT hr = S_OK;
	int i;

	m_lookAt = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	m_up = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	if (m_nCurModel == 0)
	{
		m_eye = D3DXVECTOR4(0.0f, 0.0f, -3.5f, 1.0f);
	}
	else
	{
		m_eye = D3DXVECTOR4(0.0f, 0.0f, -12.0f, 1.0f);
	}

	m_pD3DDev->SetVertexShaderConstant(10, &m_eye, 1);
	D3DXMatrixLookAtLH(&m_view, (D3DXVECTOR3*)&m_eye, &m_lookAt, &m_up);

	// Set V and IDX buffers
	m_pD3DDev->SetIndices( m_pIndexBuffer[m_nCurModel], 0);
	hr = m_pD3DDev->SetStreamSource(0, m_pVertexBuffer[m_nCurModel], sizeof(FuzzVertex));
	if (FAILED(hr))
		return hr;

	SetVertexShaderMatrices();

	// Set texture to use
	m_pD3DDev->SetTexture(0, m_pShadeTexture[m_nCurModel]);
	SetTextureStageStates(false);

	// Set constant so we have no offset on base mesh
	m_pD3DDev->SetVertexShaderConstant(15, D3DXVECTOR4(0.0f, 0.0f, 0.0f, 0.0f), 1);
	
	m_pD3DDev->SetRenderState(D3DRS_FILLMODE, (m_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID));
	m_pD3DDev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0x00BFBFBF, 1.0f, 0L);

	// Draw base mesh
	for(i = 0; i < m_dwNumSections[m_nCurModel]; i++)
	{
		m_pD3DDev->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 
										 m_pAttributes[m_nCurModel][i].VertexStart,
										 m_pAttributes[m_nCurModel][i].VertexCount, 
											m_pAttributes[m_nCurModel][i].FaceStart*3, 
											m_pAttributes[m_nCurModel][i].FaceCount );
	}

	// Draw fur/grass hulls
	SetTextureStageStates(true);
	for (int j=0;j<m_nLayers;j++)
	{
		//													vert_off						alpha
		m_pD3DDev->SetVertexShaderConstant(15, D3DXVECTOR4(m_fOffset[m_nCurModel]*(j+1), 0.9f-((0.8f/m_nLayers)*j), 0.0f, 0.0f), 1);
		for(i = 0; i < m_dwNumSections[m_nCurModel]; i++)
		{
			m_pD3DDev->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 
											 m_pAttributes[m_nCurModel][i].VertexStart,
											 m_pAttributes[m_nCurModel][i].VertexCount, 
												m_pAttributes[m_nCurModel][i].FaceStart*3, 
												m_pAttributes[m_nCurModel][i].FaceCount );
		}
	}

	return hr;
}


void CShaderFuzz::MouseButton(HWND hWnd, eButtonID button, bool bDown, int x, int y)
{
	if(button == MOUSE_LEFTBUTTON)
	{
		if(bDown)
		{
			m_pUI->OnLButtonDown(x, y);
		}
		else
		{
			m_pUI->OnLButtonUp(x, y);
		}
	}
	return;
}

void CShaderFuzz::MouseMove(HWND hWnd, int x, int y)
{
	m_pUI->OnMouseMove(x, y);
	return;
}

void CShaderFuzz::Keyboard(DWORD dwKey, UINT nFlags, bool bDown)
{
    if ( bDown )
    {
        switch ( dwKey )
        {

            case 'W' :
            {
                m_bWireframe = !m_bWireframe;
				m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
                break;
            }

            case 'R' :
            {
                m_bWireframe = false;
				m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
                break;
            }

			case 'M' :
            {
				m_nLayers++;
                break;
            }
            case 'L' :
            {
				if (m_nLayers>0) m_nLayers--;
                break;
            }

            case VK_HOME :
            case VK_END :
			case VK_NUMPAD7 :
			case VK_NUMPAD1 :
            case '7' :
            case '1' :
            {
                m_pUI->Reset();
                m_bWireframe = false;
				m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
                break;
            }

            case VK_LEFT :
			case VK_NUMPAD4 :
            case '4' :
            {
                m_pUI->Translate( -0.10f, 0.0f, 0.0f );
                break;
            }

            case VK_RIGHT :
			case VK_NUMPAD6 :
            case '6' :
            {
                m_pUI->Translate(  0.10f, 0.0f, 0.0f );
                break;
            }

            case VK_UP :
			case VK_NUMPAD8 :
            case '8' :
            {
                m_pUI->Translate(  0.0f, 0.0f, -0.1f );
                break;
            }

            case VK_DOWN :
			case VK_NUMPAD2 :
            case '2' :
            {
                m_pUI->Translate( 0.0f, 0.0f, 0.1f );
                break;
            }

            case VK_PRIOR :
			case VK_NUMPAD9 :
            case '9' :
            {
                m_pUI->Translate(  0.0f, 0.1f, 0.0f );
                break;
            }
            case VK_NEXT :
			case VK_NUMPAD3 :
            case '3' :
            {
                m_pUI->Translate(  0.0f, -0.1f, 0.0f );
                break;
            }

            case VK_ADD :
            {
                m_pUI->Translate( 0.0f, 0.0f, 1.0f );
                break;
            }
            case VK_SUBTRACT :
            {
                m_pUI->Translate( 0.0f, 0.0f, -1.0f );
                break;
            }

            default :
            {
                break;
            }
        };
    }
	return;
}