// -------------------------------------------------------------
// Hair shader effect
// -------------------------------------------------------------

#include <d3d8.h>
#include <d3dx8.h>
#include <dxerr8.h>
#include <iomanip>

#include "eb_file.h"
#include "Hair.h"
#include "Splines.h"
#include "ShaderConstants.h"

#include <functional>
#include <algorithm>

using namespace std;

namespace { // local data
	DWORD shaderData0[] = { // kUseVertexData
		D3DVSD_STREAM(0), 
		D3DVSD_REG(0, D3DVSDT_FLOAT3), // pos
		D3DVSD_REG(1, D3DVSDT_FLOAT3), // ts
		D3DVSD_REG(2, D3DVSDT_FLOAT3), // ty
		D3DVSD_REG(3, D3DVSDT_FLOAT2), // texture
		D3DVSD_REG(4, D3DVSDT_FLOAT4), // parameters
		D3DVSD_END()
	};
	DWORD shaderData1[] = { // kUseHairData
		D3DVSD_STREAM(0), 
		D3DVSD_REG(0, D3DVSDT_FLOAT2), // s, t
		D3DVSD_REG(4, D3DVSDT_FLOAT4), // parameters
		D3DVSD_END()
	};

	struct DepthSorter: public std::binary_function<const HairPatch::SortData&, const HairPatch::SortData&, bool> { 
		bool operator()(const HairPatch::SortData &lhs, const HairPatch::SortData &rhs) const { return lhs.depth > rhs.depth; }
	};
}

HairPatch::HairPatch(IDirect3DDevice8 *pD3DDev, ParameterSurface *pSurface, ShaderType shaderType): pD3DDev(pD3DDev), pSurface(pSurface), shaderType(shaderType), parameterBegin(0, 0), parameterEnd(1, 1), lightDirection(1, 1, 1) 
{
	pD3DDev->AddRef();
	pVertexBuffer = NULL;
	nHairs = 0; nSegments = 0;
	hairShader = 0;
	length = 1; radius = 0.1;
	useLast = false;
}
HairPatch::~HairPatch() {
	free();
}

void HairPatch::free() {
	SAFE_RELEASE(pVertexBuffer);
	if (pD3DDev)
	{
		if (hairShader)
		{
			pD3DDev->DeleteVertexShader(hairShader);
		}
		SAFE_RELEASE(pD3DDev);
	}
}

void HairPatch::create(int nHairs, int nSegments) {
	HairPatch::nHairs = nHairs;
	HairPatch::nSegments = nSegments;
	sortList.clear();
	srand(0xabba);


	DWORD *shaderData = shaderType == kUseVertexData ? shaderData0 : shaderData1; 
		
	
	_ExecuteDX(effect_api::LoadAndCreateShader(pD3DDev, effect_api::GetFilePath("Hair.vso"), shaderData, 0, SHADERTYPE_VERTEX, &hairShader));

	HairVertex *pVertex; { // get a pointer to the vertex data
		int vertexLength = nHairs*(nSegments*2 + 1) * sizeof(HairVertex);
		_ExecuteDX(pD3DDev->CreateVertexBuffer(vertexLength, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &pVertexBuffer));	
		_ExecuteDX(pVertexBuffer->Lock(0, vertexLength,(BYTE**)&pVertex, 0));
	}


	int iHair = 0; 
	while ( iHair < nHairs ) {
		float s = random(parameterBegin.x, parameterEnd.x);
		float t = random(parameterBegin.y, parameterEnd.y);

		if ( nHairs == 1 ) {
			s = 0.2;
			t = 0.0001;
		}

//		float rad = cos(t*D3DX_PI - D3DX_PI/2);
//		if ( random(-1, 1)*rad < 0 ) continue;

		D3DXVECTOR3 p = pSurface->pos(s, t);
		D3DXVECTOR3 ts = pSurface->ds(s, t);
		D3DXVECTOR3 tt = pSurface->dt(s, t);

		{ // save the hair root position for later depth sort
			SortData data;
			data.pos = p; // hair root position
			data.vertexIndex = iHair*(nSegments*2 + 1);  // the start index of this hair
			sortList.push_back(data);
		}

		D3DXVec3Normalize(&ts, &ts);
		D3DXVec3Normalize(&tt, &tt);

		D3DXVECTOR2 tex(s, t);
		for ( int iSeg = 0; iSeg <= nSegments; ++iSeg ) {
			float h = length/nSegments*iSeg;
			float r = radius/nSegments*(nSegments-iSeg);

			D3DXVECTOR4 parm(h/length, r, length, 0); // .x = u parameter value, .y = radius, .z = length
			*pVertex++ = HairVertex(p, ts, tt, tex, parm);
			if ( iSeg != nSegments ) { // last segment only has one triangle
				parm.y = -r;
				*pVertex++ = HairVertex(p, ts, tt, tex, parm);
			}
		}
		++iHair;
	}			
	
	pVertexBuffer->Unlock();


	// shader parameters
	pD3DDev->SetVertexShaderConstant(_Spline, polynomial[1].x, 3);
	pD3DDev->SetVertexShaderConstant(_Diffuse, &color, 1);
	pD3DDev->SetVertexShaderConstant(_Specular, &D3DXVECTOR4(1, 1, 1, 1), 1);
	pD3DDev->SetVertexShaderConstant(_BendParameters, &D3DXVECTOR4(5, 2, 0, 0), 1);
	pD3DDev->SetVertexShaderConstant(_Constants, &D3DXVECTOR4(0, 1, 2, 3), 1);
}


void HairPatch::setTransform(const D3DXMATRIX &matWorld, const D3DXMATRIX &matView, const D3DXMATRIX &matProjection) {
	D3DXMATRIX matWorldView;
	D3DXMATRIX matWorldViewProj;

	// Setup the composite matrix to get vertices to projection space
	D3DXMatrixMultiply(&matWorldView, &matWorld, &matView);
	D3DXMatrixMultiply(&matWorldViewProj, &matWorldView, &matProjection);
	
	{ // sort by screen z depth, from back to front
		D3DXVECTOR3 zrow(matWorldView(0, 2), matWorldView(1, 2), matWorldView(2, 2));
		for ( std::vector<SortData>::iterator i = sortList.begin(); i != sortList.end(); ++i ) 
			i->depth = D3DXVec3Dot(&zrow, &i->pos);

		std::sort(sortList.begin(), sortList.end(), DepthSorter());
	}
	
	// set matrix constants
	D3DXMatrixTranspose(&matWorldView, &matWorldView);
	_ExecuteDX(pD3DDev->SetVertexShaderConstant(_WorldMatrix, &matWorldView(0, 0), 4));

	D3DXMatrixTranspose(&matWorldViewProj, &matWorldViewProj);
	_ExecuteDX(pD3DDev->SetVertexShaderConstant(_ProjectionMatrix, &matWorldViewProj(0, 0), 4));

	D3DXMATRIX matWorldInv;
	{ // lighting calculations
		D3DXMATRIX matViewInv;
		D3DXMatrixInverse(&matWorldInv, NULL, &matWorld);
		D3DXMatrixInverse(&matViewInv, NULL, &matView);

		D3DXVECTOR3 L; { // vector pointing towards the light, in model space
			D3DXVec3TransformNormal(&L, &lightDirection, &matWorldInv);
			D3DXVec3Normalize(&L, &L);
			L = -L;
			_ExecuteDX(pD3DDev->SetVertexShaderConstant(_LightVector, &L, 1));
		}
	}

	{ // write shader parameters
		D3DXVECTOR4 shaderParameters(0, 0, 0, 0);
		shaderParameters.x = sqrt(sq(matWorldView(0, 0)) + sq(matWorldView(1, 0)) + sq(matWorldView(2, 0))) * matProjection(0, 0);
		shaderParameters.y = 1; // ks
		shaderParameters.z = 100; // kn/2
		shaderParameters.w = -0.5; // transparency fade
		_ExecuteDX(pD3DDev->SetVertexShaderConstant(_ShaderParameters, &shaderParameters, 1));
	}
	
	{ // dynamic bending
		// mDiff = mLast*mNowInv - mIdent
		// bending vactor = mDiff*modelPosition
		D3DXMATRIX matDiff;
		if ( useLast ) {
			D3DXMatrixMultiply(&matDiff, &matWorldLast, &matWorldInv);
			for ( int i= 0; i < 4; ++i ) matDiff(i, i) -= 1;
		} else
			matDiff = matWorld - matWorld;

		matWorldLast = matWorld;
		useLast = true;

		D3DXMatrixTranspose(&matDiff, &matDiff);
		_ExecuteDX(pD3DDev->SetVertexShaderConstant(_DynamicBendMatrix, &matDiff(0, 0), 4));
	}
}

void HairPatch::draw() {

	pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_ADDSMOOTH);
	pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
	
	pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);
	pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
	pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);

	pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, FALSE); 
	pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE); 

	_ExecuteDX(pD3DDev->SetStreamSource(0, pVertexBuffer, sizeof(HairVertex)));

	pD3DDev->SetVertexShaderConstant(_Bend, bend, 1);
	pD3DDev->SetVertexShaderConstant(_Gravity, gravity, 1);
	_ExecuteDX(pD3DDev->SetVertexShader(hairShader));

	int nVert = nSegments*2 + 1;
	int iStartVert = 0;
	int nPrimitives = nVert - 2;

	// draw hairs in back to front order
	for ( std::vector<SortData>::iterator i = sortList.begin(); i != sortList.end(); ++i ) 
		_ExecuteDX(pD3DDev->DrawPrimitive(D3DPT_TRIANGLESTRIP, i->vertexIndex, nPrimitives));

	pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, TRUE); 
}


