/*----------------------------------------------------------------------
 * Project:		Rain	
 * File:		shader_rain.cpp
 * Description:	Rain animated VS Renderer
 * Created by:	Sebastien St-Laurent
 *				on 4/22/01
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Overview:
 *  This renderer uses an accelerated vertex shader program to render a "rain splash"
 * effect onto a mesh. The shader takes some time indexes as input and a rain direction
 * vector. From this information, the water splash sprite is scaled and alpha-blended properly
 * for create an animating rain-splash effect.
 * 
 * In this sample, a seperate vertex buffer is calculated (based on some random function and some
 * triangle suface area calculation). But if a mesh is suficently dense and non-uniform, the base mesh
 * of a model could be used as the source of rain-splashes.
 *
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *  NAME                   DATE      DESCRIPTION
 *  Sebastien St-Laurent   4/22/01   Created
 *  ins
 *----------------------------------------------------------------------*/

#include "eb_effect.h"
#include "nvinc.h"
#include "shader_rain.h"
#include <sys/types.h>
#include <sys/timeb.h>
using namespace std;

#define AREA_SPLASH_RATIO	100

DECLARE_EFFECT_MAIN()

extern "C"
{
	
	__declspec(dllexport) unsigned int GetNumEffects() { return 1; }
	
	__declspec(dllexport) EBEffect* CreateEffect(unsigned int EffectNum)
	{	
		switch(EffectNum)
		{
		case 0:
			return new CShaderRain();
			break;
		default:
			return NULL;
		}
	}
	
};

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::CShaderRain
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Constructor for the effect
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
CShaderRain::CShaderRain()
: m_bWireframe(false),
  m_pUI(NULL),
  m_bUseTrilinear(true),
  m_dwCurrentRainShader(0),
  m_dwCurrentShader(0)
{
	for(int i=0;i<NUM_MODELS;i++)
	{
		m_pVertexBuffer[i]=NULL;
		m_pIndexBuffer[i]=NULL;
		m_pShadeTexture[i]=NULL;
		m_pShadeRainTexture[i]=NULL;
		m_pRainVertexBuffer[i]=NULL;
		m_pAttributes[i]=NULL;
	}

	// Setup the effect for the browser
	m_strEffectLocation = "3rd Party\\Particles";
	m_strEffectName = "Rain";
	m_strEffectVertexShader = GetFilePath("base.nvv");
	m_strEffectRainVertexShader = GetFilePath("rain.nvv");
	m_strEffectPixelShader = "";
	m_strEffectVersion = "1.0";

	// Init our matrices
	D3DXMatrixIdentity(&m_world);
	D3DXMatrixIdentity(&m_view);
	D3DXMatrixIdentity(&m_proj);

	m_strFileName[0] = "tiger.x";
	m_strTextureName[0] = "fuzz3.dds";
	m_strRainTextureName[0] = "rain.bmp";
	m_nCurModel = 0;

}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::~CShaderRain
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Destructor for the effect
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
CShaderRain::~CShaderRain()
{
	Free(); 
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::UpdateProperties
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		This adds options to the menu (click on "File" or right-click
 *				on the display.
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
void CShaderRain::UpdateProperties()
{
	EBEffect::UpdateProperties();

	AddProperty(new EBProperty("Wireframe", OBJECT_MEMBER(m_bWireframe), EBTYPE_BOOL_PROP));

	// Vertex shaders
	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum, "Basic Rendering", GetFilePath("base.nvv"), EBTYPE_STRING_PROP));
	m_pVertexShaderEnum->AddEnumerant(new EBEnumValue(m_pVertexShaderEnum, "Rain Rendering", GetFilePath("rain.nvv"), EBTYPE_STRING_PROP));

    string aboutText = "file://";
    aboutText += GetFilePath("..\\..\\..\\NVEffectsExplained.htm");
    aboutText += "#rain";
	SetAboutInfo( NULL, _T("Rain"), _T(aboutText.c_str()));

}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::ConfirmDevice
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Performs caps-bit checking to make sure the selected device 
 *				supports this demo.  In this particular case we check for nothing!
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderRain::ConfirmDevice(D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format)
{
	if (!(pCaps->TextureCaps & D3DPTEXTURECAPS_MIPMAP))
	{
		m_strLastError = "Device does not support mipmaps!";
		return E_FAIL;
	}

	if (!(pCaps->TextureCaps & D3DPTFILTERCAPS_MIPFLINEAR))
	{
		m_bUseTrilinear = false;
	}

	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::Initialize
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Initialize render-states, vertex shader and vertex buffers
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderRain::Initialize(IDirect3DDevice8* pDev)
{
	HRESULT hr;
	
	// get the D3D device 
	m_pD3DDev = pDev;
	pDev->AddRef();
	
	//initialize mouse UI
	RECT rect;
	rect.left = rect.top = 0;
	D3DVIEWPORT8 viewport;
	m_pD3DDev->GetViewport(&viewport);
	rect.bottom = viewport.Height;
	rect.right  = viewport.Width;
	m_pUI = new MouseUI((const RECT)rect);
	
	vector<DWORD> Declaration;
	Declaration.push_back(D3DVSD_STREAM(0));
	Declaration.push_back(D3DVSD_REG(D3DVSDE_POSITION, D3DVSDT_FLOAT3)); // position
	Declaration.push_back(D3DVSD_REG(D3DVSDE_NORMAL, D3DVSDT_FLOAT3)); // normal
	Declaration.push_back(D3DVSD_REG(D3DVSDE_TEXCOORD0, D3DVSDT_FLOAT2)); // texture coords 0
	Declaration.push_back(D3DVSD_END());

	vector<DWORD> Declaration2;
	Declaration2.push_back(D3DVSD_STREAM(0));
	Declaration2.push_back(D3DVSD_REG(D3DVSDE_POSITION, D3DVSDT_FLOAT3)); // position
	Declaration2.push_back(D3DVSD_REG(D3DVSDE_NORMAL, D3DVSDT_FLOAT3)); // normal
	Declaration2.push_back(D3DVSD_REG(1, D3DVSDT_FLOAT1)); 
	Declaration2.push_back(D3DVSD_END());
	
	// Load the vertex shader object file and create it.
	m_dwCurrentShader = 0;
	hr = LoadAndCreateShader(GetFilePath("base.vso"), &Declaration[0], 0, SHADERTYPE_VERTEX, &m_dwCurrentShader);
	if (FAILED(hr))
		return hr;
	m_dwCurrentRainShader = 0;
	hr = LoadAndCreateShader(GetFilePath("rain.vso"), &Declaration2[0], 0, SHADERTYPE_VERTEX, &m_dwCurrentRainShader);
	if (FAILED(hr))
		return hr;
	
	
	for (int i=0;i<NUM_MODELS;i++)
	{
		hr = LoadXFile(i,GetFilePath( m_strFileName[i] ).c_str(), BaseVertex::FVF_Flags );
		if(FAILED(hr))
			return hr;
		CreateRainVertices(i);
	}
	
	// set base render states
	m_pD3DDev->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	m_pD3DDev->SetRenderState(D3DRS_LIGHTING, FALSE);
	m_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
	m_pD3DDev->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
	m_pD3DDev->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	
	//load textures
	for (i=0;i<NUM_MODELS;i++)
	{
		D3DXCreateTextureFromFile(m_pD3DDev, GetFilePath(m_strTextureName[i]).c_str(), &m_pShadeTexture[i]);
		D3DXCreateTextureFromFile(m_pD3DDev, GetFilePath(m_strRainTextureName[i]).c_str(), &m_pShadeRainTexture[i]);
	}
	
	
	// write the constant constants to constant memory now
	D3DXVECTOR4 lightDir(0.0f, 0.0f, -1.0f, 0.0f);
	D3DXVec4Normalize(&lightDir, &lightDir);
	
	m_pD3DDev->SetVertexShaderConstant(4, &lightDir, 1);
	
	D3DXVECTOR4 constNums(0.0f, 0.5f, 1.0f, 2.0f);
	m_pD3DDev->SetVertexShaderConstant(9, &constNums, 1);

	D3DXVECTOR4 constNums2(100.0f, 0.05f, 0.1f, 0.2f);
	m_pD3DDev->SetVertexShaderConstant(18, &constNums2, 1);
	
	D3DXVECTOR3 eye, lookAt, up;
	
	eye.x    = 0.0f; eye.y	  = 0.0f; eye.z	   = -3.5f;
	lookAt.x = 0.0f; lookAt.y = 0.0f; lookAt.z = 0.0f;
	up.x	 = 0.0f; up.y 	  = 1.0f; up.z	   = 0.0f;
	
	m_pD3DDev->SetVertexShaderConstant(10, &eye, 1);
	
	D3DXMatrixIdentity(&m_world);
	D3DXMatrixLookAtLH(&m_view, &eye, &lookAt, &up);
	
	D3DXMatrixPerspectiveFovLH(&m_proj,
		D3DXToRadian(60.0f),
		1,
		1.0f,
		50.0f);
	m_pD3DDev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX*)&m_proj);
	
	SetVertexShaderMatrices();
	
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::SetVertexShaderMatrices
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Sets up the matrices used by the vertex shader
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderRain::SetVertexShaderMatrices()
{
	D3DXMATRIX mvpMat, worldITMat, worldMat;
	
	D3DXMATRIX tempMat;
	D3DXMatrixMultiply( &tempMat, &m_pUI->GetRotationMatrix(), &m_pUI->GetTranslationMatrix() );

	D3DXMatrixTranspose(&worldMat, &tempMat);

	D3DXMatrixIdentity(&mvpMat);
	
	D3DXMatrixMultiply(&mvpMat, &tempMat, &m_view);
	
	D3DXMatrixInverse(&worldITMat, NULL, &tempMat );
	
	D3DXMatrixMultiply(&mvpMat, &mvpMat, &m_proj);
	D3DXMatrixTranspose(&mvpMat, &mvpMat);
	
	m_pD3DDev->SetVertexShaderConstant(0, &mvpMat, 4);
	
	m_pD3DDev->SetVertexShaderConstant(5, &worldITMat, 4);
	m_pD3DDev->SetVertexShaderConstant(11, &worldMat, 4);
	
	return S_OK;
}


/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::PropertyUpdateCallback
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Called when some property is changed from the UI
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
void CShaderRain::PropertyUpdateCallback(const EBProperty* pProperty, bool bWritten)
{
	if (!bWritten)
		return;

	EBString name;
	if (pProperty->IsKindOf(EBTYPE_TRIGGER_PROP))
	{

		name = pProperty->GetPropertyName();
	
	}
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::Free
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Called when we switch away from this demo.	Free up all the 
 *				memory that is in use.
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderRain::Free()
{
	for (int i=0;i<NUM_MODELS;i++)
	{
		SAFE_RELEASE(m_pVertexBuffer[i]);
		SAFE_RELEASE(m_pIndexBuffer[i]);
		SAFE_RELEASE(m_pShadeTexture[i]);
		SAFE_RELEASE(m_pShadeRainTexture[i]);
		SAFE_RELEASE(m_pRainVertexBuffer[i]);
		SAFE_DELETE_ARRAY(m_pAttributes[i]);
	}
	
	if (m_pD3DDev)
	{
		m_pD3DDev->DeleteVertexShader(m_dwCurrentRainShader);
		m_pD3DDev->DeleteVertexShader(m_dwCurrentShader);
		m_dwCurrentRainShader = 0;
		m_dwCurrentShader = 0;
		SAFE_RELEASE(m_pD3DDev);
	}
	delete m_pUI, m_pUI = 0;
	
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::LoadXFile
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Loads a mesh from a .X file
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderRain::LoadXFile(int idx,const char* fileName, const DWORD dwFVF)
{
	ID3DXMesh *tempMesh, *tempMeshFVF, *tempMeshOpt;
		
	HRESULT hr = D3DXLoadMeshFromX(const_cast<char*>(fileName), D3DXMESH_SYSTEMMEM , m_pD3DDev, NULL, 
		NULL, &m_dwNumSections[idx], &tempMesh);
	
	//convert to our format
	hr = tempMesh->CloneMeshFVF(D3DXMESH_DYNAMIC, dwFVF, m_pD3DDev, &tempMeshFVF);

	D3DXComputeNormals(tempMeshFVF);

	tempMeshFVF->Optimize(D3DXMESHOPT_ATTRSORT, NULL, NULL, NULL, NULL, &tempMeshOpt);
	DWORD attribSize = m_dwNumSections[idx];
	hr = tempMeshOpt->GetAttributeTable(NULL, &attribSize);

	m_numFace[idx] = tempMeshOpt->GetNumFaces();

	SAFE_DELETE_ARRAY( m_pAttributes[idx] );

	m_pAttributes[idx] = new D3DXATTRIBUTERANGE[attribSize];
	
	hr = tempMeshOpt->GetAttributeTable(m_pAttributes[idx], &attribSize);


	SAFE_RELEASE( m_pVertexBuffer[idx] );
	SAFE_RELEASE( m_pIndexBuffer[idx] );

	tempMeshOpt->GetVertexBuffer(&m_pVertexBuffer[idx]);
	tempMeshOpt->GetIndexBuffer(&m_pIndexBuffer[idx]);

	SAFE_RELEASE(tempMesh);
	SAFE_RELEASE(tempMeshFVF);
	SAFE_RELEASE(tempMeshOpt);
		
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::Start
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Called when the effect is inovked
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderRain::Start()
{
	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::SetTextureStageStates
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		Sets the texture renderstates required to render our mesh
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
HRESULT CShaderRain::SetTextureStageStates(bool rain)
{
	if (rain==false)
	{
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
		m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		m_pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);

		m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
		m_pD3DDev->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
		m_pD3DDev->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	}
	else
	{
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
		m_pD3DDev->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
		m_pD3DDev->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
		m_pD3DDev->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );

		m_pD3DDev->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE );
		m_pD3DDev->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE );
	}

	return S_OK;
}

/*----------------------------------------------------------------------------
 * FUNCTION:		CShaderRain::Tick
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Purpose:		This function is called every frame.  Here is where we write
 *				and update constants, clear the z- and back-buffer and render our 
 *				primitives
 * Arguments:	
 * Returns:		
 * NOTES:		
 *----------------------------------------------------------------------------*/
inline DWORD FtoDW( FLOAT f ) { return *((DWORD*)&f); }
HRESULT CShaderRain::Tick(EBTimer* pTimer)
{
	HRESULT hr = S_OK;
	static _timeb   start;
	static int	firstTime = 1;
	int i;

	if (firstTime!=0)
	{
		_ftime(&start);
		firstTime = 0;
	}

    struct _timeb   now;
    _ftime(&now);
    float deltaTime(static_cast<float>(now.time - start.time));
    deltaTime += static_cast<float>(0.001f * (now.millitm - start.millitm));
	deltaTime *= 5.0f;

	// Set V and IDX buffers
	m_pD3DDev->SetIndices( m_pIndexBuffer[m_nCurModel], 0);
	hr = m_pD3DDev->SetStreamSource(0, m_pVertexBuffer[m_nCurModel], sizeof(BaseVertex));
	if (FAILED(hr))
		return hr;

	SetVertexShaderMatrices();

	// select the just created vertex shader as the thing to transform and light vertices
	hr = m_pD3DDev->SetVertexShader(m_dwCurrentShader);
	if (FAILED(hr))
		return hr;

	m_pD3DDev->SetRenderState(D3DRS_POINTSPRITEENABLE, FALSE);

	// Set texture to use
	m_pD3DDev->SetTexture(0, m_pShadeTexture[m_nCurModel]);
	SetTextureStageStates(false);


	m_pD3DDev->SetRenderState(D3DRS_FILLMODE, (m_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID));
	m_pD3DDev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0x002B2B2B, 1.0f, 0L);

	// Draw base mesh
	for(i = 0; i < m_dwNumSections[m_nCurModel]; i++)
	{
		m_pD3DDev->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 
										 m_pAttributes[m_nCurModel][i].VertexStart,
										 m_pAttributes[m_nCurModel][i].VertexCount, 
											m_pAttributes[m_nCurModel][i].FaceStart*3, 
											m_pAttributes[m_nCurModel][i].FaceCount );
	}

	// Set point sprite params for rain layer
	m_pD3DDev->SetRenderState( D3DRS_POINTSIZE_MIN, FtoDW(0.1f) );
	m_pD3DDev->SetRenderState( D3DRS_POINTSIZE_MAX, FtoDW(64.0f) );
	m_pD3DDev->SetRenderState( D3DRS_POINTSIZE,     FtoDW(32.0f) );
	m_pD3DDev->SetRenderState(D3DRS_POINTSPRITEENABLE, true);
    m_pD3DDev->SetRenderState( D3DRS_POINTSCALEENABLE,  false );

	SetTextureStageStates(true);

	hr = m_pD3DDev->SetStreamSource(0, m_pRainVertexBuffer[m_nCurModel], sizeof(RainVertex));
	if (FAILED(hr))
		return hr;

	hr = m_pD3DDev->SetVertexShader(m_dwCurrentRainShader);
	if (FAILED(hr))
		return hr;
	m_pD3DDev->SetTexture(0, m_pShadeRainTexture[m_nCurModel]);
	m_pD3DDev->SetVertexShaderConstant(17, D3DXVECTOR4(0.40f, 0.9f, 0.0f, 0.0f), 1);		// Rain direction
	m_pD3DDev->SetVertexShaderConstant(19, D3DXVECTOR4(deltaTime, 0.0f, 0.0f, 0.0f), 1);	// Anim Time index
	m_pD3DDev->SetRenderState(D3DRS_FILLMODE, (m_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID));
	m_pD3DDev->DrawPrimitive( D3DPT_POINTLIST, 0, m_dwRainVerts[m_nCurModel]);

	return hr;
}


// This function computes the surface area of a triangle
float CalcSurfaceArea(BaseVertex* pFV[3])
{
	float s;
	float len[3];

	for (int i=0;i<3;i++)
	{
		float x,y,z;
		x = pFV[(i+1)%3]->x - pFV[i]->x;
		y = pFV[(i+1)%3]->y - pFV[i]->y;
		z = pFV[(i+1)%3]->z - pFV[i]->z;
		len[i] = sqrt(x*x+y*y+z*z);
	}
	
	s = (len[0]+len[1]+len[2])/2.0f;
	return sqrt(s*(s-len[0])*(s-len[1])*(s-len[2]));
}

void CShaderRain::CreateRainVertices(int idx)
{
	int attribs = m_dwNumSections[idx];
	RainVertex* pRainVert;
	BaseVertex* pFuzzVert;
	WORD*		pFuzzIdx;
	BaseVertex* Array[3];
	float area=0.0f;

	// Lock the mesh VB and gather the info we need
	m_pVertexBuffer[idx]->Lock(0,0,(BYTE**)&pFuzzVert,0);
	m_pIndexBuffer[idx]->Lock(0,0,(BYTE**)&pFuzzIdx,0);
	for(int attrs=0;attrs<m_dwNumSections[idx];attrs++)
	{
		WORD*	pFuzzCurIdx = pFuzzIdx + (m_pAttributes[idx][attrs].FaceStart*3);
		for(int j=0;j<m_pAttributes[idx][attrs].FaceCount;j++)
		{
			Array[0] = &pFuzzVert[pFuzzCurIdx[j*3+0]];
			Array[1] = &pFuzzVert[pFuzzCurIdx[j*3+1]];
			Array[2] = &pFuzzVert[pFuzzCurIdx[j*3+2]];
			area += CalcSurfaceArea(Array);
		}
	}
	m_pVertexBuffer[idx]->Unlock();
	m_pIndexBuffer[idx]->Unlock();

	// Number of vertices is proprotional to surface area. This way the disribution of rain
	// splashes should be roughly uniform.
	m_dwRainVerts[idx] = area*AREA_SPLASH_RATIO;

	// Create our rain container buffer
	m_pD3DDev->CreateVertexBuffer(m_dwRainVerts[idx]*sizeof(RainVertex),0,0,D3DPOOL_MANAGED,&m_pRainVertexBuffer[idx]);

	// Lock it and fill it
	area = 0.0f;
	int vert = 0;
	m_pRainVertexBuffer[idx]->Lock(0,m_dwRainVerts[idx]*sizeof(RainVertex),(BYTE**)&pRainVert,0);
	m_pVertexBuffer[idx]->Lock(0,0,(BYTE**)&pFuzzVert,0);
	m_pIndexBuffer[idx]->Lock(0,0,(BYTE**)&pFuzzIdx,0);
	for(attrs=0;attrs<m_dwNumSections[idx];attrs++)
	{
		WORD*	pFuzzCurIdx = pFuzzIdx + (m_pAttributes[idx][attrs].FaceStart*3);
		for(int j=0;j<m_pAttributes[idx][attrs].FaceCount;j++)
		{
			Array[0] = &pFuzzVert[pFuzzCurIdx[j*3+0]];
			Array[1] = &pFuzzVert[pFuzzCurIdx[j*3+1]];
			Array[2] = &pFuzzVert[pFuzzCurIdx[j*3+2]];
			area += AREA_SPLASH_RATIO*CalcSurfaceArea(Array);

			// Here we create random vertices in the triangle. The randomness is defines
			// based on 2 edge vector of the triangle. We compute a ratio r1 and r2 which represent a fraction
			// of the edge. If we compute the point as pt = r1*v1 + r2*v2 and we satisfy r1+r2 <= 1.0, the point
			// will be within the triangle.
			while (area>1.0f)
			{
				float r1,r2;
				D3DXVECTOR3 s,t;
				// Create new vertex
				r1 = ((float)(rand()%1000)) / 1000.0f;
				r2 = 1.0f-r1;
				r2 = r2 * ((rand()%1000) / 1000.0f);
				s.x = Array[1]->x - Array[0]->x;
				s.y = Array[1]->y - Array[0]->y;
				s.z = Array[1]->z - Array[0]->z;
				t.x = Array[2]->x - Array[0]->x;
				t.y = Array[2]->y - Array[0]->y;
				t.z = Array[2]->z - Array[0]->z;
				pRainVert->x = Array[0]->x + s.x*r1 + t.x*r2;
				pRainVert->y = Array[0]->y + s.y*r1 + t.y*r2;
				pRainVert->z = Array[0]->z + s.z*r1 + t.z*r2;
				pRainVert->nx = (Array[0]->nx + Array[1]->nx + Array[2]->nx)/3.0f;
				pRainVert->ny = (Array[0]->ny + Array[1]->ny + Array[2]->ny)/3.0f;
				pRainVert->nz = (Array[0]->nz + Array[1]->nz + Array[2]->nz)/3.0f;
				pRainVert->loop = (100+rand()%900) / 100.0f;

				area -= 1.0f;pRainVert++;vert++;
			}
		}
	}
	m_pVertexBuffer[idx]->Unlock();
	m_pIndexBuffer[idx]->Unlock();
	m_pRainVertexBuffer[idx]->Unlock();
}


void CShaderRain::MouseButton(HWND hWnd, eButtonID button, bool bDown, int x, int y)
{
	if(button == MOUSE_LEFTBUTTON)
	{
		if(bDown)
		{
			m_pUI->OnLButtonDown(x, y);
		}
		else
		{
			m_pUI->OnLButtonUp(x, y);
		}
	}
	return;
}

void CShaderRain::MouseMove(HWND hWnd, int x, int y)
{
	m_pUI->OnMouseMove(x, y);
	return;
}

void CShaderRain::Keyboard(DWORD dwKey, UINT nFlags, bool bDown)
{
    if ( bDown )
    {
        switch ( dwKey )
        {

            case 'W' :
            {
                m_bWireframe = !m_bWireframe;
				m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
                break;
            }

            case 'R' :
            {
                m_bWireframe = false;
				m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
                break;
            }

            case VK_HOME :
            case VK_END :
			case VK_NUMPAD7 :
			case VK_NUMPAD1 :
            case '7' :
            case '1' :
            {
                m_pUI->Reset();
                m_bWireframe = false;
				m_dwEffectDirtyFlags |= EBEFFECT_DIRTY_PUBLICSTATE;
                break;
            }

            case VK_LEFT :
			case VK_NUMPAD4 :
            case '4' :
            {
                m_pUI->Translate( -0.10f, 0.0f, 0.0f );
                break;
            }

            case VK_RIGHT :
			case VK_NUMPAD6 :
            case '6' :
            {
                m_pUI->Translate(  0.10f, 0.0f, 0.0f );
                break;
            }

            case VK_UP :
			case VK_NUMPAD8 :
            case '8' :
            {
                m_pUI->Translate(  0.0f, 0.0f, -0.1f );
                break;
            }

            case VK_DOWN :
			case VK_NUMPAD2 :
            case '2' :
            {
                m_pUI->Translate( 0.0f, 0.0f, 0.1f );
                break;
            }

            case VK_PRIOR :
			case VK_NUMPAD9 :
            case '9' :
            {
                m_pUI->Translate(  0.0f, 0.1f, 0.0f );
                break;
            }
            case VK_NEXT :
			case VK_NUMPAD3 :
            case '3' :
            {
                m_pUI->Translate(  0.0f, -0.1f, 0.0f );
                break;
            }

            case VK_ADD :
            {
                m_pUI->Translate( 0.0f, 0.0f, 1.0f );
                break;
            }
            case VK_SUBTRACT :
            {
                m_pUI->Translate( 0.0f, 0.0f, -1.0f );
                break;
            }

            default :
            {
                break;
            }
        };
    }
	return;
}